/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/** \mainpage

    \hline
    \section dcsect Quick Reference to User's Guide
    - \ref hadronic
    - \ref quark_matter
    - \ref tovtoc
    - \ref cnstar
    - \ref nrTeos_section
    - \ref eossource_section
    - \ref eosref_section

    \htmlonly
    <a href="../latex/refman.pdf">PDF documentation</a>
    \endhtmlonly

    \hline
    \section hadronic Hadronic equations of state

    The hadronic equations of state are all inherited from
    hadronic_eos: schematic_eos, skyrme_eos, rmf_eos, apr_eos, and
    gen_potential_eos. 

    hadronic_eos includes several methods that can be used
    to calculate the saturation properties of nuclear matter.
    These methods are sometimes overloaded in descendants when
    exact formulas are available.

    There is also a set of classes to modify the quartic term
    of the symmetry energy: rmf4_eos, apr4_eos, skyrme4_eos, and 
    mdi4_eos all based on sym4_eos_base which can be used
    in sym4_eos.

    \hline
    \section quark_matter Equations of state of quark matter

    The equations of state of \quark matter are all inherited from
    quark_eos: bag_eos is a simple bag model, nambujl_eos is the
    Nambu--Jona-Lasinio model.

    \hline
    \section tovtoc Solution of the Tolman-Oppenheimer-Volkov equations

    The class \ref tov_solve provide a solution to the
    Tolman-Oppenheimer-Volkov (TOV) equations given an equation of
    state. This is particularly useful for static neutron star
    structure: given any equation of state one can calculate the mass
    vs. radius curve and the properties of any star of a given
    mass. An adaptive integration is employed and calculates the
    gravitational mass, the baryonic mass (if the baryon density is
    supplied), and the gravitational potential.  The remaining columns
    is the equation of state are also interpolated into the solution,
    e.g. if a chemical potential is given, then the radial dependence
    of the chemical potential for a 1.4 solar mass star can be
    automatically computed. The equation of state may be specified in
    arbitrary units so long as an appropriate conversion factor is
    supplied. An equation of state for low densities (baryon density <
    0.08 \f$ \mathrm{fm}^{-3} \f$ ) is provided and can be
    automatically appended to the user-defined equation of state.

    This is still experimental.

    \hline
    \section cnstar Naive Cold Neutron Stars

    There is also a class to calculate zero-temperature neutron
    stars: \ref cold_nstar. It uses \ref tov_solve to compute
    the structure, given a hadronic equation of state (of type
    \ref hadronic_eos). It also computes the adiabatic index, the
    speed of sound, and determines the possibility of the
    direct Urca process as a function of density or radius.

    This is still experimental.
    
    \hline
    \section nrTeos_section Non-relativistic Finite Temperature Approximations
    
    This is taken from the \ref Prakash87.

    The entropy is 
    \f[
    s = -\sum_k \left[ n_k \ln n_k + \left(1-n_k\right) \ln
    \left(1-n_k\right)\right]
    \f]

    The low-temperature (degenerate) approximation to the entropy is
    \f[
    s=\pi^2/3 N(0) T
    \f]
    where the density of states at the Fermi surface is
    \f[
    N(0)=\sum_k \delta(\epsilon_k -\mu) = \frac{3 \rho }{k_F v_F}
    \f]
    where the Fermi velocity is 
    \f[
    v_F = \left.\frac{\partial \epsilon_k}{\partial k}\right|_{k_F}
    = \frac{k_F}{m^{*}}
    \f]
    The latter equation defines the effective mass.
    The level density parameter is given by 
    \f[
    a=\frac{\pi^2 N(0)}{6 \rho}
    \f]
    Defining the Fermi temperature:
    \f[
    T_F=\frac{1}{2} k_F v_F = k_F^2/2/m^{*}
    \f]
    another expression for the entropy is
    \f[
    s=\frac{\pi^2}{2} \rho (T/T_F)
    \f]
    Expressions for the remaining quantities are
    \f[
    P=P(T=0)+\frac{\rho}{3} 
    a T^2 \left(1+\frac{d \ln v_F}{d \ln k_F }\right)
    \f]
    \f[
    E/A=E/A(T=0)+a T^2
    \f]
    \f[
    \mu=\mu(T=0)-\frac{1}{3} a T^2\left(2-\frac{d \ln v_F}{d \ln k_F }\right)
    \f]
    Typically, the leading correction to 
    \f[
    s=\frac{\pi^2}{2} \rho (T/T_F)
    \f]
    is of order  \f$ (T/T_F)^2 \f$  unless soft collective modes give
    rise to a  \f$ (T/T_F)^3 \ln (T/T_F) \f$  correction.

    At high temperature (non-degenerate approximation), 
    a stationary phase approximation gives
    \f{eqnarray*}
    \rho(T) &\sim& \frac{\gamma}{2 \pi^2} e^{\mu/T} \cdot k^2
    e^{-\epsilon_k/T} \sqrt{2 \pi} \left[\frac{2}{k^2}+
    \frac{1}{T}\frac{\partial v_k}{\partial k}\right]^{-1/2} \\
    &=& e^{\mu/T} \left.f(T)\right|_{k=k_{\rho}}
    \f}
    where  \f$ \gamma \f$  is the spin and isospin degeneracy and the
    velocity function is  \f$ v_k=\partial \epsilon_k/\partial k \f$ .
    The function  \f$ f(T) \f$  is evaluated at momentum  \f$ k_{\rho} \f$ 
    which is obtained by solving  \f$ T-k v_k/2=0 \f$ .
    The chemical potential is obtained by inverting the above
    relation for  \f$ \rho(T) \f$ :
    \f[
    \mu \sim T \ln \rho - T \ln f(T)
    \f]
    From this value of  \f$ \mu \f$  we can derive the entropy density
    using
    \f[
    T s \sim \sum_k n_k \epsilon_k + \rho T - \mu \rho
    \f]
    Using the stationary phase method:
    \f{eqnarray*}
    \sum_k n_k \epsilon_k &\sim& \frac{\gamma}{2 \pi^2} e^{\mu/T}
    \cdot k^2 \epsilon_k e^{-\epsilon_k/T} \sqrt{2 \pi}
    \left[\frac{2}{k^2}-\left(\frac{1}{\epsilon_k}-\frac{1}{T}\right)
    \frac{\partial v_k}{\partial k}+
    \left(\frac{v_k}{\epsilon_k}\right)^2\right]^{-1/2} \\
    &=& e^{\mu/T} \left.g(T)\right|_{k=k_E}
    \f}
    where  \f$ k_E \f$  is the solution of 
    \f[
    \frac{2}{k}+v_k \left(\frac{1}{\epsilon_k}-\frac{1}{T}\right)=0
    \f]
    This provides a first approximation to the energy and together
    with the thermodynamic identity gives the pressure.

    \hline
    \section eossource_section Example source code

    \subsection exlist_eos_subsect Example list

    - \ref ex_cold_nstar_sect

    \subsection ex_cold_nstar_sect Cold neutron star example

    \dontinclude ex_cold_nstar.cpp
    \skip Example:
    \until End of example

    \hline
    \section todo_section Other Todos
    
    \future Right now, the equation of state classes depend on the
    user to input the correct value of \c non_interacting
    for the particle inputs. This is not very graceful...

    \hline
    \section eosref_section Bibliography

    Some of the references which contain links should direct you to
    the work referred to directly through dx.doi.org.

    \anchor Akmal98 Akmal98:
    \htmlonly
    <a href="http://dx.doi.org/10.1103/PhysRevC.58.1805">
    Akmal, Pandharipande, and Ravenhall</a>, 
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1103/PhysRevC.58.1805}{
    Akmal, Pandharipande, and Ravenhall},
    \endlatexonly
    Phys. Rev. C \b 58, 1805 (1998).

    \anchor Bartel79 Bartel79:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0375-9474(82)90403-1">
    J. Bartel, P. Quentin, M. Brack, C. Guet, and H&aring;kansson</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0375-9474(82)90403-1}{
    J. Bartel, P. Quentin, M. Brack, C. Guet, and H{\aa}kansson},
    \endlatexonly
    Nucl. Phys. A \b 386 (1982) 79.

    \anchor Baym71 Baym71:
    G. Baym, C. Pethick, and P. Sutherland, Astrophys. J. \b 170 (1971) 299.
    
    \anchor Beiner75 Beiner75:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0375-9474(75)90338-3">
    M. Beiner, H. Flocard, Nguyen van Giai, and P. Quentin</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0375-9474(75)90338-3}{
    M. Beiner, H. Flocard, Nguyen van Giai, and P. Quentin}, 
    \endlatexonly
    Nucl. Phys. A \b 238 (1975) 29.

    \anchor Bernard88 Bernard88:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0550-3213(88)90127-7">
    V. Bernard, R.L. Jaffe, and U.-G. Meissner</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0550-3213(88)90127-7}{
    V. Bernard, R.~L. Jaffe, and U.-G. Meissner},
    \endlatexonly
    Nucl. Phys. B \b 308 (1988) 753.

    \anchor Bombaci01 Bombaci01:
    I. Bombaci, "Equation of State for Dense Isospin Asymmetric 
    Nuclear Matter for Astrophysical Applications Equation of 
    State for Isospin-Asymmetric Nuclear Matter and Neutron Star
    Properties", in "Isospin physics in heavy-ion collisions at
    intermediate energies" 
    \htmlonly
    ed. by B-A. Li and W. U. Schr&ouml;der (2001) Nova Science, New York. 
    \endhtmlonly
    \latexonly
    ed. by B-A. Li and W. U. Schr\"{o}der (2001) Nova Science, New York. 
    \endlatexonly
    
    \anchor Brack85 Brack85:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0370-1573(86)90078-5">
    M. Brack, C. Guet, and H.-B. H&aring;kansson</a>
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0370-1573(86)90078-5}{
    M. Brack, C. Guet, and H.-B. H{\aa}kansson},
    \endlatexonly
    Phys. Rep. \b 123 (1985) 275.

    \anchor Buballa99 Buballa99:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/S0370-2693(99)00533-X">
    M. Buballa and M. Oertel</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/S0370-2693(99)00533-X}{
    M. Buballa and M. Oertel}, 
    \endlatexonly
    Phys. Lett. B \b 457 (1999) 261.

    \anchor Buballa04 Buballa04:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/j.physrep.2004.11.004">
    M. Buballa</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/j.physrep.2004.11.004}{
    M. Buballa}, 
    \endlatexonly
    Phys. Rep. \b 407 (2005) 205-376.
    
    \anchor Chabanat95 Chabanat95:
    E. Chabanat, Ph. D. Thesis 1995

    \anchor Chabanat97 Chabanat97:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/S0375-9474(97)00596-4">
    E. Chabanat, J. Meyer, P. Bonche, R. Schaeffer, and P. Haensel</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/S0375-9474(97)00596-4}{
    E. Chabanat, J. Meyer, P. Bonche, R. Schaeffer, and P. Haensel}, 
    \endlatexonly
    Nucl. Phys. A \b 627 (1997) 710.

    \anchor Danielewicz08 Danielewicz08:
    \htmlonly
    <a href="http://arxiv.org/abs/0807.3743">
    P. Danielewicz, J. Lee</a>,
    \endhtmlonly
    \latexonly
    \href{http://arxiv.org/abs/0807.3743}{
    P. Danielewicz, J. Lee},
    \endlatexonly
    arXiv.org:0807.3743.

    \anchor Das03 Das03:
    \htmlonly
    <a href="http://dx.doi.org/10.1103/PhysRevC.67.034611">
    C.B. Das, S. Das Gupta, C. Gale, and B.-A. Li</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1103/PhysRevC.67.034611}{
    C.B. Das, S. Das Gupta, C. Gale, and B.-A. Li},
    \endlatexonly
    Phys. Rev. C \b 67 (2003) 034611.

    \anchor Dobaczewski94 Dobaczewski94:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0375-9474(84)90433-0">
    J. Dobaczewski, H. Flocard, and J. Treiner</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0375-9474(84)90433-0}{
    J. Dobaczewski, H. Flocard, and J. Treiner}, 
    \endlatexonly
    Nucl. Phys. A \b 422 (1984) 103.

    \anchor Dutta86 Dutta86:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0375-9474(86)90283-6">
    A.K.Dutta, J.-P.Arcoragi, J.M.Pearson, R.Behrman, F.Tondeur</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0375-9474(86)90283-6}{
    A.K.Dutta, J.-P.Arcoragi, J.M.Pearson, R.Behrman, F.Tondeur},
    \endlatexonly
    Nucl. Phys. A \b 458 (1986) 77.

    \anchor Friedrich86 Friedrich86:
    \htmlonly
    <a href="http://dx.doi.org/10.1103/PhysRevC.33.335">
    J. Friedrich, and P.-G. Reinhard</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1103/PhysRevC.33.335}{
    J. Friedrich, and P.-G. Reinhard}, 
    \endlatexonly
    Phys. Rev. C \b 33 (1986) 335.

    \anchor Gaitanos04 Gaitanos04:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/j.nuclphysa.2003.12.001">
    T. Gaitanos, M. Di Toro, S. Typel, V. Barana, C. Fuchs, 
    V. Greco, and d H. H. Wolter</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/j.nuclphysa.2003.12.001}{
    T. Gaitanos, M. Di Toro, S. Typel, V. Barana, C. Fuchs, 
    V. Greco, and d H. H. Wolter},
    \endlatexonly
    Nucl. Phys. A \b 732 (2004) 24.

    \anchor Gale87 Gale87:
    \htmlonly
    <a href="http://dx.doi.org/10.1103/PhysRevC.35.1666">
    C. Gale, G. Bertsch, and S. Das Gupta</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1103/PhysRevC.35.1666}{
    C. Gale, G. Bertsch, and S. Das Gupta},
    \endlatexonly
    Phys. Rev. C \b 35 (1986) 1666.

    \anchor Hatsuda94 Hatsuda94:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0370-1573(94)90022-1">
    T. Hatsuda and T. Kunihiro</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0370-1573(94)90022-1}{
    T. Hatsuda and T. Kunihiro}, 
    \endlatexonly
    Phys. Rep. \b 247 (1994) 221.

    \anchor Horowitz01 Horowitz01: 
    \htmlonly
    <a href="http://dx.doi.org/10.1103/PhysRevLett.86.5647">
    C.J. Horowitz and J. Piekarewicz</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1103/PhysRevLett.86.5647}{
    C.~J. Horowitz and J. Piekarewicz},
    \endlatexonly
    Phys. Rev. Lett. \b 86 (2001) 5647.

    \anchor Kubis97 Kubis97: 
    \htmlonly
    <a href="http://dx.doi.org/10.1016/S0370-2693(97)00306-7">
    S. Kubis and M. Kutschera</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/S0370-2693(97)00306-7}{
    S. Kubis and M. Kutschera},
    \endlatexonly
    Phys. Lett. B \b 399 (1997) 191.

    \anchor Margueron02 Margueron02:
    \htmlonly
    <a href="http://dx.doi.org/10.1103/PhysRevC.66.014303">
    J. Margueron, J. Navarro, and N.V. Giai</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1103/PhysRevC.66.014303}{
    J. Margueron, J. Navarro, and N.~V. Giai},
    \endlatexonly
    Phys. Rev. C \b 66 (2002) 014303.

    \anchor Muller96 Muller96: 
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0375-9474(96)00187-X">
    H. Muller and B. D. Serot</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0375-9474(96)00187-X}{
    H. Muller and B.~D. Serot}, 
    \endlatexonly
    Nucl. Phys. A \b 606 (1996), 508.

    \anchor Osni94 Osni94:
    \htmlonly
    <a href="http://dx.doi.org/10.1103/PhysRevC.50.460">
    M. Onsi, H. Przysiezniak, J.M. Pearson</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1103/PhysRevC.50.460}{
    M. Onsi, H. Przysiezniak, J.M. Pearson},
    \endlatexonly
    Phys. Rev. C \b 50 (1994) 460.

    \anchor Pethick95 Pethick95:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0375-9474(94)00506-I">
    C.J. Pethick, D.G. Ravenhall, and C.P. Lorenz</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0375-9474(94)00506-I}{
    C.~J. Pethick, D.~G. Ravenhall, and C.~P. Lorenz}, 
    \endlatexonly
    Nucl. Phys. A \b 584 (1995) 675.

    \anchor Prakash87 Prakash87:
    \htmlonly
    M. Prakash, T.L. Ainsworth, J.P. Blaizot, and H. Wolter,
    in Windsurfing the Fermi Sea, Volume II
    edited by T.T.S. Kuo and J. Speth, Elsevier 1987, pg. 357.
    \endhtmlonly
    \latexonly
    M. Prakash, T.~L. Ainsworth, J.~P. Blaizot, and H. Wolter,
    in Windsurfing the Fermi Sea, Volume II
    edited by T.~T.~S. Kuo and J. Speth, Elsevier 1987, pg. 357.
    \endlatexonly

    \anchor Prakash88 Prakash88:
    \htmlonly
    <a href="http://dx.doi.org/10.1103/PhysRevLett.61.2518">
    M. Prakash, T. L. Ainsworth, and J. M. Lattimer</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1103/PhysRevLett.61.2518}{
    M. Prakash, T.~L. Ainsworth, and J.~M. Lattimer},
    \endlatexonly
    Phys. Rev. Lett. \b 61 (1988) 2518.

    \anchor Prakash97 Prakash97:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/S0370-1573(96)00023-3">
    Madappa Prakash, I. Bombaci, Manju Prakash, P.J. Ellis, J.M. Lattimer,
    and R. Knorren</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/S0370-1573(96)00023-3}{
    Madappa Prakash, I. Bombaci, Manju Prakash, P.~J. Ellis, J.~M. Lattimer,
    and R. Knorren}, 
    \endlatexonly
    Phys. Rep. \b 280 (1997) 1.

    \anchor Reinhard95 Reinhard95:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0375-9474(94)00770-N">
    P.-G. Reinhard and H. Flocard</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0375-9474(94)00770-N}{
    P.-G. Reinhard and H. Flocard}, 
    \endlatexonly
    Nucl. Phys. A \b 584 (1995) 467.

    \anchor Reinhard99 Reinhard99: 
    \htmlonly
    <a href="http://dx.doi.org/10.1103/PhysRevC.60.014316">
    P.-G. Reinhard, D.J. Dean, W. Nazarewicz, J. Dobaczewski, 
    J.A. Maruhn, M.R. Strayer</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1103/PhysRevC.60.014316}{
    P.-G. Reinhard, D.J. Dean, W. Nazarewicz, J. Dobaczewski, 
    J.A. Maruhn, M.R. Strayer},
    \endlatexonly
    Phys. Rev. C \b 60, (1999) 014316.
    
    \anchor Shapiro83 Shapiro83:
    S. L. Shapiro and S. A. Teukolsky, "Black Holes, White Dwarfs,
    and Neutron Stars: The Physics of Compact Objects", John
    Wiley and Sons, New York, 1983.

    \anchor Steiner00 Steiner00:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/S0370-2693(00)00780-2">
    A. W. Steiner, M. Prakash, and J. M. Lattimer</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/S0370-2693(00)00780-2}{
    A. W. Steiner, M. Prakash, and J.M. Lattimer},
    \endlatexonly
    Phys. Lett. B, \b 486 (2000) 239.

    \anchor Steiner02 Steiner02:
    \htmlonly
    <a href="http://dx.doi.org/10.1103/PhysRevD.66.094007">
    A. W. Steiner, S. Reddy, and M. Prakash</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1103/PhysRevD.66.094007}{
    A. W. Steiner, S. Reddy, and M. Prakash},
    \endlatexonly
    Phys. Rev. D, \b 66 (2002) 094007.

    \anchor Steiner05 Steiner05:
    \htmlonly
    <a href="http://dx.doi.org/10.1103/PhysRevD.72.054024">
    A. W. Steiner</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1103/PhysRevD.72.054024}{
    A. W. Steiner},
    \endlatexonly
    Phys. Rev. D, \b 72 (2005) 054024.

    \anchor Steiner06 Steiner06:
    \htmlonly
    <a href="http://dx.doi.org/10.1103/PhysRevC.74.045808">
    A. W. Steiner</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1103/PhysRevC.74.045808}{
    A. W. Steiner},
    \endlatexonly
    Phys. Rev. C, \b 74 (2006) 045808.

    \anchor Tondeur84 Tondeur84:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0375-9474(84)90444-5">
    F. Tondeur, M. Brack, M. Farine, and J.M. Pearson</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0375-9474(84)90444-5}{
    F. Tondeur, M. Brack, M. Farine, and J.M. Pearson}, 
    \endlatexonly
    Nucl. Phys. A \b 420 (1984) 297.

    \anchor Typel99 Typel99:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/S0375-9474(99)00310-3">
    S. Typel and H. H. Wolter</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/S0375-9474(99)00310-3}
    {S. Typel and H. H. Wolter},
    \endlatexonly
    Nucl. Phys. A \b 656 (1999) 331.
    
    \anchor VanGiai81 VanGiai81:
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0370-2693(81)90646-8">
    Nguyen van Giai and H. Sagawa</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0370-2693(81)90646-8}{
    Nguyen van Giai and H. Sagawa}, 
    \endlatexonly
    Phys. Lett. B \b 106 (1981) 379.

    \anchor Zimanyi90 Zimanyi90:
    \htmlonly
    <a href="http://dx.doi.org/10.1103/PhysRevC.42.1416">
    Zimanyi and Moszkowski</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1103/PhysRevC.42.1416}{
    Zimanyi and Moszkowski}, 
    \endlatexonly
    Phys. Rev. C \b 42 (1990) 1416.

    \comment
    A blank reference

    \anchor :
    \htmlonly
    <a href="">
    </a>,
    \endhtmlonly
    \latexonly
    \href{}{
    },
    \endlatexonly

    \endcomment

*/
