/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

/* Example: ex_contour.cpp
   -------------------------------------------------------------------
   Example for generating contour lines
*/

#include <iostream>
#include <o2scl/contour.h>
#include <o2scl/ovector_tlate.h>

using namespace std;
using namespace o2scl;

// A function defining the three-dimensional surface
// for which we want to compute contour levels
double fun(double x, double y) {
  return 15.0*exp(-pow(x-20.0,2.0)/400.0-pow(y-5.0,2.0)/25.0)
    +40.0*exp(-pow(x-70.0,2.0)/500.0-pow(y-2.0,2.0)/4.0);
}

// A function for outputting the data to cout
int print_data(int nx, int ny, ovector_base &x, ovector_base &y,
	       omatrix_base &data);

// A function for printing the contour information to a file
int file_out(vector<contour_line> &conts, vector<contour_line> &conts2);

int main(void) {
  test_mgr t;
  t.set_output_level(1);

  cout.setf(ios::scientific);

  contour co;

  // Initialize the data

  ovector x(12), y(10);
  omatrix data(10,12);
  for(size_t i=0;i<10;i++) {
    y[i]=((double)i);
  }
  for(size_t i=0;i<12;i++) {
    x[i]=((double)i)*((double)i);
  }
  
  for(size_t j=0;j<12;j++) {
    for(size_t k=0;k<10;k++) {
      data[k][j]=fun(x[j],y[k]);
    }
  }
  co.set_data(12,10,x,y,data);
  
  // Print out the data

  print_data(12,10,x,y,data);

  // Set the contour levels

  ovector levels(7);
  levels[0]=5.0;
  levels[1]=10.0;
  levels[2]=15.0;
  levels[3]=20.0;
  levels[4]=25.0;
  levels[5]=30.0;
  levels[6]=35.0;
  
  co.set_levels(7,levels);

  // Compute the contours

  vector<contour_line> conts;
  co.calc_contours(conts);

  // Print the contours to the screen and test to make sure
  // that they match the requested level

  size_t nc=conts.size();
  for(size_t i=0;i<nc;i++) {
    cout << "Contour " << i << " at level " << conts[i].level << ":" << endl;
    size_t cs=conts[i].x.size();
    for(size_t j=0;j<cs;j++) {
      cout << "(" << conts[i].x[j] << ", " << conts[i].y[j] << ")" << endl;
      //t.test_rel(fun(conts[i].x[j],conts[i].y[j]),conts[i].level,
      //1.5e-1,"curve");
    }
    cout << endl;
  }

  // Refine the data using cubic spline interpolation

  def_interp_mgr<ovector_const_view,cspline_interp> dim1;
  def_interp_mgr<ovector_const_subvector,cspline_interp> dim2;
  co.regrid_data(5,5,dim1,dim2);

  // Recompute the contours

  vector<contour_line> conts2;
  co.calc_contours(conts2);

  // Output the contour information to a file for the documentation
  file_out(conts,conts2);
  
  t.report();

  return 0;
}
// End of example

int file_out(vector<contour_line> &conts, vector<contour_line> &conts2) {

  // This file is used by to generate the plots for the documentation
  ofstream fout("ex_contour.out");

  size_t nc=conts.size();
  fout << nc << endl;
  for(size_t i=0;i<nc;i++) {
    size_t cs=conts[i].x.size();
    fout << cs << " ";
    for(size_t j=0;j<cs;j++) {
      fout << conts[i].x[j] << " " << conts[i].y[j] << " ";
    }
    fout << endl;
  }
  
  nc=conts2.size();
  fout << nc << endl;
  for(size_t i=0;i<nc;i++) {
    size_t cs=conts2[i].x.size();
    fout << cs << " ";
    for(size_t j=0;j<cs;j++) {
      fout << conts2[i].x[j] << " " << conts2[i].y[j] << " ";
    }
    fout << endl;
  }

  fout.close();

  return 0;
}
  
int print_data(int nx, int ny, ovector_base &x, ovector_base &y,
	       omatrix_base &data) {
  int j, k;
  cout.setf(ios::showpos);
  cout.precision(3);
  cout << "              ";
  for(k=0;k<ny;k++) {
    string stx="y";
    stx+=itos(k);
    cout.width(11);
    cout << stx;
  }
  cout << endl;
  cout << "               ";
  for(k=0;k<ny;k++) {
    cout << y[k] << " ";
  }
  cout << endl;
  for(j=0;j<nx;j++) {
    string stx="x";
    stx+=itos(j);
    cout.width(3);
    cout << stx << " ";
    cout << x[j] << " ";
    for(k=0;k<ny;k++) {
      cout << data[k][j] << " ";
    }
    cout << endl;
  }
  cout << endl;
  cout.unsetf(ios::showpos);
  cout.precision(6);
  return 0;
}

