/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

/* Example: ex_ode.cpp
   -------------------------------------------------------------------
   An example to demonstrate solving differential equations
*/

#include <gsl/gsl_sf_bessel.h>
#include <gsl/gsl_sf_airy.h>
#include <gsl/gsl_sf_gamma.h>
#include <o2scl/test_mgr.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/ode_funct.h>
#include <o2scl/gsl_rkck.h>
#include <o2scl/gsl_rk8pd.h>
#include <o2scl/gsl_astep.h>
#include <o2scl/ode_iv_solve.h>

using namespace std;
using namespace o2scl;

// Differential equation defining the Bessel function. This assumes
// the second derivative at x=0 is 0 and thus only works for odd alpha.
int derivs(double x, size_t nv, const ovector_base &y, 
	   ovector_base &dydx, double &alpha) {
  dydx[0]=y[1];
  if (x==0.0) dydx[1]=0.0;
  else dydx[1]=(-x*y[1]+(-x*x+alpha*alpha)*y[0])/x/x;
  return 0;
}

// Differential equation defining the Airy Ai(x) function.
int derivs2(double x, size_t nv, const ovector_base &y, 
	    ovector_base &dydx, double &alpha) {
  dydx[0]=y[1];
  dydx[1]=y[0]*x;
  return 0;
}

int main(void) {
  int i;
  double x, dx=1.0e-1;
  ovector y(2), dydx(2), yout(2), yerr(2), dydx_out(2);
  void *vp=NULL;
  test_mgr t;
  t.set_output_level(1);

  cout.setf(ios::scientific);
  cout.setf(ios::showpos);

  ode_funct_fptr<double,ovector_base> od(derivs);
  ode_funct_fptr<double,ovector_base> od2(derivs2);
  gsl_rkck<double> ode;
  gsl_rk8pd<double> ode2;
  double alpha=1.0;
  
  // Solve using the non-adaptive Cash-Karp stepper.

  cout << "Bessel function, Cash-Karp: " << endl;
  x=0.0;
  y[0]=0.0;
  y[1]=0.5;
  derivs(x,2,y,dydx,alpha);
  cout << " x             J1(calc)      J1(exact)     rel. diff.    "
       << "err" << endl;
  while (x<1.0) {
    ode.step(x,dx,2,y,dydx,y,yerr,dydx,alpha,od);
    x+=dx;
    cout << x << " " << y[0] << " " 
	 << gsl_sf_bessel_J1(x) << " ";
    cout << fabs((y[0]-gsl_sf_bessel_J1(x))/gsl_sf_bessel_J1(x)) << " ";
    cout << yerr[0] << endl;
    t.test_rel(y[0],gsl_sf_bessel_J1(x),5.0e-5,"rkck");
  }
  cout << "Accuracy at end: " 
       << fabs(y[0]-gsl_sf_bessel_J1(x))/gsl_sf_bessel_J1(x) << endl;
  cout << endl;
  
  // Solve using the non-adaptive Prince-Dormand stepper. Note that
  // for the Bessel function, the 8th order stepper performs worse
  // than the 4th order. The error returned by the stepper is 
  // larger near x=0, as expected. 
  
  cout << "Bessel function, Prince-Dormand: " << endl;
  x=0.0;
  y[0]=0.0;
  y[1]=0.5;
  derivs(x,2,y,dydx,alpha);
  cout << " x             J1(calc)      J1(exact)     rel. diff.    "
       << "err" << endl;
  while (x<1.0) {
    ode2.step(x,dx,2,y,dydx,y,yerr,dydx,alpha,od);
    x+=dx;
    cout << x << " " << y[0] << " " 
	 << gsl_sf_bessel_J1(x) << " ";
    cout << fabs((y[0]-gsl_sf_bessel_J1(x))/gsl_sf_bessel_J1(x)) << " ";
    cout << yerr[0] << endl;
    t.test_rel(y[0],gsl_sf_bessel_J1(x),5.0e-4,"rk8pd");
  }
  cout << "Accuracy at end: " 
       << fabs(y[0]-gsl_sf_bessel_J1(x))/gsl_sf_bessel_J1(x) << endl;
  cout << endl;

  // Solve using the non-adaptive Cash-Karp stepper.

  cout << "Airy function, Cash-Karp: " << endl;
  x=0.0;
  y[0]=1.0/pow(3.0,2.0/3.0)/gsl_sf_gamma(2.0/3.0);
  y[1]=-1.0/pow(3.0,1.0/3.0)/gsl_sf_gamma(1.0/3.0);
  derivs2(x,2,y,dydx,alpha);
  cout << " x             Ai(calc)      Ai(exact)     rel. diff.    "
       << "err" << endl;
  while (x<1.0) {
    ode.step(x,dx,2,y,dydx,y,yerr,dydx,alpha,od2);
    x+=dx;
    cout << x << " " << y[0] << " " 
	 << gsl_sf_airy_Ai(x,GSL_PREC_DOUBLE) << " ";
    cout << fabs((y[0]-gsl_sf_airy_Ai(x,GSL_PREC_DOUBLE))/
		 gsl_sf_airy_Ai(x,GSL_PREC_DOUBLE)) << " ";
    cout << yerr[0] << endl;
    t.test_rel(y[0],gsl_sf_airy_Ai(x,GSL_PREC_DOUBLE),1.0e-8,"rkck");
  }
  cout << "Accuracy at end: " 
       << fabs(y[0]-gsl_sf_airy_Ai(x,GSL_PREC_DOUBLE))/
    gsl_sf_airy_Ai(x,GSL_PREC_DOUBLE) << endl;
  cout << endl;
  
  // Solve using the non-adaptive Prince-Dormand stepper. On this
  // function, the higher-order routine performs significantly
  // better. 
  
  cout << "Airy function, Prince-Dormand: " << endl;
  x=0.0;
  y[0]=1.0/pow(3.0,2.0/3.0)/gsl_sf_gamma(2.0/3.0);
  y[1]=-1.0/pow(3.0,1.0/3.0)/gsl_sf_gamma(1.0/3.0);
  derivs2(x,2,y,dydx,alpha);
  cout << " x             Ai(calc)      Ai(exact)     rel. diff.    "
       << "err" << endl;
  while (x<1.0) {
    ode2.step(x,dx,2,y,dydx,y,yerr,dydx,alpha,od2);
    x+=dx;
    cout << x << " " << y[0] << " " 
	 << gsl_sf_airy_Ai(x,GSL_PREC_DOUBLE) << " ";
    cout << fabs((y[0]-gsl_sf_airy_Ai(x,GSL_PREC_DOUBLE))/
		 gsl_sf_airy_Ai(x,GSL_PREC_DOUBLE)) << " ";
    cout << yerr[0] << endl;
    t.test_rel(y[0],gsl_sf_airy_Ai(x,GSL_PREC_DOUBLE),1.0e-14,"rk8pd");
  }
  cout << "Accuracy at end: " 
       << fabs(y[0]-gsl_sf_airy_Ai(x,GSL_PREC_DOUBLE))/
    gsl_sf_airy_Ai(x,GSL_PREC_DOUBLE) << endl;
  cout << endl;

  // Solve using the GSL adaptive stepper

  cout << "Adaptive stepper: " << endl;
  gsl_astep<double> ode3;
  x=0.0;
  y[0]=0.0;
  y[1]=0.5;
  cout << "   x             J1(calc)      J1(exact)     rel. diff.";
  cout << "    err_0         err_1" << endl;
  int k=0;
  while (x<10.0) {
    int retx=ode3.astep(x,dx,10.0,2,y,dydx,yerr,alpha,od);
    if (k%3==0) {
      cout << retx << " " << x << " " << y[0] << " " 
	   << gsl_sf_bessel_J1(x) << " ";
      cout << fabs((y[0]-gsl_sf_bessel_J1(x))/gsl_sf_bessel_J1(x)) << " ";
      cout << yerr[0] << " " << yerr[1] << endl;
    }
    t.test_rel(y[0],gsl_sf_bessel_J1(x),5.0e-3,"astep");
    t.test_rel(y[1],0.5*(gsl_sf_bessel_J0(x)-gsl_sf_bessel_Jn(2,x)),
               5.0e-3,"astep 2");
    t.test_rel(yerr[0],0.0,4.0e-6,"astep 3");
    t.test_rel(yerr[1],0.0,4.0e-6,"astep 4");
    t.test_gen(retx==0,"astep 5");
    k++;
  }
  cout << "Accuracy at end: "
       << fabs(y[0]-gsl_sf_bessel_J1(x))/gsl_sf_bessel_J1(x) << endl;
  cout << endl;

  // Decrease the tolerances, and the adaptive stepper takes
  // smaller step sizes.

  cout << "Adaptive stepper with smaller tolerances: " << endl;
  ode3.con.eps_abs=1.0e-8;
  ode3.con.a_dydt=1.0;
  x=0.0;
  y[0]=0.0;
  y[1]=0.5;
  cout << "   x             J1(calc)      J1(exact)     rel. diff.";
  cout << "    err_0         err_1" << endl;
  k=0;
  while (x<10.0) {
    int retx=ode3.astep(x,dx,10.0,2,y,dydx,yerr,alpha,od);
    if (k%3==0) {
      cout << retx << " " << x << " " << y[0] << " " 
	   << gsl_sf_bessel_J1(x) << " ";
      cout << fabs((y[0]-gsl_sf_bessel_J1(x))/gsl_sf_bessel_J1(x)) << " ";
      cout << yerr[0] << " " << yerr[1] << endl;
    }
    t.test_rel(y[0],gsl_sf_bessel_J1(x),5.0e-3,"astep");
    t.test_rel(y[1],0.5*(gsl_sf_bessel_J0(x)-gsl_sf_bessel_Jn(2,x)),
               5.0e-3,"astep 2");
    t.test_rel(yerr[0],0.0,4.0e-8,"astep 3");
    t.test_rel(yerr[1],0.0,4.0e-8,"astep 4");
    t.test_gen(retx==0,"astep 5");
    k++;
  }
  cout << "Accuracy at end: "
       << fabs(y[0]-gsl_sf_bessel_J1(x))/gsl_sf_bessel_J1(x) << endl;
  cout << endl;

  // Use the higher-order stepper, and less steps are required. The
  // stepper automatically takes more steps near x=0 in order since
  // the higher-order routine has more trouble there.

  cout << "Adaptive stepper, Prince-Dormand: " << endl;
  ode3.set_step(ode2);
  x=0.0;
  y[0]=0.0;
  y[1]=0.5;
  cout << "   x             J1(calc)      J1(exact)     rel. diff.";
  cout << "    err_0         err_1" << endl;
  k=0;
  while (x<10.0) {
    int retx=ode3.astep(x,dx,10.0,2,y,dydx,yerr,alpha,od);
    if (k%3==0) {
      cout << retx << " " << x << " " << y[0] << " " 
	   << gsl_sf_bessel_J1(x) << " ";
      cout << fabs((y[0]-gsl_sf_bessel_J1(x))/gsl_sf_bessel_J1(x)) << " ";
      cout << yerr[0] << " " << yerr[1] << endl;
    }
    t.test_rel(y[0],gsl_sf_bessel_J1(x),5.0e-3,"astep");
    t.test_rel(y[1],0.5*(gsl_sf_bessel_J0(x)-gsl_sf_bessel_Jn(2,x)),
               5.0e-3,"astep");
    t.test_rel(yerr[0],0.0,4.0e-8,"astep 3");
    t.test_rel(yerr[1],0.0,4.0e-8,"astep 4");
    t.test_gen(retx==0,"astep 5");
    k++;
  }
  cout << "Accuracy at end: "
       << fabs(y[0]-gsl_sf_bessel_J1(x))/gsl_sf_bessel_J1(x) << endl;
  cout << endl;

  // Solve using the O2scl initial value solver
  
  cout << "Initial value solver: " << endl;
  ode_iv_solve<double> ode4;
  const int ngrid=101;
  ovector xg(ngrid), yinit(2);
  omatrix yg(ngrid,2);
  for(i=0;i<ngrid;i++) xg[i]=((double)i)/10.0;
  yinit[0]=0.0;
  yinit[1]=0.5;
  ode4.solve_grid(0.0,10.0,0.1,2,yinit,ngrid,xg,yg,alpha,od);
  
  cout << " x             J1(calc)      J1(exact)     rel. diff." << endl;
  for(i=1;i<ngrid;i+=10) {
    cout << xg[i] << " " << yg[i][0] << " "
         << gsl_sf_bessel_J1(xg[i]) << " ";
    cout << fabs((yg[i][0]-gsl_sf_bessel_J1(xg[i]))/
		 gsl_sf_bessel_J1(xg[i])) << endl;
    t.test_rel(yg[i][0],gsl_sf_bessel_J1(xg[i]),5.0e-7,"astep");
    t.test_rel(yg[i][1],0.5*(gsl_sf_bessel_J0(xg[i])-
                             gsl_sf_bessel_Jn(2,xg[i])),5.0e-7,"astep 2");
  }
  cout << "Accuracy at end: "
       << fabs(yg[ngrid-1][0]-gsl_sf_bessel_J1(xg[ngrid-1]))/
    gsl_sf_bessel_J1(xg[ngrid-1]) << endl;
  cout << endl;

  cout.unsetf(ios::showpos);
  t.report();

  return 0;
}
// End of example
