/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ERR_HND_H
#define O2SCL_ERR_HND_H

#include <iostream>
#include <string>
#include <gsl/gsl_errno.h>

/** \file err_hnd.h
    \brief File for definitions for err_base and err_class
*/

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief The integer error definitions 
      
      The errors associated with the integers between -2 and 32
      are from GSL, the rest are specific to \o2 . 
  */
  enum {
    /// Success (GSL)
    gsl_success=0,
    /// Failure (GSL)
    gsl_failure=-1,
    /// iteration has not converged (GSL)
    gsl_continue=-2,  
    /// input domain error, e.g sqrt(-1) (GSL)
    gsl_edom=1,   
    /// output range error, e.g. exp(1e100) (GSL)
    gsl_erange=2,   
    /// invalid pointer (GSL)
    gsl_efault=3,   
    /// invalid argument supplied by user (GSL)
    gsl_einval=4,   
    /// generic failure (GSL)
    gsl_efailed=5,   
    /// factorization failed (GSL)
    gsl_efactor=6,   
    /// sanity check failed - shouldn't happen (GSL)
    gsl_esanity=7,   
    /// malloc failed (GSL)
    gsl_enomem=8,   
    /// problem with user-supplied function (GSL)
    gsl_ebadfunc=9,   
    /// iterative process is out of control (GSL)
    gsl_erunaway=10,  
    /// exceeded max number of iterations (GSL)
    gsl_emaxiter=11,  
    /// tried to divide by zero (GSL)
    gsl_ezerodiv=12,  
    /// user specified an invalid tolerance (GSL)
    gsl_ebadtol=13,  
    /// failed to reach the specified tolerance (GSL)
    gsl_etol=14,  
    /// underflow (GSL)
    gsl_eundrflw=15,  
    /// overflow (GSL)
    gsl_eovrflw=16,  
    /// loss of accuracy (GSL)
    gsl_eloss=17,  
    /// failed because of roundoff error (GSL)
    gsl_eround=18,  
    /// matrix, vector lengths are not conformant (GSL)
    gsl_ebadlen=19,  
    /// matrix not square (GSL)
    gsl_enotsqr=20,  
    /// apparent singularity detected (GSL)
    gsl_esing=21,  
    /// integral or series is divergent (GSL)
    gsl_ediverge=22,  
    /// requested feature is not supported by the hardware (GSL)
    gsl_eunsup=23,  
    /// requested feature not (yet) implemented (GSL)
    gsl_eunimpl=24,  
    /// cache limit exceeded (GSL)
    gsl_ecache=25,  
    /// table limit exceeded (GSL)
    gsl_etable=26,  
    /// iteration is not making progress toward solution (GSL)
    gsl_enoprog=27,  
    /// jacobian evaluations are not improving the solution (GSL)
    gsl_enoprogj=28,  
    /// cannot reach the specified tolerance in f (GSL)
    gsl_etolf=29,  
    /// cannot reach the specified tolerance in x (GSL)
    gsl_etolx=30,  
    /// cannot reach the specified tolerance in gradient (GSL)
    gsl_etolg=31,  
    /// end of file (GSL)
    gsl_eof=32,  
    /// Generic "not found" result
    gsl_enotfound=33,
    /// Incorrect type for memory object
    gsl_ememtype=34,
    /// File not found
    gsl_efilenotfound=35,
    /// Invalid index for array or matrix
    gsl_eindex=36,
    /// Outside constraint region
    gsl_outsidecons=37
  };

  // Forward declaration
  class err_base;

  /** \brief The global error handler pointer
   */      
  extern err_base *err_hnd;

  /** 
      \brief Class defining an error handler [abstract base]

      A global object of this type is defined, \ref err_hnd .

      \future There may be an issue associated with the string
      manipulations causing errors in the error handler.
   */
  class err_base {
    
  public:

    err_base() {
      gsl_set_error_handler(err_hnd->gsl_hnd);
    }

    virtual ~err_base() {}

    /** 
	\brief Set an error 
    
	This is separate from set(), since the gsl error handler
	needs to be a static function.
    */
    static void gsl_hnd(const char *reason, const char *file, 
			int line, int lerrno) {
      err_hnd->set(reason,file,line,lerrno);
    }

    /// Set an error 
    virtual void set(const char *reason, const char *file, 
		     int line, int lerrno)=0;
    
    /// Get the last error
    virtual void get(const char *&reason, const char *&file,
		     int &line, int &lerrno)=0;

    /// Return the last error number
    virtual int get_errno()=0;

    /// Return the line number of the last error
    virtual int get_line()=0;

    /// Return the reason for the last error
    virtual const char *get_reason()=0;

    /// Return the file name of the last error
    virtual const char *get_file()=0;

    /// Return a string summarizing the last error
    virtual const char *get_str()=0;

    /// Remove last error information
    virtual void reset()=0;

  };
  
  /** 
      \brief The error handler 
      
      An error handler for use in \o2 which replaces the GSL error handler
      
      Note that the string arguments to set() can refer to temporary
      storage, since they are copied when the function is called and
      an error is set.
  */
  class err_class : public err_base {

  public:  

    err_class();

    virtual ~err_class() {}

    /// Set an error 
    virtual void set(const char *reason, const char *file, 
		     int line, int lerrno);

    /// Get the last error
    virtual void get(const char *&reason, const char *&file,
		     int &line, int &lerrno);

    /// Return the last error number
    virtual int get_errno();

    /// Return the line number of the last error
    virtual int get_line();

    /// Return the reason for the last error
    virtual const char *get_reason();

    /// Return the file name of the last error
    virtual const char *get_file();

    /// Return a string summarizing the last error
    virtual const char *get_str();

    /// Remove last error information
    virtual void reset();

    /** \brief Set error handling mode

	- 0 - Continue execution after an error occurs
	- 1 - Continue execution after an error occurs
	- 2 - Abort execution after an error occurs, and print
	out the error information (default)
     */
    void set_mode(int m) {
      if (m<0) mode=0;
      else if (m>2) mode=2;
      else mode=m;
    }

    /// Return the error handling mode
    int get_mode() {
      return mode;
    }

    /** \brief If true, call exit() when an array index error is set 
	(default true)
	
	This is ignored if \c O2SCL_ARRAY_ABORT is not defined.
    */
    bool array_abort;

    /// Number of characters from filename to print (default 35)
    size_t fname_size;

  protected:

#ifndef DOXYGEN_INTERNAL

    /// The maximum size of error explanations
    static const int rsize=300;
    /// The maximum size of error explanations with the line and file info
    static const int fsize=400;

    /// The error number
    int a_errno;
    /// The line number
    int a_line;
    /// The mode of error handling (default 2)
    int mode;

    /// The filename
    char *a_file;
    /// The error explanation
    char a_reason[rsize];
    /// A full string with explanation and line and file info
    char fullstr[fsize];

#endif

  };

  /** \brief The default error handler
   */      
  extern err_class def_err_hnd;

  /** \brief Set an error with message \c d and code \c n
   */
#define O2SCL_ERR(d,n) o2scl::set_err_fn(d,__FILE__,__LINE__,n);
  
  /** \brief Set a "convergence" error
   */
#define O2SCL_CONV(d,n,b) {if (b) o2scl::set_err_fn(d,__FILE__,__LINE__,n);}
  
  /** \brief Set an error, two-string version
   */
#define O2SCL_ERR2(d,d2,n) o2scl::set_err_fn((std::string(d)+d2).c_str(), \
					     __FILE__,__LINE__,n);
  
  /** \brief Set a "convergence" error, two-string version
   */
#define O2SCL_CONV2(d,d2,n,b) {if (b)					\
      o2scl::set_err_fn((std::string(d)+d2).c_str(),			\
			__FILE__,__LINE__,n);}
  
  /** \brief Set an error and return the error value
   */
#define O2SCL_ERR_RET(d,n)						\
  do { o2scl::set_err_fn(d,__FILE__,__LINE__,n); return n; } while (0)
  
  /** \brief Set a "convergence" error and return the error value
   */
#define O2SCL_CONV_RET(d,n,b)						\
  do { if (!b) { return 0; } else {					\
      o2scl::set_err_fn(d,__FILE__,__LINE__,n); return n; } } while (0)
  
  /** \brief Set an error and return the error value, two-string version
   */
#define O2SCL_ERR2_RET(d,d2,n)						\
  do { o2scl::set_err_fn((std::string(d)+d2).c_str(),			\
			 __FILE__,__LINE__,n); return n; } while (0)
  
  /** \brief Set an error and return the error value, two-string version
   */
#define O2SCL_CONV2_RET(d,d2,n,b)					\
  do { if (!b) {return 0; } else {					\
      o2scl::set_err_fn((std::string(d)+d2).c_str(),			\
			__FILE__,__LINE__,n); return n; } } while (0)
  
  /** \brief Set an error
   */
  inline void set_err_fn(const char *desc, const char *file, int line,
			 int errnum) {
    err_hnd->set(desc,file,line,errnum);
    return;
  }
  
  /** \brief Update an error value \c err with the value in \c ret

      If \c ret is zero, this sets \c ret to the value \c err, and 
      if \c ret is nonzero this function does nothing.
   */
  inline void error_update(int &ret, int err) { if (ret==0) ret=err; }
  
  /** \brief A version of \c assert, i.e. exit if the error value is
      non-zero and do nothing otherwise
      
      \future Make this consistent with assert() using NDEBUG?
  */
#define O2SCL_ASSERT(ev)						\
  do { if (ev!=0) { std::cout << "O2scl: Macro err_assert() causing exit" \
			      << " from error " << ev << " at "		\
			      << __LINE__ << " in file:\n "		\
			      << __FILE__ << std::endl;			\
      std::cout << "Error handler string:\n " << err_hnd->get_str()	\
		<< std::endl; exit(ev); } } while (0)
  
  /** \brief A version of \c assert for bool types. Exit if the argument
      is false
  */
#define O2SCL_BOOL_ASSERT(ev,str)					\
  do { if (ev==false) {							\
      std::cout << "O2scl: Macro bool_assert() causing exit at line "	\
		<< __LINE__ << " in file:\n "				\
		<< __FILE__ << std::endl;				\
      std::cout << "Given string: " << str				\
		<< std::endl; exit(-1); } } while (0)
  
#ifndef DOXYGENP
}
#endif

#endif
