/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_FUNCT_H
#define O2SCL_FUNCT_H

#include <string>
#include <o2scl/collection.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief One-dimensional function [abstract base]
      
      This class generalizes a function y(x).

      This class is one of a large number of function object classes
      in \o2 designed to provide a mechanism for the user to 
      supply functions to solvers, minimizers, integrators, etc.
      See \ref funct_section for a general description.

  */
  template<class param_t> class funct {
    
  public:  
    
    /*
      This empty constructor is required for the child funct classes
    */
    funct() {}

    virtual ~funct() {}
    
    /** \brief Compute the function at point \c x, with result \c y
     */
    virtual int operator()(double x, double &y, param_t &pa)=0;
    
    /** \brief Compute the function at point \c x, returning the result
     */
    virtual double operator()(double x, param_t &pa) {
      double y;
      operator()(x,y,pa);
      return y;
    }

#ifndef DOXYGENP

  private:

    funct(const funct &);

    funct& operator=(const funct&);
    
#endif

  };
  
  /** \brief Function pointer to a function
   */
  template<class param_t> class funct_fptr : public funct<param_t> {

  public:
    
    /** \brief Specify the function pointer
     */
    funct_fptr(int (*fp)(double x, double &y, param_t &pa)) {
      fptr=fp;
    }
    
    virtual ~funct_fptr() {}

    /** \brief Compute the function at point \c x, with result \c y
     */
    virtual int operator()(double x, double &y, param_t &pa) {
      return fptr(x,y,pa);
    }
    
    /** \brief Compute the function at point \c x, returning the result

	\comment
	This is here because if we overload operator()(double,double,void*),
	we have to reimplement all operator()'s
	\endcomment
    */
    virtual double operator()(double x, param_t &pa) {
      double y;
      operator()(x,y,pa);
      return y;
    }
    
#ifndef DOXYGEN_INTERNAL

  protected:

    friend class io_tlate<funct_fptr>;
  
    /// Storage for the function pointer
    int (*fptr)(double x, double &y, param_t &pa);

#endif
#ifndef DOXYGENP

  private:

    funct_fptr(const funct_fptr &);
    funct_fptr& operator=(const funct_fptr&);

#endif

  };

  /** \brief Function pointer to a function returning the function value
   */
  template<class param_t>
    class funct_fptr_noerr : public funct<param_t> {
  public:

    /** \brief Specify the function pointer
     */
    funct_fptr_noerr(double (*fp)(double x, param_t &pa)) {
      fptr=fp;
    }

    virtual ~funct_fptr_noerr() {}

    /** \brief Compute the function at point \c x, with result \c y
     */
    virtual int operator()(double x, double &y, param_t &pa) {
      y=fptr(x,pa);
      return 0;
    }

    /** \brief Compute the function at point \c x, returning the result

	\comment
	This is here because if we overload operator()(double,double,void*),
	we have to reimplement all operator()'s
	\endcomment
     */
    virtual double operator()(double x, param_t &pa) {
      double y;
      operator()(x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

  protected:
  
    /// Storage for the function pointer
    double (*fptr)(double x, param_t &pa);

#endif
#ifndef DOXYGENP

  private:

    funct_fptr_noerr(const funct_fptr_noerr &);
    funct_fptr_noerr& operator=(const funct_fptr_noerr&);

#endif

  };

  /** \brief Function pointer to a function with no parameters
   */
  template<class param_t>
    class funct_fptr_nopar : public funct<param_t> {

  public:

    /** \brief Specify the function pointer
     */
    funct_fptr_nopar(double (*fp)(double x)) {
      fptr=fp;
    }

    virtual ~funct_fptr_nopar() {}

    /** \brief Compute the function at point \c x, with result \c y
     */
    virtual int operator()(double x, double &y, param_t &pa) {
      y=fptr(x);
      return 0;
    }

    /** \brief Compute the function at point \c x, returning the result

	\comment
	This is here because if we overload operator()(double,double,void*),
	we have to reimplement all operator()'s
	\endcomment
     */
    virtual double operator()(double x, param_t &pa) {
      double y;
      operator()(x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

  protected:
  
    /// Storage for the function pointer
    double (*fptr)(double x);

#endif
#ifndef DOXYGENP

  private:

    funct_fptr_nopar(const funct_fptr_nopar &);
    funct_fptr_nopar& operator=(const funct_fptr_nopar&);

#endif

  };

  /** 
      \brief Member function pointer to a one-dimensional function
  */
  template <class tclass, class param_t> class funct_mfptr : 
  public funct<param_t> {
  public:
  
    /** \brief Specify the member function pointer
     */
    funct_mfptr(tclass *tp, int (tclass::*fp)(double x, double &y, 
					      param_t &pa)) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~funct_mfptr() {};
  
    /** \brief Compute the function at point \c x, with result \c y
     */
    virtual int operator()(double x, double &y, param_t &pa) {
      return (*tptr.*fptr)(x,y,pa);
    }

    /** \brief Compute the function at point \c x, returning the result

	\comment
	This is here because if we overload operator()(double,double,void*),
	we have to reimplement all operator()'s
	\endcomment
     */
    virtual double operator()(double x, param_t &pa) {
      double y;
      operator()(x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

  protected:
  
    /// Storage for the member function pointer
    int (tclass::*fptr)(double x, double &y, param_t &pa);

    /// Store the pointer to the class instance
    tclass *tptr;

#endif
#ifndef DOXYGENP

  private:

    funct_mfptr(const funct_mfptr &);
    funct_mfptr& operator=(const funct_mfptr&);

#endif

  };

  /** 
      \brief Const member function pointer to a one-dimensional function

      \note While this is designed to accept a pointer to a const
      member function, the choice of whether the class pointer
      given in the template type <tt>tclass</tt> is const or 
      not is up to the user.
  */
  template <class tclass, class param_t> class funct_cmfptr : 
  public funct<param_t> {
  public:
  
    /** \brief Specify the member function pointer
     */
    funct_cmfptr(tclass *tp, int (tclass::*fp)(double x, double &y, 
					       param_t &pa) const) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~funct_cmfptr() {};
  
    /** \brief Compute the function at point \c x, with result \c y
     */
    virtual int operator()(double x, double &y, param_t &pa) {
      return (*tptr.*fptr)(x,y,pa);
    }

    /** \brief Compute the function at point \c x, returning the result

	\comment
	This is here because if we overload operator()(double,double,void*),
	we have to reimplement all operator()'s
	\endcomment
     */
    virtual double operator()(double x, param_t &pa) {
      double y;
      operator()(x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

  protected:
  
    /// Storage for the const member function pointer
    int (tclass::*fptr)(double x, double &y, param_t &pa) const;

    /// Store the pointer to the class instance
    tclass *tptr;

#endif
#ifndef DOXYGENP

  private:

    funct_cmfptr(const funct_cmfptr &);
    funct_cmfptr& operator=(const funct_cmfptr&);

#endif

  };

 /** \brief Member function pointer to a one-dimensional function
      returning the function value
  */
  template <class tclass, class param_t> class funct_mfptr_noerr : 
  public funct<param_t> {
  public:
  
    /** \brief Specify the member function pointer
     */
    funct_mfptr_noerr(tclass *tp, 
		      double (tclass::*fp)(double x, param_t &pa)) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~funct_mfptr_noerr() {};
  
    /** \brief Compute the function at point \c x, with result \c y
     */
    virtual int operator()(double x, double &y, param_t &pa) {
      y=(*tptr.*fptr)(x,pa);
      return 0;
    }

    /** \brief Compute the function at point \c x, returning the result

	\comment
	This is here because if we overload operator()(double,double,void*),
	we have to reimplement all operator()'s
	\endcomment
     */
    virtual double operator()(double x, param_t &pa) {
      double y;
      operator()(x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

  protected:
  
    /// Storage for the member function pointer
    double (tclass::*fptr)(double x, param_t &pa);

    /// Store the pointer to the class instance
    tclass *tptr;

#endif
#ifndef DOXYGENP

  private:

    funct_mfptr_noerr(const funct_mfptr_noerr &);
    funct_mfptr_noerr& operator=(const funct_mfptr_noerr&);

#endif

  };

  /** \brief Const member function pointer to a one-dimensional function
      returning the function value

      \note While this is designed to accept a pointer to a const
      member function, the choice of whether the class pointer
      given in the template type <tt>tclass</tt> is const or 
      not is up to the user.
  */
  template <class tclass, class param_t> class funct_cmfptr_noerr : 
  public funct<param_t> {
  public:
  
    /** \brief Specify the member function pointer
     */
    funct_cmfptr_noerr(tclass *tp, 
		      double (tclass::*fp)(double x, param_t &pa) const) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~funct_cmfptr_noerr() {};
  
    /** \brief Compute the function at point \c x, with result \c y
     */
    virtual int operator()(double x, double &y, param_t &pa) {
      y=(*tptr.*fptr)(x,pa);
      return 0;
    }

    /** \brief Compute the function at point \c x, returning the result

	\comment
	This is here because if we overload operator()(double,double,void*),
	we have to reimplement all operator()'s
	\endcomment
     */
    virtual double operator()(double x, param_t &pa) {
      double y;
      operator()(x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

  protected:
  
    /// Storage for the const member function pointer
    double (tclass::*fptr)(double x, param_t &pa) const;

    /// Store the pointer to the class instance
    tclass *tptr;

#endif
#ifndef DOXYGENP

  private:

    funct_cmfptr_noerr(const funct_cmfptr_noerr &);
    funct_cmfptr_noerr& operator=(const funct_cmfptr_noerr&);

#endif

  };

  /** \brief Member function pointer to a one-dimensional function
      with no parameters
  */
  template <class tclass, class param_t> class funct_mfptr_nopar : 
  public funct<param_t> {
  public:
    
    /** \brief Specify the member function pointer
     */
    funct_mfptr_nopar(tclass *tp, double (tclass::*fp)(double x)) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~funct_mfptr_nopar() {};
  
    /** \brief Compute the function at point \c x, with result \c y
     */
    virtual int operator()(double x, double &y, param_t &pa) {
      y=(*tptr.*fptr)(x);
      return 0;
    }
  
    /** \brief Compute the function at point \c x, returning the result

	\comment
	This is here because if we overload operator()(double,double,void*),
	we have to reimplement all operator()'s
	\endcomment
     */
    virtual double operator()(double x, param_t &pa) {
      double y;
      operator()(x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

  protected:
  
    /// Storage for the member function pointer
    double (tclass::*fptr)(double x);
    /// Store the pointer to the class instance
    tclass *tptr;

#endif
#ifndef DOXYGENP

  private:

    funct_mfptr_nopar(const funct_mfptr_nopar &);
    funct_mfptr_nopar& operator=(const funct_mfptr_nopar&);

#endif

  };
  
  /** \brief Const member function pointer to a one-dimensional function
      with no parameters

      \note While this is designed to accept a pointer to a const
      member function, the choice of whether the class pointer
      given in the template type <tt>tclass</tt> is const or 
      not is up to the user.
  */
  template <class tclass, class param_t> class funct_cmfptr_nopar : 
  public funct<param_t> {
  public:
    
    /** \brief Specify the member function pointer
     */
    funct_cmfptr_nopar(tclass *tp, double (tclass::*fp)(double x) const) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~funct_cmfptr_nopar() {};
  
    /** \brief Compute the function at point \c x, with result \c y
     */
    virtual int operator()(double x, double &y, param_t &pa) {
      y=(*tptr.*fptr)(x);
      return 0;
    }
  
    /** \brief Compute the function at point \c x, returning the result

	\comment
	This is here because if we overload operator()(double,double,void*),
	we have to reimplement all operator()'s
	\endcomment
     */
    virtual double operator()(double x, param_t &pa) {
      double y;
      operator()(x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

  protected:
  
    /// Storage for the const member function pointer
    double (tclass::*fptr)(double x) const;
    /// Store the pointer to the class instance
    tclass *tptr;

#endif
#ifndef DOXYGENP

  private:

    funct_cmfptr_nopar(const funct_cmfptr_nopar &);
    funct_cmfptr_nopar& operator=(const funct_cmfptr_nopar&);

#endif

  };
  
#ifndef DOXYGENP
}
#endif

#endif
