/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#include <o2scl/interp.h>
#include <o2scl/test_mgr.h>

using namespace std;
using namespace o2scl;

int main(void) {
  ovector x(5), y(5), rx(5), ry(5);
  double xa[5], ya[5], rxa[5], rya[5];
  test_mgr t;
  def_err_hnd.set_mode(2);
  t.set_output_level(1);
  
  cout.setf(ios::scientific);

  // ---------------------------------------------------------------
  // Create test data

  for(size_t i=0;i<5;i++) {
    x[i]=((double)i);
    y[i]=((double)i*i);
    rx[4-i]=((double)i);
    ry[4-i]=((double)i*i);
    xa[i]=((double)i);
    ya[i]=((double)i*i);
    rxa[4-i]=((double)i);
    rya[4-i]=((double)i*i);
  }
  
  // ---------------------------------------------------------------
  // Test reversing arrays

  ovector_reverse rx2(x);
  for(size_t i=0;i<5;i++) {
    t.test_rel(rx[i],rx2[i],1.0e-10,"rev1");
  }
  
  array_reverse<5> rxa2(xa);
  for(size_t i=0;i<5;i++) {
    t.test_rel(rxa[i],rxa2[i],1.0e-10,"rev2");
  }

  // ---------------------------------------------------------------
  // Use linear interpolation

  def_interp_mgr<ovector_const_view,linear_interp> itx;
  def_interp_mgr<ovector_const_reverse,linear_interp> ritx;
  o2scl_interp<> gi(itx,ritx);
  o2scl_interp_vec<> giv1(itx,5,x,y);
  o2scl_interp_vec<> giv2(itx,5,rx,ry);

  // ---------------------------------------------------------------
  // Test normal interpolation

  double x0=2.5;
  double y0;
  
  y0=gi.interp(x0,5,x,y);
  t.test_rel(y0,6.5,1.0e-5,"intp 1a.");
  y0=gi.deriv(x0,5,x,y);
  t.test_rel(y0,5.0,1.0e-5,"intp 1b.");
  y0=gi.deriv2(x0,5,x,y);
  t.test_rel(y0,0.0,1.0e-5,"intp 1c.");
  y0=gi.integ(x0,3.5,5,x,y);
  t.test_rel(y0,9.25,1.0e-5,"intp 1d.");

  // ---------------------------------------------------------------
  // Test normal interpolation with a reversed array
  
  y0=gi.interp(x0,5,rx,ry);
  t.test_rel(y0,6.5,1.0e-5,"rintp 1a.");
  y0=gi.deriv(x0,5,rx,ry);
  t.test_rel(y0,5.0,1.0e-5,"rintp 1b.");
  y0=gi.deriv2(x0,5,rx,ry);
  t.test_rel(y0,0.0,1.0e-5,"rintp 1c.");
  y0=gi.integ(x0,3.5,5,rx,ry);
  t.test_rel(y0,9.25,1.0e-5,"rintp 1d.");

  // ---------------------------------------------------------------
  // Test normal interpolation with alternative interface

  y0=giv1.interp(x0);
  t.test_rel(y0,6.5,1.0e-5,"intp 2a.");
  y0=giv1.deriv(x0);
  t.test_rel(y0,5.0,1.0e-5,"intp 2b.");
  y0=giv1.deriv2(x0);
  t.test_rel(y0,0.0,1.0e-5,"intp 2c.");
  y0=giv1.integ(x0,3.5);
  t.test_rel(y0,9.25,1.0e-5,"intp 2d.");

  // ---------------------------------------------------------------
  // Test normal interpolation with a reversed array and 
  // alternate interface
  
  y0=giv2.interp(x0);
  t.test_rel(y0,6.5,1.0e-5,"rintp 2a.");
  y0=giv2.deriv(x0);
  t.test_rel(y0,5.0,1.0e-5,"rintp 2b.");
  y0=giv2.deriv2(x0);
  t.test_rel(y0,0.0,1.0e-5,"rintp 2c.");
  y0=giv2.integ(x0,3.5);
  t.test_rel(y0,9.25,1.0e-5,"rintp 2d.");

  // ---------------------------------------------------------------
  // Use cspline interpolation

  def_interp_mgr<ovector_const_view,cspline_interp> itz;
  def_interp_mgr<ovector_const_reverse,cspline_interp> ritz;
  gi.set_type(itz,ritz);
  o2scl_interp_vec<> giv3(itz,5,x,y);
  o2scl_interp_vec<> giv4(itz,5,rx,ry);

  // ---------------------------------------------------------------
  // Test normal interpolation

  y0=gi.interp(x0,5,x,y);
  t.test_rel(y0,6.232143,1.0e-5,"intp 3a.");
  y0=gi.deriv(x0,5,x,y);
  t.test_rel(y0,4.964286,1.0e-5,"intp 3b.");
  y0=gi.deriv2(x0,5,x,y);
  t.test_rel(y0,2.142857,1.0e-5,"intp 3c.");
  y0=gi.integ(x0,3.5,5,x,y);
  t.test_rel(y0,9.098214,1.0e-5,"intp 3d.");

  // ---------------------------------------------------------------
  // Test normal interpolation with a reversed array
  
  y0=gi.interp(x0,5,rx,ry);
  t.test_rel(y0,6.232143,1.0e-5,"rintp 3a.");
  y0=gi.deriv(x0,5,rx,ry);
  t.test_rel(y0,4.964286,1.0e-5,"rintp 3b.");
  y0=gi.deriv2(x0,5,rx,ry);
  t.test_rel(y0,2.142857,1.0e-5,"rintp 3c.");
  y0=gi.integ(x0,3.5,5,rx,ry);
  t.test_rel(y0,9.098214,1.0e-5,"rintp 3d.");

  // ---------------------------------------------------------------
  // Test normal interpolation with alternative interface
  
  y0=giv3.interp(x0);
  t.test_rel(y0,6.232143,1.0e-5,"intp 4a.");
  y0=giv3.deriv(x0);
  t.test_rel(y0,4.964286,1.0e-5,"intp 4b.");
  y0=giv3.deriv2(x0);
  t.test_rel(y0,2.142857,1.0e-5,"intp 4c.");
  y0=giv3.integ(x0,3.5);
  t.test_rel(y0,9.098214,1.0e-5,"intp 4d.");

  // ---------------------------------------------------------------
  // Test normal interpolation with a reversed array and 
  // alternate interface
  
  y0=giv4.interp(x0);
  t.test_rel(y0,6.232143,1.0e-5,"rintp 4a.");
  y0=giv4.deriv(x0);
  t.test_rel(y0,4.964286,1.0e-5,"rintp 4b.");
  y0=giv4.deriv2(x0);
  t.test_rel(y0,2.142857,1.0e-5,"rintp 4c.");
  y0=giv4.integ(x0,3.5);
  t.test_rel(y0,9.098214,1.0e-5,"rintp 4d.");

  // ---------------------------------------------------------------
  // Testing with C-style arrays

  // ---------------------------------------------------------------
  // Use linear interpolation

  def_interp_mgr<double[5],linear_interp> itax;
  def_interp_mgr<array_const_reverse<5>,linear_interp> ritax;

  array_interp<5> gia(itax,ritax);
  array_interp_vec<double[5]> giv1a(itax,5,xa,ya);
  array_interp_vec<double[5]> giv2a(itax,5,rxa,rya);

  // ---------------------------------------------------------------
  // Test normal interpolation
  
  x0=2.5;
  
  y0=gia.interp(x0,5,xa,ya);
  t.test_rel(y0,6.5,1.0e-5,"intp 1a.");
  y0=gia.deriv(x0,5,xa,ya);
  t.test_rel(y0,5.0,1.0e-5,"intp 1b.");
  y0=gia.deriv2(x0,5,xa,ya);
  t.test_rel(y0,0.0,1.0e-5,"intp 1c.");
  y0=gia.integ(x0,3.5,5,xa,ya);
  t.test_rel(y0,9.25,1.0e-5,"intp 1d.");

  // ---------------------------------------------------------------
  // Test normal interpolation with a reversed array
  
  y0=gia.interp(x0,5,rxa,rya);
  t.test_rel(y0,6.5,1.0e-5,"rintp 1a.");
  y0=gia.deriv(x0,5,rxa,rya);
  t.test_rel(y0,5.0,1.0e-5,"rintp 1b.");
  y0=gia.deriv2(x0,5,rxa,rya);
  t.test_rel(y0,0.0,1.0e-5,"rintp 1c.");
  y0=gia.integ(x0,3.5,5,rxa,rya);
  t.test_rel(y0,9.25,1.0e-5,"rintp 1d.");

  // ---------------------------------------------------------------
  // Test normal interpolation with alternative interface

  y0=giv1a.interp(x0);
  t.test_rel(y0,6.5,1.0e-5,"intp 2a.");
  y0=giv1a.deriv(x0);
  t.test_rel(y0,5.0,1.0e-5,"intp 2b.");
  y0=giv1a.deriv2(x0);
  t.test_rel(y0,0.0,1.0e-5,"intp 2c.");
  y0=giv1a.integ(x0,3.5);
  t.test_rel(y0,9.25,1.0e-5,"intp 2d.");

  // ---------------------------------------------------------------
  // Test normal interpolation with a reversed array and 
  // alternate interface
  
  y0=giv2a.interp(x0);
  t.test_rel(y0,6.5,1.0e-5,"rintp 2a.");
  y0=giv2a.deriv(x0);
  t.test_rel(y0,5.0,1.0e-5,"rintp 2b.");
  y0=giv2a.deriv2(x0);
  t.test_rel(y0,0.0,1.0e-5,"rintp 2c.");
  y0=giv2a.integ(x0,3.5);
  t.test_rel(y0,9.25,1.0e-5,"rintp 2d.");

  // ---------------------------------------------------------------
  // Use cspline interpolation
  
  def_interp_mgr<double[5],cspline_interp> itaz;
  def_interp_mgr<array_const_reverse<5>,cspline_interp> ritaz;
  gia.set_type(itaz,ritaz);
  array_interp_vec<double[5]> giv3a(itaz,5,xa,ya);
  array_interp_vec<double[5]> giv4a(itaz,5,rxa,rya);
  
  // ---------------------------------------------------------------
  // Test normal interpolation

  y0=gia.interp(x0,5,xa,ya);
  t.test_rel(y0,6.232143,1.0e-5,"intp 3a.");
  y0=gia.deriv(x0,5,xa,ya);
  t.test_rel(y0,4.964286,1.0e-5,"intp 3b.");
  y0=gia.deriv2(x0,5,xa,ya);
  t.test_rel(y0,2.142857,1.0e-5,"intp 3c.");
  y0=gia.integ(x0,3.5,5,xa,ya);
  t.test_rel(y0,9.098214,1.0e-5,"intp 3d.");

  // ---------------------------------------------------------------
  // Test normal interpolation with a reversed array
  
  y0=gia.interp(x0,5,rxa,rya);
  t.test_rel(y0,6.232143,1.0e-5,"rintp 3a.");
  y0=gia.deriv(x0,5,rxa,rya);
  t.test_rel(y0,4.964286,1.0e-5,"rintp 3b.");
  y0=gia.deriv2(x0,5,rxa,rya);
  t.test_rel(y0,2.142857,1.0e-5,"rintp 3c.");
  y0=gia.integ(x0,3.5,5,rxa,rya);
  t.test_rel(y0,9.098214,1.0e-5,"rintp 3d.");

  // ---------------------------------------------------------------
  // Test normal interpolation with alternative interface
  
  y0=giv3a.interp(x0);
  t.test_rel(y0,6.232143,1.0e-5,"intp 4a.");
  y0=giv3a.deriv(x0);
  t.test_rel(y0,4.964286,1.0e-5,"intp 4b.");
  y0=giv3a.deriv2(x0);
  t.test_rel(y0,2.142857,1.0e-5,"intp 4c.");
  y0=giv3a.integ(x0,3.5);
  t.test_rel(y0,9.098214,1.0e-5,"intp 4d.");

  // ---------------------------------------------------------------
  // Test normal interpolation with a reversed array and 
  // alternate interface
  
  y0=giv4a.interp(x0);
  t.test_rel(y0,6.232143,1.0e-5,"rintp 4a.");
  y0=giv4a.deriv(x0);
  t.test_rel(y0,4.964286,1.0e-5,"rintp 4b.");
  y0=giv4a.deriv2(x0);
  t.test_rel(y0,2.142857,1.0e-5,"rintp 4c.");
  y0=giv4a.integ(x0,3.5);
  t.test_rel(y0,9.098214,1.0e-5,"rintp 4d.");

  // ---------------------------------------------------------------
  // Test akima

  def_interp_mgr<double[5],akima_interp> akita;
  def_interp_mgr<array_const_reverse<5>,akima_interp> rakita;
  gia.set_type(akita,rakita);
  y0=gia.interp(x0,5,xa,ya);
  t.test_rel(y0,6.25,1.0e-7,"intp 3a.");
  y0=gia.deriv(x0,5,xa,ya);
  t.test_rel(y0,5.0,1.0e-7,"intp 3b.");
  y0=gia.deriv2(x0,5,xa,ya);
  t.test_rel(y0,2.0,1.0e-7,"intp 3c.");
  y0=gia.integ(x0,3.5,5,xa,ya);
  t.test_rel(y0,9.0833333,1.0e-7,"intp 3d.");
  
  t.report();

  return 0;
}
