/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_MM_FUNCT_H
#define O2SCL_MM_FUNCT_H

#include <string>
#include <o2scl/collection.h>
#include <o2scl/ovector_tlate.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Array of multi-dimensional functions [abstract base]
      
      This class generalizes \c nv functions of \c nv variables, i.e.
      \f$ y_j(x_0,x_1,...,x_{nv-1}) \f$ for \f$ 0\leq j \leq 
      nv-1 \f$ .

      For functions with C-style arrays, use
      the corresponding children of \ref mm_vfunct .

      This class is one of a large number of function object classes
      in \o2 designed to provide a mechanism for the user to 
      supply functions to solvers, minimizers, integrators, etc.
      See \ref funct_section for a general description.
  */
  template<class param_t, class vec_t=ovector_base>
    class mm_funct {
    public:  

    mm_funct() {}
    
    virtual ~mm_funct() {}
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, vec_t &y, 
			   param_t &pa)=0;
    
#ifndef DOXYGENP

    private:

    mm_funct(const mm_funct &);
    mm_funct& operator=(const mm_funct&);

#endif

  };
  
  /** \brief Function pointer to array of multi-dimensional functions
   */
  template<class param_t, class vec_t=ovector_base>
    class mm_funct_fptr : public mm_funct<param_t,vec_t> {

    public:
    
    mm_funct_fptr() {}

    virtual ~mm_funct_fptr() {}

    /** \brief Specify the function pointer
     */
    mm_funct_fptr(int (*fp)(size_t nv, const vec_t &x, vec_t &y, 
			    param_t &pa)) {
      fptr=fp;
    }

    /** \brief Specify the function pointer
     */
    int set_function(int (*fp)(size_t nv, const vec_t &x, vec_t &y, 
			       param_t &pa)) {
      fptr=fp;
      return 0;
    }
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, vec_t &y, param_t &pa) {
      return fptr(nv,x,y,pa);
    }
    
    
#ifndef DOXYGEN_INTERNAL
    
    protected:
    
    friend class io_tlate<mm_funct_fptr>;

    /// The function pointer to the user-supplied function
    int (*fptr)(size_t nv, const vec_t &x, vec_t &y, param_t &pa);
    
    private:
    
    mm_funct_fptr(const mm_funct_fptr &);
    mm_funct_fptr& operator=(const mm_funct_fptr&);
    
#endif
    
  };
  
  /** \brief Function pointer to array of multi-dimensional functions with
      no parameters
  */
  template<class param_t, class vec_t=ovector_base>
    class mm_funct_fptr_nopar : public mm_funct<param_t,vec_t> {
    
    public:
    
    mm_funct_fptr_nopar() {}

    virtual ~mm_funct_fptr_nopar() {}

    /** \brief Specify the function pointer
     */
    mm_funct_fptr_nopar(int (*fp)(size_t nv, const vec_t &x, vec_t &y)) {
      fptr=fp;
    }

    /** \brief Specify the function pointer
     */
    int set_function(int (*fp)(size_t nv, const vec_t &x, vec_t &y)) {
      fptr=fp;
      return 0;
    }
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, vec_t &y, param_t &pa) {
      return fptr(nv,x,y);
    }
    
    
#ifndef DOXYGEN_INTERNAL
    
    protected:
    
    friend class io_tlate<mm_funct_fptr_nopar>;

    /// The function pointer to the user-supplied function
    int (*fptr)(size_t nv, const vec_t &x, vec_t &y);
    
    private:
    
    mm_funct_fptr_nopar(const mm_funct_fptr_nopar &);
    mm_funct_fptr_nopar& operator=(const mm_funct_fptr_nopar&);
    
#endif
    
  };
  
  /** 
      \brief Function pointer to a gsl_multiroot_function

      This works because with the template parameter \c vec_t as an
      \ref ovector_base class because \ref ovector_base is inherited
      from \c gsl_vector.
  */
  template<class param_t, class vec_t=ovector_base>
    class mm_funct_gsl : public mm_funct<param_t,vec_t> {
    public:
    
    /** \brief Specify the function pointer
     */
    mm_funct_gsl(int (*fp)(const gsl_vector *x, param_t &pa, gsl_vector *f)) {
      fptr=fp;
    }
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, vec_t &y, param_t &pa) {
      const gsl_vector *gx=(const gsl_vector *)(&x);
      gsl_vector *gy=(gsl_vector *)(&y);
      return fptr(gx,pa,gy);
    }
    
#ifndef DOXYGEN_INTERNAL
    
    protected:
    
    friend class io_tlate<mm_funct_gsl>;
    
    /// The function pointer to the user-supplied function
    int (*fptr)(const gsl_vector *x, param_t &pa, gsl_vector *f);
    
    private:
    
    mm_funct_gsl(const mm_funct_gsl &);
    mm_funct_gsl& operator=(const mm_funct_gsl&);
    
#endif
    
  };
  
  /** \brief Member function pointer to an array of 
      multi-dimensional functions
  */
  template<class tclass, class param_t, class vec_t=ovector_base>
    class mm_funct_mfptr : public mm_funct<param_t,vec_t> {
    public:
    
    /** \brief Empty constructor
     */
    mm_funct_mfptr() {
    }

    /** \brief Specify the member function pointer
     */
    mm_funct_mfptr(tclass *tp, int (tclass::*fp)
		   (size_t nv, const vec_t &x, vec_t &y, param_t &pa)) {
      tptr=tp;
      fptr=fp;
    }

    /** \brief Specify the member function pointer
     */
    int set_function(tclass *tp, int (tclass::*fp)
		     (size_t nv, const vec_t &x, vec_t &y, param_t &pa)) 
    {
      tptr=tp;
      fptr=fp;
      return 0;
    }
    
    virtual ~mm_funct_mfptr() {};
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, vec_t &y, param_t &pa) {
      return (*tptr.*fptr)(nv,x,y,pa);
    }
    
#ifndef DOXYGEN_INTERNAL
    
    protected:
    
    /// The member function pointer
    int (tclass::*fptr)(size_t nv, const vec_t &x, vec_t &y, param_t &pa);

    /// The class pointer
    tclass *tptr;
    
    private:
    
    mm_funct_mfptr(const mm_funct_mfptr &);
    mm_funct_mfptr& operator=(const mm_funct_mfptr&);
    
#endif
    
  };

  /** \brief Member function pointer to an array of 
      multi-dimensional functions with no parameters
  */
  template<class tclass, class param_t, class vec_t=ovector_base>
    class mm_funct_mfptr_nopar : public mm_funct<param_t,vec_t> {
    public:
    
    /** \brief Empty constructor
     */
    mm_funct_mfptr_nopar() {
    }

    /** \brief Specify the member function pointer
     */
    mm_funct_mfptr_nopar(tclass *tp, int (tclass::*fp)
			 (size_t nv, const vec_t &x, vec_t &y)) {
      tptr=tp;
      fptr=fp;
    }
    
    /** \brief Specify the member function pointer
     */
    int set_function(tclass *tp, int (tclass::*fp)
		     (size_t nv, const vec_t &x, vec_t &y)) 
    {
      tptr=tp;
      fptr=fp;
      return 0;
    }
    
    virtual ~mm_funct_mfptr_nopar() {};
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, vec_t &y, param_t &pa) {
      return (*tptr.*fptr)(nv,x,y);
    }
    
#ifndef DOXYGEN_INTERNAL
    
    protected:
    
    /// The member function pointer
    int (tclass::*fptr)(size_t nv, const vec_t &x, vec_t &y);

    /// The class pointer
    tclass *tptr;
    
    private:
    
    mm_funct_mfptr_nopar(const mm_funct_mfptr_nopar &);
    mm_funct_mfptr_nopar& operator=(const mm_funct_mfptr_nopar&);
    
#endif
    
  };

  /** \brief Const member function pointer to an array of 
      multi-dimensional functions
  */
  template<class tclass, class param_t, class vec_t=ovector_base>
    class mm_funct_cmfptr : public mm_funct<param_t,vec_t> {
    public:
    
    /** \brief Empty constructor
     */
    mm_funct_cmfptr() {
    }

    /** \brief Specify the member function pointer
     */
    mm_funct_cmfptr(tclass *tp, int (tclass::*fp)
		   (size_t nv, const vec_t &x, vec_t &y, param_t &pa) const) {
      tptr=tp;
      fptr=fp;
    }

    /** \brief Specify the member function pointer
     */
    int set_function(tclass *tp, int (tclass::*fp)
		     (size_t nv, const vec_t &x, vec_t &y, param_t &pa) const) 
    {
      tptr=tp;
      fptr=fp;
      return 0;
    }
    
    virtual ~mm_funct_cmfptr() {};
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, vec_t &y, param_t &pa) {
      return (*tptr.*fptr)(nv,x,y,pa);
    }
    
#ifndef DOXYGEN_INTERNAL
    
    protected:
    
    /// The member function pointer
    int (tclass::*fptr)(size_t nv, const vec_t &x, vec_t &y, param_t &pa) const;

    /// The class pointer
    tclass *tptr;
    
    private:
    
    mm_funct_cmfptr(const mm_funct_cmfptr &);
    mm_funct_cmfptr& operator=(const mm_funct_cmfptr&);
    
#endif
    
  };

  /** \brief Const member function pointer to an array of 
      multi-dimensional functions with no parameters
  */
  template<class tclass, class param_t, class vec_t=ovector_base>
    class mm_funct_cmfptr_nopar : public mm_funct<param_t,vec_t> {
    public:
    
    /** \brief Empty constructor
     */
    mm_funct_cmfptr_nopar() {
    }

    /** \brief Specify the member function pointer
     */
    mm_funct_cmfptr_nopar(tclass *tp, int (tclass::*fp)
			 (size_t nv, const vec_t &x, vec_t &y) const) {
      tptr=tp;
      fptr=fp;
    }
    
    /** \brief Specify the member function pointer
     */
    int set_function(tclass *tp, int (tclass::*fp)
		     (size_t nv, const vec_t &x, vec_t &y)) 
    {
      tptr=tp;
      fptr=fp;
      return 0;
    }
    
    virtual ~mm_funct_cmfptr_nopar() {};
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, vec_t &y, param_t &pa) {
      return (*tptr.*fptr)(nv,x,y);
    }
    
#ifndef DOXYGEN_INTERNAL
    
    protected:
    
    /// The member function pointer
    int (tclass::*fptr)(size_t nv, const vec_t &x, vec_t &y) const;

    /// The class pointer
    tclass *tptr;
    
    private:
    
    mm_funct_cmfptr_nopar(const mm_funct_cmfptr_nopar &);
    mm_funct_cmfptr_nopar& operator=(const mm_funct_cmfptr_nopar&);
    
#endif
    
  };

  /** 
      \brief Array of multi-dimensional functions with arrays [abstract
      base]
      
      This class generalizes \c nv functions of \c nv variables, i.e.
      \f$ y_j(x_0,x_1,...,x_{nv-1}) \f$ for \f$ 0\leq j \leq 
      nv-1 \f$.

      To use ovector_base objects instead of C-style arrays, 
      use \ref mm_funct.
      
      This class is one of a large number of function object classes
      in \o2 designed to provide a mechanism for the user to 
      supply functions to solvers, minimizers, integrators, etc.
      See \ref funct_section for a general description.
  */
  template<class param_t, size_t nv>
    class mm_vfunct {
  public:  
    
    mm_vfunct() {}

    virtual ~mm_vfunct() {}
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nvar, const double x[nv], 
			   double y[nv], param_t &pa)=0;
    
#ifndef DOXYGENP

  private:

    mm_vfunct(const mm_vfunct &);
    mm_vfunct& operator=(const mm_vfunct&);

#endif

  };
  
  /** \brief Function pointer to array of multi-dimensional functions
      with arrays
  */
  template<class param_t, size_t nv> class mm_vfunct_fptr : 
  public mm_vfunct<param_t,nv> {
  public:
    
    /** \brief Specify the function pointer
     */
    mm_vfunct_fptr(int (*fp)(size_t nvar, const double x[nv], double y[nv], 
			     param_t &pa)) {
      fptr=fp;
    }
    
    
    virtual ~mm_vfunct_fptr() {};
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nvar, const double x[nv], double y[nv], 
			   param_t &pa) {
      return fptr(nv,x,y,pa);
    }
    
    
#ifndef DOXYGEN_INTERNAL
    
  protected:
    
    friend class io_tlate<mm_vfunct_fptr>;
    
    mm_vfunct_fptr() {};
    
    /// The function pointer
    int (*fptr)(size_t nvar, const double x[nv], double y[nv], param_t &pa);
    
  private:
    
    mm_vfunct_fptr(const mm_vfunct_fptr &);
    mm_vfunct_fptr& operator=(const mm_vfunct_fptr&);
    
#endif
    
  };
    
  /** \brief Function pointer to array of multi-dimensional functions
      with arrays and no parameters
  */
  template<class param_t, size_t nv> class mm_vfunct_fptr_nopar : 
  public mm_vfunct<param_t,nv> 
    {
    public:
	
      /** \brief Specify the function pointer
       */
      mm_vfunct_fptr_nopar(int (*fp)(size_t nvar, const double x[nv], 
				     double y[nv])) {
	fptr=fp;
      }
      
      
      virtual ~mm_vfunct_fptr_nopar() {};
      
      /** \brief Compute \c nv functions, \c y, of \c nv variables
	  stored in \c x with parameter \c pa.
      */
      virtual int operator()(size_t nvar, const double x[nv], double y[nv],
			     param_t &pa) {
	return fptr(nv,x,y);
      }
    
    
#ifndef DOXYGEN_INTERNAL
    
    protected:
    
      friend class io_tlate<mm_vfunct_fptr_nopar>;
    
      mm_vfunct_fptr_nopar() {};
    
      /// The function pointer
      int (*fptr)(size_t nvar, const double x[nv], double y[nv]);
    
    private:
    
      mm_vfunct_fptr_nopar(const mm_vfunct_fptr_nopar &);
      mm_vfunct_fptr_nopar& operator=(const mm_vfunct_fptr_nopar&);
    
#endif
    
    };
  
  /** \brief Function pointer to a gsl_multiroot_function with arrays
   */
  template<class param_t, size_t nv>
    class mm_vfunct_gsl : public mm_vfunct<param_t,nv> {
  public:
    
    /** \brief Specify the function pointer
     */
    mm_vfunct_gsl(int (*fp)(const gsl_vector *x, param_t &pa, gsl_vector *f)) {
      fptr=fp;
    }
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nvar, const double x[nv], 
			   double y[nv], param_t &pa) {
      const gsl_vector *gx=(const gsl_vector *)(&x);
      gsl_vector *gy=(gsl_vector *)(&y);
      return fptr(gx,pa,gy);
    }
    
#ifndef DOXYGEN_INTERNAL
    
  protected:
    
    friend class io_tlate<mm_vfunct_gsl>;
    
    /// The function pointer
    int (*fptr)(const gsl_vector *x, param_t &pa, gsl_vector *f);
    
  private:
    
    mm_vfunct_gsl(const mm_vfunct_gsl &);
    mm_vfunct_gsl& operator=(const mm_vfunct_gsl&);
    
#endif
    
  };
  
  /** \brief Member function pointer to an array of 
      multi-dimensional functions with arrays
  */
  template<class tclass, class param_t, size_t nv>
    class mm_vfunct_mfptr : public mm_vfunct<param_t,nv> {
  public:
    
    /** \brief Specify the member function pointer
     */
    mm_vfunct_mfptr(tclass *tp, int (tclass::*fp)
		    (size_t nvar, const double x[nv], double y[nv], 
		     param_t &pa)) {
      tptr=tp;
      fptr=fp;
    }
    
    virtual ~mm_vfunct_mfptr() {};
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nvar, const double x[nv], double y[nv], 
			   param_t &pa) {
      return (*tptr.*fptr)(nv,x,y,pa);
    }
    
#ifndef DOXYGEN_INTERNAL
    
  protected:
    
    /// The member function pointer
    int (tclass::*fptr)(size_t nvar, const double x[nv], double y[nv], 
			param_t &pa);

    /// The class pointer
    tclass *tptr;
    
  private:
    
    mm_vfunct_mfptr(const mm_vfunct_mfptr &);
    mm_vfunct_mfptr& operator=(const mm_vfunct_mfptr&);
    
#endif
    
  };

  /** \brief Const member function pointer to an array of 
      multi-dimensional functions with arrays with no parameters
  */
  template<class tclass, class param_t, size_t nv>
    class mm_vfunct_mfptr_nopar : public mm_vfunct<param_t,nv> {

  public:
    
    /** \brief Specify the member function pointer
     */
    mm_vfunct_mfptr_nopar(tclass *tp, int (tclass::*fp)
			  (size_t nvar, const double x[nv], 
			   double y[nv])) {
      tptr=tp;
      fptr=fp;
    }
    
    virtual ~mm_vfunct_mfptr_nopar() {};
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nvar, const double x[nv], double y[nv], 
			   param_t &pa) {
      return (*tptr.*fptr)(nv,x,y);
    }
    
#ifndef DOXYGEN_INTERNAL
    
  protected:
    
    /// The member function pointer
    int (tclass::*fptr)(size_t nvar, const double x[nv], double y[nv]);

    /// The class pointer
    tclass *tptr;
    
  private:
    
    mm_vfunct_mfptr_nopar(const mm_vfunct_mfptr_nopar &);
    mm_vfunct_mfptr_nopar& operator=(const mm_vfunct_mfptr_nopar&);
    
#endif
    
  };

  /** \brief Const member function pointer to an array of 
      multi-dimensional functions with arrays
  */
  template<class tclass, class param_t, size_t nv>
    class mm_vfunct_cmfptr : public mm_vfunct<param_t,nv> {
  public:
    
    /** \brief Specify the member function pointer
     */
    mm_vfunct_cmfptr(tclass *tp, int (tclass::*fp)
		    (size_t nvar, const double x[nv], double y[nv], 
		     param_t &pa) const) {
      tptr=tp;
      fptr=fp;
    }
    
    virtual ~mm_vfunct_cmfptr() {};
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nvar, const double x[nv], double y[nv], 
			   param_t &pa) {
      return (*tptr.*fptr)(nv,x,y,pa);
    }
    
#ifndef DOXYGEN_INTERNAL
    
  protected:
    
    /// The member function pointer
    int (tclass::*fptr)(size_t nvar, const double x[nv], double y[nv], 
			param_t &pa) const;

    /// The class pointer
    tclass *tptr;
    
  private:
    
    mm_vfunct_cmfptr(const mm_vfunct_cmfptr &);
    mm_vfunct_cmfptr& operator=(const mm_vfunct_cmfptr&);
    
#endif
    
  };

  /** \brief Member function pointer to an array of 
      multi-dimensional functions with arrays with no parameters
  */
  template<class tclass, class param_t, size_t nv>
    class mm_vfunct_cmfptr_nopar : public mm_vfunct<param_t,nv> {

  public:
    
    /** \brief Specify the member function pointer
     */
    mm_vfunct_cmfptr_nopar(tclass *tp, int (tclass::*fp)
			   (size_t nvar, const double x[nv], double y[nv])
			   const) {
      tptr=tp;
      fptr=fp;
    }
    
    virtual ~mm_vfunct_cmfptr_nopar() {};
    
    /** \brief Compute \c nv functions, \c y, of \c nv variables
	stored in \c x with parameter \c pa.
    */
    virtual int operator()(size_t nvar, const double x[nv], double y[nv], 
			   param_t &pa) {
      return (*tptr.*fptr)(nv,x,y);
    }
    
#ifndef DOXYGEN_INTERNAL
    
  protected:
    
    /// The member function pointer
    int (tclass::*fptr)(size_t nvar, const double x[nv], double y[nv]) const;

    /// The class pointer
    tclass *tptr;
    
  private:
    
    mm_vfunct_cmfptr_nopar(const mm_vfunct_cmfptr_nopar &);
    mm_vfunct_cmfptr_nopar& operator=(const mm_vfunct_cmfptr_nopar&);
    
#endif
    
  };

#ifndef DOXYGENP
}
#endif

#endif
