/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_MULTI_FUNCT_H
#define O2SCL_MULTI_FUNCT_H

#include <string>
#include <o2scl/collection.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Multi-dimensional function [abstract base]

      This class generalizes one function of several variables,
      i.e. \f$ y(x_0,x_1,...,x_{nv-1}) \f$ where \c nv is the number
      of variables in the function y. 

      For functions with C-style arrays, use
      the corresponding children of \ref multi_vfunct .

      This class is one of a large number of function object classes
      in \o2 designed to provide a mechanism for the user to 
      supply functions to solvers, minimizers, integrators, etc.
      See \ref funct_section for a general description.

  */

  template<class param_t, class vec_t=ovector_base> class multi_funct {

    public:  

    multi_funct() {}

    virtual ~multi_funct() {}

    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, double &y, 
			   param_t &pa)=0;
    

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const vec_t &x, param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }
  
#ifndef DOXYGENP

    private:

    multi_funct(const multi_funct &);
    multi_funct& operator=(const multi_funct&);

#endif

  };

  /** \brief Function pointer to a multi-dimensional function
   */
  template<class param_t, class vec_t=ovector_base>
    class multi_funct_fptr : public multi_funct<param_t,vec_t> {

    public:

    /** \brief Specify the function pointer
     */
    multi_funct_fptr(int (*fp)(size_t nv, const vec_t &x, double &y, 
			       param_t &pa)) {
      fptr=fp;
    }
    

    virtual ~multi_funct_fptr() {};

    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, double &y, param_t &pa) {
      return fptr(nv,x,y,pa);
    }

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const vec_t &x, param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

    protected:
    
    friend class io_tlate<multi_funct_fptr>;

    /// Store the function pointer
    int (*fptr)(size_t nv, const vec_t &x, double &y, param_t &pa);

    multi_funct_fptr() {}

#ifndef DOXYGENP
#endif

    private:

    multi_funct_fptr(const multi_funct_fptr &);
    multi_funct_fptr& operator=(const multi_funct_fptr&);

#endif

  };

  /** \brief Function pointer to a gsl_multimin_function
   */
  template<class param_t, class vec_t=ovector_base>
    class multi_funct_gsl : public multi_funct<param_t,vec_t> {
    public:

    /** \brief Specify the function pointer
     */
    multi_funct_gsl(double (*fp)(const gsl_vector *x, param_t &pa)) {
      fptr=fp;
    }
    
    virtual ~multi_funct_gsl() {}

    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, double &y, param_t &pa) {
      const gsl_vector *gx=(const gsl_vector *)(&x);
      y=fptr(gx,pa);
      return 0;
    }

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const vec_t &x, param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

    protected:
  
    friend class io_tlate<multi_funct_gsl>;
    
    /// Store the function pointer
    double (*fptr)(const gsl_vector *x, param_t &pa);
    
    multi_funct_gsl() {}

#ifndef DOXYGENP
#endif

    private:

    multi_funct_gsl(const multi_funct_gsl &);
    multi_funct_gsl& operator=(const multi_funct_gsl&);

#endif

  };

  /** \brief Function pointer to a multi-dimensional function without
      error control
  */
  template<class param_t, class vec_t=ovector_base>
    class multi_funct_fptr_noerr : public multi_funct<param_t,vec_t> {
    public:

    /** \brief Specify the function pointer
     */
    multi_funct_fptr_noerr(double (*fp)(size_t nv, const vec_t &x, 
					param_t &pa)) {
      fptr=fp;
    }

    virtual ~multi_funct_fptr_noerr() {}
    
    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, double &y, param_t &pa) {
      y=fptr(nv,x,pa);
      return 0;
    }

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const vec_t &x, param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

    protected:

    friend class io_tlate<multi_funct_fptr_noerr>;
  
    multi_funct_fptr_noerr() {}

    /// Store the function pointer
    double (*fptr)(size_t nv, const vec_t &x, param_t &pa);

#ifndef DOXYGENP
#endif

    private:

    multi_funct_fptr_noerr(const multi_funct_fptr_noerr &);
    multi_funct_fptr_noerr& operator=(const multi_funct_fptr_noerr&);

#endif

  };

  /** \brief Member function pointer to a multi-dimensional function
   */
  template<class tclass, class param_t, class vec_t=ovector_base>
    class multi_funct_mfptr : public multi_funct<param_t,vec_t> {
    public:
  
    /** \brief Specify the member function pointer
     */
    multi_funct_mfptr(tclass *tp, int (tclass::*fp)
		      (size_t nv, const vec_t &x, double &y, param_t &pa)) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~multi_funct_mfptr() {}
  
    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, double &y, param_t &pa) {
      return (*tptr.*fptr)(nv,x,y,pa);
    }

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const vec_t &x, param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }
  
#ifndef DOXYGEN_INTERNAL
  
    protected:
  
    /// Store the function pointer
    int (tclass::*fptr)(size_t nv, const vec_t &x, double &y, param_t &pa);
    /// Store a pointer to the class instance
    tclass *tptr;
  
#ifndef DOXYGENP
#endif

    private:

    multi_funct_mfptr(const multi_funct_mfptr &);
    multi_funct_mfptr& operator=(const multi_funct_mfptr&);

#endif

  };

  /** \brief Member function pointer to a multi-dimensional function
   */
  template<class tclass, class param_t, class vec_t=ovector_base>
    class multi_funct_mfptr_noerr : public multi_funct<param_t,vec_t> {
    public:
  
    /** \brief Specify the member function pointer
     */
    multi_funct_mfptr_noerr(tclass *tp, double (tclass::*fp)
			    (size_t nv, const vec_t &x, param_t &pa)) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~multi_funct_mfptr_noerr() {}
  
    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, double &y, param_t &pa) {
      y=(*tptr.*fptr)(nv,x,pa);
      return 0;
    }

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const vec_t &x, param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }
  
#ifndef DOXYGEN_INTERNAL
  
    protected:
  
    /// Store the function pointer
    double (tclass::*fptr)(size_t nv, const vec_t &x, param_t &pa);
    /// Store a pointer to the class instance
    tclass *tptr;
  
#ifndef DOXYGENP
#endif

    private:

    multi_funct_mfptr_noerr(const multi_funct_mfptr_noerr &);
    multi_funct_mfptr_noerr& operator=(const multi_funct_mfptr_noerr&);

#endif

  };
  
  /** \brief Const member function pointer to a multi-dimensional function
   */
  template<class tclass, class param_t, class vec_t=ovector_base>
    class multi_funct_cmfptr : public multi_funct<param_t,vec_t> {
    public:
  
    /** \brief Specify the member function pointer
     */
    multi_funct_cmfptr(tclass *tp, int (tclass::*fp)
		       (size_t nv, const vec_t &x, double &y, 
		       param_t &pa) const) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~multi_funct_cmfptr() {}
  
    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, double &y, param_t &pa) {
      return (*tptr.*fptr)(nv,x,y,pa);
    }

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const vec_t &x, param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }
  
#ifndef DOXYGEN_INTERNAL
  
    protected:
  
    /// Store the function pointer
    int (tclass::*fptr)(size_t nv, const vec_t &x, double &y, 
			param_t &pa) const;
    /// Store a pointer to the class instance
    tclass *tptr;
  
#ifndef DOXYGENP
#endif

    private:

    multi_funct_cmfptr(const multi_funct_cmfptr &);
    multi_funct_cmfptr& operator=(const multi_funct_cmfptr&);

#endif

  };

  /** \brief Const member function pointer to a multi-dimensional function
   */
  template<class tclass, class param_t, class vec_t=ovector_base>
    class multi_funct_cmfptr_noerr : public multi_funct<param_t,vec_t> {
    public:
  
    /** \brief Specify the member function pointer
     */
    multi_funct_cmfptr_noerr(tclass *tp, double (tclass::*fp)
			    (size_t nv, const vec_t &x, param_t &pa) const) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~multi_funct_cmfptr_noerr() {}
  
    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const vec_t &x, double &y, param_t &pa) {
      y=(*tptr.*fptr)(nv,x,pa);
      return 0;
    }

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const vec_t &x, param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }
  
#ifndef DOXYGEN_INTERNAL
  
    protected:
  
    /// Store the function pointer
    double (tclass::*fptr)(size_t nv, const vec_t &x, param_t &pa) const;
    /// Store a pointer to the class instance
    tclass *tptr;
  
#ifndef DOXYGENP
#endif

    private:

    multi_funct_cmfptr_noerr(const multi_funct_cmfptr_noerr &);
    multi_funct_cmfptr_noerr& operator=(const multi_funct_cmfptr_noerr&);

#endif

  };
  
  // ---------------------------------------------------------------------
  // ---------------------------------------------------------------------

  /** 
      \brief Multi-dimensional function base with arrays [abstract base]

      This class generalizes one function of several variables,
      i.e.  \f$ y(x_0,x_1,...,x_{nv-1}) \f$  where \c nv  is the number
      of variables in the function y. 

      This class is one of a large number of function object classes
      in \o2 designed to provide a mechanism for the user to 
      supply functions to solvers, minimizers, integrators, etc.
      See \ref funct_section for a general description.
  */
  template<class param_t, size_t nvar> class multi_vfunct {

    public:  

    multi_vfunct() {}

    virtual ~multi_vfunct() {}
    
    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const double x[nvar], double &y, 
			   param_t &pa)=0;
    
    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const double x[nvar], param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }
  
#ifndef DOXYGENP

    private:

    multi_vfunct(const multi_vfunct &);
    multi_vfunct& operator=(const multi_vfunct&);

#endif

  };

  /** \brief Function pointer to a multi-dimensional function with arrays
   */
  template<class param_t, size_t nvar>
    class multi_vfunct_fptr : public multi_vfunct<param_t,nvar> {

    public:

    /** \brief Specify the function pointer
     */
    multi_vfunct_fptr(int (*fp)(size_t nv, const double x[nvar], double &y, 
			       param_t &pa)) {
      fptr=fp;
    }
    

    virtual ~multi_vfunct_fptr() {}

    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const double x[nvar], double &y, 
			   param_t &pa) {
      return fptr(nv,x,y,pa);
    }
    
    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const double x[nvar], param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

    protected:
    
    friend class io_tlate<multi_vfunct_fptr>;

    /// Store the function pointer
    int (*fptr)(size_t nv, const double x[nvar], double &y, param_t &pa);

    multi_vfunct_fptr() {}

#ifndef DOXYGENP
#endif

    private:

    multi_vfunct_fptr(const multi_vfunct_fptr &);
    multi_vfunct_fptr& operator=(const multi_vfunct_fptr&);

#endif

  };

  /** \brief Function pointer to a gsl_multimin_function with arrays
   */
  template<class param_t, size_t nvar>
    class multi_vfunct_gsl : public multi_vfunct<param_t,nvar> {
    public:

    /** \brief Specify the function pointer
     */
    multi_vfunct_gsl(double (*fp)(const gsl_vector *x, param_t &pa)) {
      fptr=fp;
    }
    
    virtual ~multi_vfunct_gsl() {}

    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const double x[nvar], double &y, 
			   param_t &pa) {
      const gsl_vector *gx=(const gsl_vector *)(&x);
      y=fptr(gx,pa);
      return 0;
    }

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const double x[nvar], param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

    protected:
  
    friend class io_tlate<multi_vfunct_gsl>;
    
    /// Store the function pointer
    double (*fptr)(const gsl_vector *x, param_t &pa);
    
    multi_vfunct_gsl() {}

#ifndef DOXYGENP
#endif

    private:

    multi_vfunct_gsl(const multi_vfunct_gsl &);
    multi_vfunct_gsl& operator=(const multi_vfunct_gsl&);

#endif

  };

  /** \brief Function pointer to a multi-dimensional function with
      arrays and without error control
  */
  template<class param_t, size_t nvar>
    class multi_vfunct_fptr_noerr : public multi_vfunct<param_t,nvar> {
    public:
    
    /** \brief Specify the function pointer
     */
    multi_vfunct_fptr_noerr(double (*fp)(size_t nv, const double x[nvar], 
					 param_t &pa)) {
      fptr=fp;
    }
    
    virtual ~multi_vfunct_fptr_noerr() {}

    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const double x[nvar], double &y, 
			   param_t &pa) {
      y=fptr(nv,x,pa);
      return 0;
    }

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const double x[nvar], param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }

#ifndef DOXYGEN_INTERNAL

    protected:

    friend class io_tlate<multi_vfunct_fptr_noerr>;
  
    multi_vfunct_fptr_noerr() {}

    /// Store the function pointer
    double (*fptr)(size_t nv, const double x[nvar], param_t &pa);

#ifndef DOXYGENP
#endif

    private:

    multi_vfunct_fptr_noerr(const multi_vfunct_fptr_noerr &);
    multi_vfunct_fptr_noerr& operator=(const multi_vfunct_fptr_noerr&);

#endif

  };

  /** \brief Member function pointer to a multi-dimensional function
      with arrays
   */
  template<class tclass, class param_t, size_t nvar>
    class multi_vfunct_mfptr : public multi_vfunct<param_t,nvar> {
    public:
  
    /** \brief Specify the member function pointer
     */
    multi_vfunct_mfptr(tclass *tp, int (tclass::*fp)
		       (size_t nv, const double x[nvar], double &y, 
			param_t &pa)) {
      tptr=tp;
      fptr=fp;
    }
    
    virtual ~multi_vfunct_mfptr() {}
    
    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const double x[nvar], double &y, 
			   param_t &pa) {
      return (*tptr.*fptr)(nv,x,y,pa);
    }

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const double x[nvar], param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }
  
#ifndef DOXYGEN_INTERNAL
  
    protected:
  
    /// Store the function pointer
    int (tclass::*fptr)(size_t nv, const double x[nvar], double &y, 
			param_t &pa);

    /// Store a pointer to the class instance
    tclass *tptr;
  
#ifndef DOXYGENP
#endif

    private:

    multi_vfunct_mfptr(const multi_vfunct_mfptr &);
    multi_vfunct_mfptr& operator=(const multi_vfunct_mfptr&);

#endif

  };

  /** \brief Member function pointer to a multi-dimensional function
      with arrays
   */
  template<class tclass, class param_t, size_t nvar>
    class multi_vfunct_mfptr_noerr : public multi_vfunct<param_t,nvar> {
    public:
  
    /** \brief Specify the member function pointer
     */
    multi_vfunct_mfptr_noerr(tclass *tp, double (tclass::*fp)
			    (size_t nv, const double x[nvar], param_t &pa)) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~multi_vfunct_mfptr_noerr() {}
  
    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const double x[nvar], double &y, 
			   param_t &pa) {
      y=(*tptr.*fptr)(nv,x,pa);
      return 0;
    }

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const double x[nvar], param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }
  
#ifndef DOXYGEN_INTERNAL
  
    protected:
  
    /// Store the function pointer
    double (tclass::*fptr)(size_t nv, const double x[nvar], param_t &pa);

    /// Store a pointer to the class instance
    tclass *tptr;
  
#ifndef DOXYGENP
#endif

    private:

    multi_vfunct_mfptr_noerr(const multi_vfunct_mfptr_noerr &);
    multi_vfunct_mfptr_noerr& operator=(const multi_vfunct_mfptr_noerr&);

#endif

  };
  
  /** \brief Member function pointer to a multi-dimensional function
      with arrays
   */
  template<class tclass, class param_t, size_t nvar>
    class multi_vfunct_cmfptr : public multi_vfunct<param_t,nvar> {
    public:
  
    /** \brief Specify the member function pointer
     */
    multi_vfunct_cmfptr(tclass *tp, int (tclass::*fp)
		       (size_t nv, const double x[nvar], double &y, 
			param_t &pa) const) {
      tptr=tp;
      fptr=fp;
    }
    
    virtual ~multi_vfunct_cmfptr() {}
    
    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const double x[nvar], double &y, 
			   param_t &pa) {
      return (*tptr.*fptr)(nv,x,y,pa);
    }

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const double x[nvar], param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }
  
#ifndef DOXYGEN_INTERNAL
  
    protected:
  
    /// Store the function pointer
    int (tclass::*fptr)(size_t nv, const double x[nvar], double &y, 
			param_t &pa) const;

    /// Store a pointer to the class instance
    tclass *tptr;
  
#ifndef DOXYGENP
#endif

    private:

    multi_vfunct_cmfptr(const multi_vfunct_cmfptr &);
    multi_vfunct_cmfptr& operator=(const multi_vfunct_cmfptr&);

#endif

  };

  /** \brief Member function pointer to a multi-dimensional function
      with arrays
   */
  template<class tclass, class param_t, size_t nvar>
    class multi_vfunct_cmfptr_noerr : public multi_vfunct<param_t,nvar> {
    public:
  
    /** \brief Specify the member function pointer
     */
    multi_vfunct_cmfptr_noerr(tclass *tp, double (tclass::*fp)
			    (size_t nv, const double x[nvar], 
			     param_t &pa) const) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~multi_vfunct_cmfptr_noerr() {}
  
    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual int operator()(size_t nv, const double x[nvar], double &y, 
			   param_t &pa) {
      y=(*tptr.*fptr)(nv,x,pa);
      return 0;
    }

    /** \brief Return the value of a function of \c nv variables
	stored in \c x with parameter \c pa.

	Note that this is reimplemented in all children because 
	if one member function operator() is reimplemented, all must be.
    */
    virtual double operator()(size_t nv, const double x[nvar], param_t &pa) {
      double y;
      operator()(nv,x,y,pa);
      return y;
    }
  
#ifndef DOXYGEN_INTERNAL
  
    protected:
  
    /// Store the function pointer
    double (tclass::*fptr)(size_t nv, const double x[nvar], param_t &pa) const;

    /// Store a pointer to the class instance
    tclass *tptr;
  
#ifndef DOXYGENP
#endif

    private:

    multi_vfunct_cmfptr_noerr(const multi_vfunct_cmfptr_noerr &);
    multi_vfunct_cmfptr_noerr& operator=(const multi_vfunct_cmfptr_noerr&);

#endif

  };
  
#ifndef DOXYGENP
}
#endif

#endif
