/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_SEARCH_VEC_H
#define O2SCL_SEARCH_VEC_H

#include <iostream>
#include <string>
#include <gsl/gsl_spline.h>
#include <o2scl/err_hnd.h>
#include <o2scl/collection.h>
#include <o2scl/ovector_tlate.h>

/* \file search_vec.h
   \brief File for definitions for \ref search_vec

   \comment 
   This file isn't included in Doxygen at the moment since there
   isn't anything in this file other than the search_vec class.
   \endcomment
*/

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Searching class for monotonic data with caching

      A searching class for monotonic vectors. A caching system
      similar to \c gsl_interp_accel is used.
      
      To find the interval containing a value, use find_interval(). If
      you happen to know in advance that the vector is increasing or
      decreasing, then you can use find_interval_inc() or
      find_interval_dec() instead. To ignore the caching and just use
      straight binary search, you can use bsearch_inc() or
      bsearch_dec() for increasing or decreasing arrays respectively.

      If you want to allow the function to return <tt>n-1</tt>, to 
      indicate the value is larger than or equal to the last element,
      use find_interval_uncons().

      Alternatively, if you just want to find the index with the
      element closest to a specified value, use ordered_lookup(). Note
      that ordered_lookup() is slightly different than \ref
      ovector_tlate::lookup(), since the latter does not assume the
      data is monotonic.
  */
  template<class vec_t> class search_vec {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// Storage for the most recent index
    size_t cache;

#endif

  public:

    search_vec() {
      cache=0;
    }

    /** \brief Search an increasing or decreasing vector for the interval
	containing <tt>x0</tt>
     */
    size_t find_interval(const double x0, size_t n, const vec_t &x) {
      if (x[0]<x[n-1]) return find_interval_inc(x0,n,x);
      return find_interval_dec(x0,n,x);
    }

    /** 
	\brief Search an increasing vector for the interval
	containing <tt>x0</tt>

	This function operates in the same way as
	<tt>gsl_interp_accel_find()</tt>, except that it does not
	record cache hits and misses.
     */
    size_t find_interval_inc(const double x0, size_t n, const vec_t &x) {
      size_t x_index=cache;

      // Make sure the cached value is not outside the array.
      if (x_index>n-1) x_index=n/2;
  
      if (x0<x[x_index]) {
	cache=bsearch_inc(x0,x,0,x_index);
      } else if (x0>=x[x_index+1]) {
	cache=bsearch_inc(x0,x,x_index,n-1);
      }
  
      return cache;
    }
    
    /** \brief Search a decreasing vector for the interval
	containing <tt>x0</tt>
    */
    size_t find_interval_dec(const double x0, size_t n, const vec_t &x) {
      size_t x_index=cache;
      
      // Make sure the cached value is not outside the array.
      if (x_index>n-1) x_index=n/2;
  
      if (x0>x[x_index]) {
	cache=bsearch_dec(x0,x,0,x_index);
      } else if (x0<x[x_index+1]) {
	cache=bsearch_dec(x0,x,x_index,n-1);
      }
  
      return cache;
    }

    /** 
	\brief Find the index of x0 in the ordered array \c x 

	This returns the index i for which x[i] is as close
	as possible to x0 if x[i] is either increasing or decreasing.

	If some of the values in the ovector are not finite, then
	the output of this function is not defined.

	If you have a non-monotonic \ref ovector or \ref uvector
	object, consider using \ref ovector_view_tlate::lookup() or
	\ref uvector_view_tlate::lookup() instead of this function.

	Generally, if there are two adjacent entries with the same
	value, this function will return the entry with the smaller
	index. (This is the same as the convention in \ref
	ovector_view_tlate::lookup() and \ref
	uvector_view_tlate::lookup() ).

	\future This is not as efficient as it could be, as it just
	uses the <tt>find_interval</tt> functions and then adjusts the
	answer at the end if necessary.
    */
    size_t ordered_lookup(const double x0, size_t n, const vec_t &x) {
      size_t row;

      if (x[0]<=x[n-1]) {

	// Increasing case

	if (x0>=x[n-1]) {
	  row=n-1;
	} else { 
	  row=find_interval_inc(x0,n,x);
	  if (row<n-1 && fabs(x[row+1]-x0)<fabs(x[row]-x0)) row++;
	}
    
      } else {

	// Decreasing case
    
	if (x0<=x[n-1]) {
	  row=n-1;
	} else {
	  row=find_interval_dec(x0,n,x);
	  if (row<n-1 && fabs(x[row+1]-x0)<fabs(x[row]-x0)) row++;
	}
      }

      return row;
    }

    /** \brief Search an increasing or decreasing vector for the interval
	containing <tt>x0</tt> including the open interval at the end

	This returns the index i for which x[i]<=x0<x[i+1]. 

	This function operates just as find_interval(), except that in
	the case of increasing arrays, it will return <tt>n-1</tt> if
	<tt>x0</tt> is greater than the last element in the array. The
	decreasing case is handled analogously.

	If some of the values in the vector are not finite, then the
	output of this function is not defined.

    */
    size_t find_interval_uncons(const double x0, size_t n, const vec_t &x) {
      size_t row;

      if (x[0]<=x[n-1]) {

	// Increasing case

	if (x0>=x[n-1]) row=n-1;
	else row=find_interval_inc(x0,n,x);

      } else {
	// Decreasing case
    
	if (x0<=x[n-1]) row=n-1;
	else row=find_interval_dec(x0,n,x);
      }

      return row;
    }

    /** 
	\brief Binary search a part of an increasing vector for
	the interval containing <tt>x0</tt>.

	This function performs a binary search of between
	<tt>x[lo]</tt> and <tt>x[hi]</tt> (inclusive). It returns 
	- \c lo if \c x0 < <tt>x[lo]</tt>
	- \c i if <tt>x[i]</tt> <= \c x0 < <tt>x[i+1]</tt> 
	for \c lo <= \c i < \c hi
	- \c hi-1 if \c x0 >= \c <tt>x[hi]</tt>
	
	This function operates in the same way as
	<tt>gsl_interp_bsearch()</tt>.
    
	The cache is not used for this function.

	\note The value of \c hi for an <tt>n</tt> element array
	is typically <tt>n-1</tt>, e.g. For <tt>double x[10]</tt>
	one would use <tt>bsearch_inc(1.0,x,0,9)</tt>. 
     */
    size_t bsearch_inc(const double x0, const vec_t &x,
		       size_t lo, size_t hi) const {
      while (hi>lo+1) {
	size_t i=(hi+lo)/2;
	if (x[i]>x0) {
	  hi=i;
	} else {
	  lo=i;
	}
      }
  
      return lo;
    }

    /** 
	\brief Binary search a part of an decreasing vector for
	the interval containing <tt>x0</tt>.
	
	This function performs a binary search of between
	<tt>x[lo]</tt> and <tt>x[hi]</tt> (inclusive). It returns 
	- \c lo if \c x0 > <tt>x[lo]</tt>
	- \c i if <tt>x[i]</tt> >= \c x0 > <tt>x[i+1]</tt> 
	for \c lo <= \c i < \c hi
	- \c hi-1 if \c x0 <= \c <tt>x[hi]</tt>
    
	The cache is not used for this function.

	\note The value of \c hi for an <tt>n</tt> element array
	is typically <tt>n-1</tt>, e.g. For <tt>double x[10]</tt>
	one would use <tt>bsearch_dec(1.0,x,0,9)</tt>. 
     */
    size_t bsearch_dec(const double x0, const vec_t &x,
		       size_t lo, size_t hi) const {
      while (hi>lo+1) {
	size_t i=(hi+lo)/2;
	if (x[i]<x0) {
	  hi=i;
	} else {
	  lo=i;
	}
      }
  
      return lo;
    }

  };
  
#ifndef DOXYGENP
}
#endif

#endif
