/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ANNEAL_MT_H
#define O2SCL_ANNEAL_MT_H

#include <o2scl/gsl_rnga.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/multi_funct.h>

// for boost::bind()
#include <boost/bind.hpp>
// for boost::thread
#include <boost/thread/thread.hpp>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief Multidimensional minimization by simulated annealing 
      (Boost multi-threaded version)

      See also the unthreaded version of this class in \ref gsl_anneal.

      \future There may be a good way to remove the function indirection
      here to make this class a bit faster.
  */
  template<class param_t, class param_vec_t=std::vector<param_t>,
    class func_t=multi_funct<param_t>, class vec_t=ovector_base,
    class alloc_vec_t=ovector, class alloc_t=ovector_alloc,
    class rng_t=gsl_rnga> class anneal_mt {

  public:
  
  anneal_mt() {
    ntrial=100;
    nstep=0;
    nproc=1;
    outs=&std::cout;
    ins=&std::cin;
  }

  virtual ~anneal_mt() {
    if (nstep>0) {
      delete[] step_sizes;
    }
  }

  /// \name Basic usage
  //@{
  /** \brief Calculate the minimum \c fmin of \c func w.r.t the 
      array \c x0 of size \c nv using \c np threads.
  */
  virtual int mmin(size_t nv, vec_t &x0, double &fmin, 
		   func_t &func, size_t np, param_vec_t &pars) {
    
    if (nv==0) {
      O2SCL_ERR2_RET("Tried to minimize over zero variables ",
		     " in anneal_mt::mmin().",gsl_einval);
    }
    if (np==0) {
      O2SCL_ERR2_RET("Tried to use 0 threads in ",
		     "anneal_mt::mmin().",gsl_einval);
    }

    allocate(nv,np);
    params=&pars;
    f=&func;
    fmin=0.0;

    double E, best_E, T, old_E;
    int i, iter=0;
    size_t j;

    for(j=0;j<nv;j++) {
      x[j]=x0[j];
      best_x[j]=x0[j];
    }
    
    E=func(nv,x,pars[0]);
    best_E=E;

    // Setup initial temperature and step sizes
    start(nv,T);

    bool done=false;
    
    boost::thread **thrd;
    
    while (!done) {

      // Copy old value of x for next() function

      for(j=0;j<nv;j++) old_x[j]=x[j];
      old_E=E;
	  
      size_t nmoves=0;

      for (i=0;i<this->ntrial;++i) {

	// Determine the stepsize and call the function
	thrd=new boost::thread *[np];
	for(size_t ip=0;ip<np;ip++) {
	  new_x[ip]=x;
	  thrd[ip]=new boost::thread
	    (boost::bind(&anneal_mt::func_wrapper,this,ip));
	}
	// Wait until all the threads are done
	for(size_t ip=0;ip<np;ip++) {
	  thrd[ip]->join();
	}
	// Delete the threads and continue
	for(size_t ip=0;ip<np;ip++) {
	  delete thrd[ip];
	}
	delete[] thrd;
	// Process the results from each thread
	for(size_t ip=0;ip<np;ip++) {

	  // Store best value obtained so far
	  if(new_E[ip]<=best_E){
	    for(j=0;j<nv;j++) best_x[j]=new_x[ip][j];
	    best_E=new_E[ip];
	    std::cout << "Best: " << best_x << " " << best_E << std::endl;
	  }

	  // Take the crucial step: see if the new point is accepted
	  // or not, as determined by the boltzmann probability
	  if (new_E[ip]<E) {
	    for(j=0;j<nv;j++) x[j]=new_x[ip][j];
	    E=new_E[ip];
	    nmoves++;
	  } else if (this->def_rng.random() < exp(-(new_E[ip]-E)/(boltz*T)) ) {
	    for(j=0;j<nv;j++) x[j]=new_x[ip][j];
	    E=new_E[ip];
	    nmoves++;
	  }
	}
	
      }
      
      if (this->verbose>0) {
	print_iter(nv,best_x,best_E,iter,T,"gsl_anneal");
	iter++;
      }
      
      // See if we're finished and proceed to the next step
      next(nv,old_x,old_E,x,E,T,nmoves,done);
      
    }
    
    for(j=0;j<nv;j++) x0[j]=best_x[j];
    fmin=best_E;

    free(np);
  
    return 0;
  }
  //@}
      
  /// \name Iteration control
  //@{
  /// Determine how to change the minimization for the next iteration
  virtual int next(size_t nv, vec_t &x_old, double min_old, vec_t &x_new, 
		   double min_new, double &T, size_t n_moves,
		   bool &finished) {
	
    if (T/2.0<this->tolx) {
      finished=true;
      return gsl_success;
    }
    if (n_moves==0) {
      for(size_t i=0;i<nv;i++) {
	if (step_sizes[i]>this->tolx*1.0e2) {
	  step_sizes[i]/=1.5;
	}
      }
    }
    T/=1.5;
    return gsl_success;
  }

  /// Setup initial temperature and stepsize
  virtual int start(size_t nv, double &T) {
    T=1.0;
    nstep=nv;
    step_sizes=new double[nv];
    for(size_t i=0;i<nv;i++) {
      step_sizes[i]=10.0;
    }
	
    return gsl_success;
  }
  //@}

  /// \name Parameters
  //@{
  /// Boltzmann factor (default 1.0).
  double boltz;

  /// Number of iterations
  int ntrial;

  /// Output control
  int verbose;
  
  /// The independent variable tolerance
  double tolx;
  //@}

  /// The default random number generator
  rng_t def_rng;

  /// Return string denoting type ("anneal_mt")
  virtual const char *type() { return "anneal_mt"; }

  /** 
      \brief Print out iteration information.
      
      Depending on the value of the variable verbose, this prints out
      the iteration information. If verbose=0, then no information is
      printed, while if verbose>1, then after each iteration, the
      present values of x and y are output to std::cout along with the
      iteration number. If verbose>=2 then each iteration waits for a
      character.  
  */
  virtual int print_iter(size_t nv, vec_t &xx, double y, int iter,
			 double tptr, std::string comment) 
  {
    if (this->verbose<=0) return 0;
    
    size_t i;
    char ch;
    
    (*this->outs) << comment << " Iteration: " << iter << std::endl;
    text_out_file outsf(this->outs,79);
    outsf.word_out("x:");
    for(i=0;i<nv;i++) outsf.double_out(xx[i]);
    outsf.end_line();
    (*this->outs) << "y: " << y << " Tptr: " << tptr << std::endl;
    if (this->verbose>1) {
      (*this->outs) << "Press a key and type enter to continue. ";
      (*this->ins) >> ch;
    }
    
    return 0;
  }
  
#ifndef DOXYGEN_INTERNAL
      
  protected:

  /// The function wrapper executed by thread with index \c ip
  void func_wrapper(size_t ip) {
    step(new_x[ip],nvar);
    (*f)(nvar,new_x[ip],new_E[ip],(*params)[ip]);
  }
  
  /// Stream for verbose output
  std::ostream *outs;
  
  /// Stream for verbose input
  std::istream *ins;

  /// The number of threads to run
  size_t nproc;

  /// The user-specified parameters
  param_vec_t *params;
  
  /// The number of variables over which we minimize
  size_t nvar;

  /// The function to minimize
  func_t *f;
  
  /// \name Storage for present, next, and best vectors
  //@{
  alloc_vec_t x, *new_x, best_x, new_E, old_x;
  //@}

  /// Allocation object
  alloc_t ao;

  /// Number of step sizes
  size_t nstep;
      
  /// Step sizes
  double *step_sizes;

  /** \brief Allocate memory for a minimizer over \c n dimensions
      with stepsize \c step
  */
  virtual int allocate(size_t nv, size_t np) {
    nvar=nv;
    nproc=np;
    ao.allocate(x,nvar);
    new_x=new alloc_vec_t[np];
    ao.allocate(old_x,nvar);
    for(size_t i=0;i<np;i++) {
      ao.allocate(new_x[i],nvar);
    }
    ao.allocate(new_E,np);
    ao.allocate(best_x,nvar);
    return 0;
  }

  /// Free allocated memory
  virtual int free(size_t np) {
    ao.free(x);
    ao.free(old_x);
    ao.free(new_E);
    ao.free(best_x);
    for(size_t i=0;i<np;i++) {
      ao.free(new_x[i]);
    }
    delete[] new_x;
    return 0;
  }

  /// Make a step to a new attempted minimum
  virtual int step(vec_t &sx, int nv) {
    for(int i=0;i<nv;i++) {
      double u=this->def_rng.random();
      sx[i]=(2.0*u-1.0)*step_sizes[i%nstep]+sx[i];
    }
    return 0;
  }
  
#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
