/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/cold_nstar.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;
using namespace o2scl_fm;

cold_nstar::cold_nstar() {
  def_n.init(mass_neutron,2.0);
  def_p.init(mass_proton,2.0);
  def_n.non_interacting=false;
  def_p.non_interacting=false;
  np=&def_n;
  pp=&def_p;

  e.init(mass_electron,2.0);
  mu.init(mass_muon,2.0);

  eost.inc_maxlines(1000);
  
  def_tov.verbose=1;
  def_tov_eos.verbose=1;

  def_tov.set_units("fm^-4","fm^-4","fm^-3");
  def_tov_eos.set_units("fm^-4","fm^-4","fm^-3");
  def_tov_eos.set_low_density_eos(true,"");
  tp=&def_tov;

  acausal=0.0;
  min_bad=0.0;
  allow_urca=0.0;
  deny_urca=0.0;

  nb_start=0.05;
  nb_end=2.0;
  dnb=0.01;

  include_muons=false;
  eos_set=false;

  rp=&def_root;

  solver_tol=1.0e-4;
}

int cold_nstar::solve_fun(double x, double &y, int &vp) {
  
  np->n=x;
  
  pp->n=barn-np->n;
  hep->calc_e(*np,*pp,hb);

  e.mu=np->mu-pp->mu;
  e.calc_mu_zerot();
  y=pp->n-e.n;
  
  if (include_muons) {
    mu.mu=e.mu;
    mu.calc_mu_zerot();
    y-=mu.n;
  }
  
  return 0;
}

int cold_nstar::calc_eos(double np_0) {
  int ret;

  if (eos_set==false) {
    O2SCL_ERR("EOS not set in calc_eos().",gsl_efailed);
    return gsl_efailed;
  }
  
  eost.clear_table();
  eost.line_of_names("ed pr nb mun mup mue nn np ne kfn kfp kfe");
  if (include_muons) {
    eost.new_column("mumu");
    eost.new_column("nmu");
    eost.new_column("kfmu");
  }
  
  double x;
  if (fabs(np_0)<1.0e-12) x=nb_start/3.0;
  else x=np_0;
  double oldpr=0.0;
  min_bad=0.0;
  
  funct_mfptr<cold_nstar,int> sf(this,&cold_nstar::solve_fun);
  
  bool success=true;
  for(barn=nb_start;barn<=nb_end+dnb/10.0;barn+=dnb) {

    int vpx=0;
    ret=rp->solve(x,vpx,sf);
    double y;
    solve_fun(x,y,vpx);
    
    if (ret!=0 || fabs(y)>solver_tol) {
      success=false;
    }

    if (include_muons) {

      h=hb+e+mu;
      
      double line[15]={h.ed,h.pr,barn,np->mu,pp->mu,e.mu,np->n,pp->n,e.n,
		       np->kf,pp->kf,e.kf,mu.mu,mu.n,mu.kf};
      eost.line_of_data(15,line);

    } else {

      h=hb+e;

      double line[12]={h.ed,h.pr,barn,np->mu,pp->mu,e.mu,np->n,pp->n,e.n,
		       np->kf,pp->kf,e.kf};
      eost.line_of_data(12,line);

    }
    
    if (barn>nb_start && min_bad<=0.0 && h.pr<oldpr) min_bad=barn;
    oldpr=h.pr;
  }

  // Calculate the squared speed of sound. If the EOS becomes acausal,
  // calculate the density and pressure at which this happens

  eost.deriv("ed","pr","cs2");
  acausal=0.0;
  acausal_pr=0.0;
  acausal_ed=0.0;
  for(size_t i=0;i<eost.get_nlines()-1;i++) {
    if (eost.get("cs2",i)>=0.99) {
      acausal=eost.interp("cs2",1.0,"nb");
      acausal_pr=eost.interp("cs2",1.0,"pr");
      acausal_ed=eost.interp("cs2",1.0,"ed");
      i=eost.get_nlines();
    }
  }

  // Naively fix the equation of state so that it is not acausal

  if (false) {
    for(size_t i=0;i<eost.get_nlines()-1;i++) {
      if (eost.get("ed",i)>acausal_ed) {
	eost.set("pr",i,eost.get("ed",i)+acausal_pr-acausal_ed);
      }
    }
  }

  // Calculate the adiabatic index and the Urca threshold

  eost.new_column("logp");
  eost.new_column("loge");
  eost.new_column("s");
  eost.new_column("urca");
  for(size_t i=0;i<eost.get_nlines();i++) {
    eost.set("logp",i,log(eost.get("pr",i)));
    eost.set("loge",i,log(eost.get("ed",i)));
    double stmp=(eost.get("kfn",i)+eost.get("kfp",i)+eost.get("kfe",i))/2.0;
    eost.set("s",i,stmp);
    double utmp=(stmp*(stmp-eost.get("kfn",i))*(stmp-eost.get("kfp",i))*
		 (stmp-eost.get("kfe",i)));
    eost.set("urca",i,utmp);
  }
  eost.deriv("loge","logp","ad_index");
  
  deny_urca=0.0, allow_urca=0.0;
  for(size_t i=0;i<eost.get_nlines()-1;i++) {
    if (deny_urca<0.001 && eost.get("urca",i)>0.0 && 
	eost.get("urca",i+1)<0.0) {
      deny_urca=eost.get("nb",i);
    } else if (allow_urca<0.001 && eost.get("urca",i)<0.0 &&
	       eost.get("urca",i+1)>0.0) {
      /// Use linear interpolation to improve the urca estimate
      allow_urca=eost.get("nb",i)+
	(-eost.get("urca",i))*(eost.get("nb",i+1)-eost.get("nb",i))/
	(eost.get("urca",i+1)-eost.get("urca",i));
      
    }
  }
  
  if (success==false) {
    O2SCL_ERR_RET("Solving for EOS failed in calc_eos().",1);
  }

  return 0;
}

double cold_nstar::calc_urca(double np_0) {
  int ret;
  double old_urca=0.0, urca;
  
  double x;
  if (fabs(np_0)<1.0e-12) x=nb_start/3.0;
  else x=np_0;

  funct_mfptr<cold_nstar,int > sf(this,&cold_nstar::solve_fun);
  
  bool success=true;
  for(barn=nb_start;barn<=nb_end+dnb/10.0;barn+=dnb) {
    
    int vpx=0;
    ret=rp->solve(x,vpx,sf);
    double y;
    solve_fun(x,y,vpx);
    
    if (ret!=0 || fabs(y)>1.0e-4) {
      success=false;
    }
    
    double s=(np->kf+pp->kf+e.kf)/2.0;
    urca=s*(s-np->kf)*(s-pp->kf)*(s-e.kf);
    if (barn>nb_start && urca>0.0 && old_urca<0.0) {
      if (success==false) {
	O2SCL_ERR("Solution failed in calc_urca().",gsl_efailed);
      }
      return barn-dnb*urca/(urca-old_urca);
    }
    
    old_urca=urca;
  }

  if (success==false) {
    O2SCL_ERR("Solution failed in calc_urca().",gsl_efailed);
  }
  
  return 0.0;
}

int cold_nstar::calc_nstar() {
  if (def_tov_eos.read_table(eost)!=0) {
    O2SCL_ERR("Reading EOS failed in calc_nstar()",gsl_efailed);
    return gsl_efailed;
  }
  
  tp->set_units("fm^-4","fm^-4","fm^-3");
  tp->set_eos(def_tov_eos);
  
  if (tp->mvsr()!=0) {
    O2SCL_ERR("Calculating M vs. R failed in calc_nstar().",gsl_efailed);
    return gsl_efailed;
  }
  
  return 0;
}

