/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/tov_solve.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

const double xmev_kg=1.782661731e-30; 
const double mev_per_fm3_msun_km3=xmev_kg/gsl_mks::solar_mass*1.0e54; 

tov_solve::tov_solve() {
  eos_set=false;
  te=0;

  presmin=-1.0e2;
  
  verbose=1;

  kmax=40000;
  hmin=1.0e-4;
  hmax=0.05;
  hstart=4.0e-3;
  generel=true;
  maxoutsize=400;
  maxradius=60.0;

  efactor=1.0;
  pfactor=1.0;
  nfactor=1.0;

  prbegin=7.0e-7;
  prend=8.0e-3;
  princ=1.1;
  prguess=5.2e-5;
  logmode=true;

  def_solver.tolx=1.0e-12;
  def_solver.tolf=1.0e-12;
  def_solver.verbose=0;
  def_solver.ntrial=100;

  def_solver.tolf=1.0e-7;
  
  mroot_ptr=&def_solver;
  min_ptr=&def_min;
  as_ptr=&def_stepper;
  
  out_table.add_constant("schwarz",(gsl_cgs::schwarzchild_radius/1.0e5));
  out_table.add_constant("Msun",gsl_mks::solar_mass);
  out_table.add_constant("pi",pi);
  out_table.add_constant("mproton",gsl_mks::mass_proton);
  
  compute_ang_vel=false;
  
  schwarz_km=gsl_cgs::schwarzchild_radius/1.0e5;

  calcgpot=false;
}

tov_solve::~tov_solve() {
  // Free storage
  rkx.free();
  for(size_t i=0;i<6;i++) {
    rky[i].free();
    rkdydx[i].free();
  }
}

int tov_solve::set_kmax(int s_maxoutsize, int s_kmax) {
  kmax=s_kmax;
  maxoutsize=s_maxoutsize;
  return gsl_success;
}

int tov_solve::set_units(double s_efactor, double s_pfactor, 
			 double s_nfactor) {
  efactor=s_efactor;
  pfactor=s_pfactor;
  nfactor=s_nfactor;
  return gsl_success;
}

int tov_solve::set_units(std::string leunits, 
			 std::string lpunits, std::string lnunits) {
  eunits=leunits;
  punits=lpunits;
  nunits=lnunits;

  if (eunits=="g/cm^3") {
    efactor=1.0e12/gsl_mks::solar_mass;
  } else if (eunits=="erg/cm^3") {
    efactor=1.0e2*xmev_kg/gsl_mks::solar_mass/ev_mks;
  } else if (eunits=="MeV/fm^3") {
    efactor=mev_per_fm3_msun_km3;
  } else if (eunits=="Msun/km^3") {
    efactor=1.0;
  } else if (eunits=="fm^-4") {
    efactor=hc_mev_fm*mev_per_fm3_msun_km3;
  } else if (eunits=="") {
    efactor=1.0;
  } else {
    if (verbose>0) {
      cout << "Known energy and pressure units are:" << endl;
      cout << "'g/cm^3'" << endl;
      cout << "'erg/cm^3'" << endl;
      cout << "'MeV/fm^3'" << endl;
      cout << "'fm^-4'" << endl;
      cout << "'Msun/km^3' (solar masses/km^3)" << endl;
      cout << "eunits='" << eunits << "' was not understood." << endl;
    }
    O2SCL_ERR_RET("Incorrect energy density units in tov_solve::set_units().",
		gsl_efailed);
  }
  
  if (punits=="g/cm^3") {
    pfactor=1.0e12/gsl_mks::solar_mass;
  } else if (punits=="erg/cm^3") {
    pfactor=1.0e2*xmev_kg/gsl_mks::solar_mass/ev_mks;
  } else if (punits=="MeV/fm^3") {
    pfactor=mev_per_fm3_msun_km3;
  } else if (punits=="Msun/km^3") {
    pfactor=1.0;
  } else if (punits=="fm^-4") {
    pfactor=hc_mev_fm*mev_per_fm3_msun_km3;
  } else if (punits=="") {
    pfactor=1.0;
  } else {
    if (verbose>0) {
      cout << "Known energy and pressure units are:" << endl;
      cout << "'g/cm^3'" << endl;
      cout << "'erg/cm^3'" << endl;
      cout << "'MeV/fm^3'" << endl;
      cout << "'fm^-4'" << endl;
      cout << "'Msun/km^3' (solar masses/km^3)" << endl;
      cout << "eunits='" << punits << "' was not understood." << endl;
      cout << "Exiting." << endl;
    }
    O2SCL_ERR_RET("Incorrect pressure units in tov_solve::set_units().",
		gsl_efailed);
  }

  if (nunits=="cm^-3") {
    nfactor=1.0e-39;
  } else if (nunits=="m^-3") {
    nfactor=1.0e-42;
  } else if (nunits=="fm^-3") {
    nfactor=1.0;
  } else if (nunits=="") {
    nfactor=1.0;
  } else {
    if (verbose>0) {
      cout << "Known baryon density units are:" << endl;
      cout << "'m^-3' (baryons/m^3)" << endl;
      cout << "'cm^-3' (baryons/cm^3)" << endl;
      cout << "'fm^-3' (baryons/fm^3)" << endl;
      cout << "nunits='" << nunits << "' was not understood." << endl;
      cout << "Exiting." << endl;
    }
    O2SCL_ERR_RET("Incorrect baryon density units in tov_solve::set_units().",
		gsl_efailed);
  }

  out_table.add_constant("efactor",efactor);
  out_table.add_constant("pfactor",pfactor);
  out_table.add_constant("nfactor",nfactor);

  return gsl_success;
}
 
int tov_solve::derivs(double r, size_t nv, const ovector_base &y, 
		      ovector_base &dydx, int &pa) {
  double ed, term1, term2, term3, pres, nb;
  size_t ix;

  if (y[1]<=presmin) {
    return gsl_efailed;
  }

  if (r==0.0) {
    dydx[0]=0.0;
    dydx[1]=0.0;
    ix=2;
    if (calcgpot) {
      dydx[ix]=0.0;
      ix++;
    }
    if (te->baryon_column) {
      dydx[ix]=0.0;
      ix++;
    }
    return gsl_success;
  }

  pres=exp(y[1]);

  if (!finite(pres)) {
    return gsl_efailed;
  }

  double pnew=pres/pfactor;
  te->get_eden(pnew,ed,nb);
  ed*=efactor;
  nb*=nfactor;
  
  if (isinf(ed) || isnan(ed)) {
    if (verbose>0) {
      cout << "Energy density from tov_eos object not finite " 
	   << "for derivative evaluation." << endl;
    }
    O2SCL_ERR2_RET("Energy density from tov_eos object not finite ",
		   "in tov_solve::derivs().",gsl_efailed);
  }
  
  dydx[0]=4.0*pi*r*r*ed;
  if (generel) {
    term1=ed+pres;
    if (y[0]<0.0) term2=4.0*pi*pow(r,3.0)*pres;
    else term2=y[0]+4.0*pi*pow(r,3.0)*pres;
    term3=r-schwarz_km*y[0];
  } else {
    term1=ed;
    if (y[0]<0.0) term2=0.0;
    else term2=y[0];
    term3=r;
  }
  dydx[1]=-schwarz_km/2.0/r*term1*term2/term3/pres;

  ix=2;
  if (calcgpot) {
    dydx[ix]=-1.0/term1*dydx[1]*pres;
    ix++;
  }
  if (te->baryon_column) {
    if (y[0]<0.0) {
      dydx[ix]=4.0*pi*r*r*nb*gsl_mks::mass_proton/gsl_mks::solar_mass*1.0e54;
    } else {
      dydx[ix]=4.0*pi*r*r*nb*gsl_mks::mass_proton/gsl_mks::solar_mass*1.0e54/
	sqrt(1.0-schwarz_km*y[0]/r);
    }
    ix++;
  }

  return gsl_success;
}

int tov_solve::derivs_ang_vel(double r, size_t nv, const ovector_base &y, 
			      ovector_base &dydx, int &pa) {
  double rjw=y[0];
  double omega=y[1];
  
  double enc_mass=smi.interp(r,rkx.size(),rkx,rky[0]);
  double phi=smi.interp(r,rkx.size(),rkx,rky[2]);
  
  double j, djdr;

  if (r<=0.0) {

    j=0.0;
    djdr=0.0;
    dydx[0]=0.0;
    dydx[1]=0.0;

  } else {

    double pr=exp(smi.interp(r,rkx.size(),rkx,rky[1]));
    if (!finite(pr)) {
      O2SCL_ERR2_RET("Pressure not finite in ",
		     "tov_solve::derivs_ang_vel().",gsl_esanity);
    }
    double ed, nb;
    te->get_eden(pr/pfactor,ed,nb);
    ed*=efactor;
    double term2=enc_mass+4.0*pi*pow(r,3.0)*pr;
    double term3=r-schwarz_km*enc_mass;

    // This is actually dpdr, not d(log P)/dr, which is what
    // derivs() computes.
    double dpdr=-schwarz_km/2.0/r*(ed+pr)*term2/term3;
    
    double dmdr=4.0*pi*r*r*ed;
    double dphidr=-1.0/(ed+pr)*dpdr;
    if (!finite(dphidr)) {
      O2SCL_ERR2_RET("Variable 'dphidr' not finite in ",
		     "tov_solve::derivs_ang_vel().",gsl_esanity);
    }
    
    j=(1.0-schwarz_km*enc_mass/r)*exp(-phi);
    djdr=(-schwarz_km*dmdr/r+schwarz_km*enc_mass/r/r)*exp(-phi)+
      (1.0-schwarz_km*enc_mass/r)*exp(-phi)*dphidr;
    dydx[0]=-4.0*r*r*r*djdr*omega;
    dydx[1]=rjw/pow(r,4.0)/j;
  
  }
  
  if (!finite(dydx[0]) || !finite(dydx[1])) {
    O2SCL_ERR_RET("Not finite in tov_solve::derivs_ang_vel().",
		  gsl_efailed);
  }

  return 0;
}

int tov_solve::make_unique_name(string &col, std::vector<string> &cnames) {
  bool done;
  do {
    done=true;
    for(size_t i=0;i<cnames.size();i++) {
      if (col==cnames[i]) {
	done=false;
	i=cnames.size();
      }
    }
    if (done==false) {
      col+='_';
    }
  } while (done==false);
  return 0;
}

int tov_solve::profile_out(double xx) {
  int i,j,k;

  // ---------------------------------------------------------------
  // Clear previously stored data.

  out_table.clear_table();

  // ---------------------------------------------------------------
  // Add internal column names

  vector<string> inames;
  string line2;
  inames.push_back("gm");
  inames.push_back("r");
  if (calcgpot) {
    inames.push_back("gp");
  }
  if (te->baryon_column) {
    inames.push_back("bm");
  }
  inames.push_back("pr");
  inames.push_back("ed");
  inames.push_back("nb");
  inames.push_back("sg");
  inames.push_back("rs");
  inames.push_back("dmdr");
  inames.push_back("dlogpdr");
  if (calcgpot) {
    inames.push_back("dgpdr");
  }
  if (te->baryon_column) {
    inames.push_back("dbmdr");
  }
  for(size_t il=0;il<inames.size();il++) {
    line2+=inames[il];
    line2+=' ';
  }

  // ---------------------------------------------------------------
  // Add external column names

  size_t naux;
  vector<string> enames;
  te->get_names(naux,enames);
  for(size_t il=0;il<enames.size();il++) {
    make_unique_name(enames[il],inames);
  }

  for(size_t il=0;il<enames.size();il++) {
    line2+=enames[il];
    line2+=' ';
  }
  out_table.line_of_names(line2);

  // ---------------------------------------------------------------
  // Create output file of profile for star

  int klast=rkx.size()-1;

  if (klast+1<maxoutsize) maxoutsize=klast+1;
  for(i=0;i<maxoutsize;i++) {
      
    if (klast+1>maxoutsize && i!=maxoutsize-1) {
      j=i*(klast)/(maxoutsize-1);
    } else if (klast+1>maxoutsize) {
      j=klast;
    } else {
      j=i;
    }

    std::vector<double> line;
    size_t ix;

    // Output enclosed mass and radius
    line.push_back(rky[0][j]);
    line.push_back(rkx[j]);

    ix=2;
    // Gravitational potential
    if (calcgpot) {
      line.push_back(rky[ix][j]);
      ix++;
    }
    // Enclosed baryon mass
    if (te->baryon_column) {
      line.push_back(rky[ix][j]);
      ix++;
    }

    // Energy density, pressure, and baryon density
    if (rky[1][j]>presmin) {
      double ed, nb;
      te->get_eden(exp(rky[1][j])/pfactor,ed,nb);
      line.push_back(exp(rky[1][j])/pfactor);
      line.push_back(ed);
      line.push_back(nb);
    } else {
      line.push_back(0.0);
      line.push_back(0.0);
      line.push_back(0.0);
    }

    // surface gravity and redshift
    if (rkx[j]!=0.0) {
      line.push_back(schwarz_km/2.0*rky[0][j]/rkx[j]/rkx[j]/
		     sqrt(1.0-schwarz_km*rky[0][j]/rkx[j]));
      line.push_back(1.0/sqrt(1.0-rky[0][j]*schwarz_km/rkx[j])-1.0);
    } else {
      line.push_back(0.0);
      line.push_back(0.0);
    }

    // derivatives
    line.push_back(0.0);
    line.push_back(0.0);
    if (calcgpot) line.push_back(0.0);
    if (te->baryon_column) line.push_back(0.0);
    
    // Add extra EOS quantities to vector
    ovector auxp;
    if (naux>0) auxp.allocate(naux);
    te->get_aux(exp(rky[1][j])/pfactor,naux,auxp);
    for(size_t ik=0;ik<naux;ik++) line.push_back(auxp[ik]);
    
    // Check for not finite values
    for(size_t ik=0;ik<line.size();ik++) {
      if (!finite(line[ik])) {
	O2SCL_ERR_RET("Non-finite values in tov_solve::profile_out().",
		      gsl_efailed);
      }
    }

    // Copy vector to table
    int ret=out_table.line_of_data(line.size(),&(line[0]));
    if (line.size()!=out_table.get_ncolumns()) {
      O2SCL_ERR2_RET("Sanity check failed in ",
		     "tov_solve::profile_out().",gsl_esanity);
    }
      
  }

  return gsl_success;
}

int tov_solve::solution_check() {

  out_table.new_column("P_int");
  out_table.new_column("e_int");
  out_table.new_column("nb_int");
  out_table.new_column("dgmdr");
  out_table.new_column("dPdr");
  out_table.new_column("dgpdr");
  out_table.new_column("dbmdr");

  for (int i=0;i<((int)out_table.get_nlines());i++) {
    out_table.set("P_int",i,out_table.get("P",i)/pfactor);
    out_table.set("e_int",i,out_table.get("e",i)/efactor);
    out_table.set("nb_int",i,out_table.get("nb",i)/nfactor);
  }

  std::string form;
  form="dgmdr=4*pi*r*r*e_int ";
  if (generel) {
    form+="term1=e_int+P_int ";
    form+="term2=gm+4*pi*r^3*P_int ";
    form+="term3=r-schwarz*gm ";
  } else {
    form+="term1=e_int ";
    form+="term2=gm ";
    form+="term3=r ";
  }
  form+="dPdr=-schwarz/2/r*term1*term2/term3/P_int ";
  if (calcgpot) {
    form+="dgpdr=-1/term1*dPdr2*P_int ";
    if (te->baryon_column) {
      form+="dbmdr=4*pi*r^2*nb_int*mproton*Msun/sqrt(1-schwarz*P_int/r) ";
    }
  }
  //out_table.make_cols(form);
  
  out_table.deriv("r","P_int","dPdr2");
  out_table.deriv("r","gm","dgmdr2");
  out_table.deriv("r","bm","dbmdr2");
  out_table.deriv("r","gp","dgpdr2");

  return 0;
}

double tov_solve::maxfun(double maxx, int &pa) {
  ovector x(1), y(1);
  double retval;
  x[0]=maxx;

  int ret=starfun(1,x,y,pa);

  if (maxx<0.0) retval=1000.0*fabs(maxx);
  else if (ret!=0) retval=1000.0;
  else retval=-mass;

  return retval;
}

int tov_solve::starfun(size_t ndvar, const ovector_base &ndx, 
		       ovector_base &ndy, int &pa) {

  int i, k, nvar=2, test;
  size_t ix;
  bool done;
  double lastgpot=0.0;
  double x, h, hnext;
  
  if (calcgpot) nvar++;
  if (te->baryon_column) nvar++;

  ovector v(nvar), dvdx(nvar);

  if (ndx[0]<0.0) {
    O2SCL_ERR_RET("Initial pressure less than zero in tov_solve::starfun().",
		gsl_efailed);
  }

  // ---------------------------------------------------------------
  // Set up inital point

  v[0]=0.0;
  v[1]=log(ndx[0]);
  ix=2;
  if (calcgpot) {
    v[ix]=0.0;
    ix++;
  }
  if (te->baryon_column) {
    v[ix]=0.0;
  }
  
  x=0.0;
  h=hstart;
  done=false;

  // ---------------------------------------------------------------
  // Free previous storage
  
  rkx.free();
  for(i=0;i<6;i++) {
    rky[i].free();
    rkdydx[i].free();
  }

  // ---------------------------------------------------------------
  // Copy initial point to storage

  rkx.push_back(0.0);
  for (i=0;i<nvar;i++) {
    rky[i].push_back(v[i]);
    rkdydx[i].push_back(0.0);
  }

  // ---------------------------------------------------------------
  // Main loop
  
  ode_funct_mfptr<tov_solve,int > ofm(this,&tov_solve::derivs);
  
  double outrad=0.0;
  if (verbose>=2) {
    cout << "Central pressure: " << ndx[0] << endl;
    cout << "    k Radius       Enc. mass    X=Log(P)"
	 << endl;
  }

  for (k=0;done==false && k<kmax;k++) {

    if (done==false) {
      
      // ---------------------------------------------------------------
      // Fix step size if too large or too small
      
      if (h>hmax) h=hmax;
      if (h<hmin) h=hmin;
      if (x+h>maxradius) h=maxradius-x;

      // ---------------------------------------------------------------
      // Take an adaptive step 
      
      ovector vp(nvar), ve(nvar);
      test=as_ptr->astep(x,h,x+h,nvar,v,vp,ve,pa,ofm);
      if (test!=0) done=true;
      
      if ((verbose>=3 && k%20==0) || (verbose>1 && x>outrad)) {
	cout.width(5);
	cout << k << " ";
	cout << x << " " << v[0] << " " << v[1] << " ";
	{
	  // FIXME: What is this doing here? Document this as it
	  // doesn't look like this calculation is being used
	  double et, nt;
	  te->get_eden(exp(v[1])/pfactor,et,nt);
	}
	cout << endl;
	outrad+=1.0;
	
      }

      
      // ---------------------------------------------------------------
      // If we're at the right edge already, store the point and exit
      // from the loop.  This probably isn't necessary.
      
      if (x>=maxradius) {
	if (verbose>0) {
	  cout << "Warning: Reached largest radius in tov_solve::starfun()." 
	       << endl;
	}
	done=true;
      }
      
      // ---------------------------------------------------------------
      // Copy point to storage
      
      if (done==false) {
	rkx.push_back(x);
	for (i=0;i<nvar;i++) {
	  rky[i].push_back(v[i]);
	}
	derivs(x,nvar,v,dvdx,pa);
	for (i=0;i<nvar;i++) {
	  rkdydx[i].push_back(dvdx[i]);
	}
      }
    }
  }
  
  // --------------------------------------------------------------
  // Stop if we've overrun the limit 

  if (kmax>0 && k==kmax) {
    if (verbose>0) {
      cout << "Ran out of memory! (kmax=" <<  kmax << ") ";
      cout << "Reached a radius of " <<  rkx[kmax-1] 
	   << " km " << endl;
    }
    O2SCL_ERR_RET("Ran out of memory in tov_solve::starfun().",gsl_efailed);
  }

  // --------------------------------------------------------------
  // Final calculations of mass and radius. This simple linear
  // interpolation to zero pressure could be improved, possibly by
  // using an analytical solution
  
  test=derivs(x,nvar,v,dvdx,pa);

  if (dvdx[1]==0.0) rad=x;
  else rad=-1.0/dvdx[1]+x;
  mass=v[0]-dvdx[0]*(x-rad);
  ix=2;
  if (calcgpot) {
    lastgpot=v[ix]-dvdx[ix]*(x-rad);
    ix++;
  }
  if (te->baryon_column) {
    bmass=v[ix]-dvdx[ix]*(x-rad);
  }

  ndy[0]=mass-tmass;
  
  // --------------------------------------------------------------
  // Store the last point for radius, mass, and pressure
  
  rkx.push_back(rad);
  rky[0].push_back(mass);
  rky[1].push_back(presmin);

  // --------------------------------------------------------------
  // Now that we have determined the mass and radius, we can add a
  // constant to the grav. pot to satisfy the boundary conditions

  ix=2;
  if (calcgpot) {
    rky[ix].push_back(lastgpot);
    for(k=0;k<((int)rky[ix].size());k++) {
      rky[ix][k]+=0.5*log(1-schwarz_km*mass/rad)-lastgpot;
    }
    gpot=rky[ix][rky[ix].size()-1];
    ix++;
  }
  
  // --------------------------------------------------------------
  // Store the last point for baryonic mass

  if (te->baryon_column) {
    rky[ix].push_back(bmass);
  }

  // --------------------------------------------------------------
  
  for(int ik=0;ik<nvar;ik++) {
    rkdydx[ik].push_back(0.0);
  }

  // --------------------------------------------------------------
  // Output if verbose > 0

  if (verbose>0) {
    cout.precision(4);
    cout << "Central P: " << ndx[0] << " (Msun/km^3), M: " 
	 <<  mass << " (Msun), R: " <<  rad << " (km)" << endl;
    cout.precision(6);
    if (verbose>=3) {
      cout << "Press a key and 'Enter' to continue." << endl;
      char ch;
      cin >> ch;
    }
  }

  return 0;
}

int tov_solve::mvsr() {

  ovector x(1), y(1);
  int vpx=0;
  
  if (eos_set==false) {
    O2SCL_ERR_RET
      ("EOS not specified tov_solve::mvsr().",gsl_efailed);
  }

  if (verbose>0) cout << "Mass versus radius mode." << endl;

  // ---------------------------------------------------------------
  // Clear previously stored data.
  
  out_table.clear_table();

  // ---------------------------------------------------------------
  // Add internal column names

  vector<string> inames;
  string line2;
  inames.push_back("gm");
  inames.push_back("r");
  if (calcgpot) {
    inames.push_back("gp");
  }
  if (te->baryon_column) {
    inames.push_back("bm");
  }
  inames.push_back("pr");
  inames.push_back("ed");
  inames.push_back("nb");
  inames.push_back("sg");
  inames.push_back("rs");
  inames.push_back("dmdr");
  inames.push_back("dlogpdr");
  if (calcgpot) {
    inames.push_back("dgpdr");
  }
  if (te->baryon_column) {
    inames.push_back("dbmdr");
  }
  for(size_t il=0;il<inames.size();il++) {
    line2+=inames[il];
    line2+=' ';
  }

  // ---------------------------------------------------------------
  // Add external column names

  size_t naux;
  vector<string> enames;
  te->get_names(naux,enames);
  for(size_t il=0;il<enames.size();il++) {
    make_unique_name(enames[il],inames);
  }

  for(size_t il=0;il<enames.size();il++) {
    line2+=enames[il];
    line2+=' ';
  }
  out_table.line_of_names(line2);

  // ---------------------------------------------------------------
  // Main loop

  for (x[0]=prbegin;((prend>prbegin && x[0]<=prend) ||
		     (prend<prbegin && x[0]>=prend));) {
    
    starfun(1,x,y,vpx);
      
    // --------------------------------------------------------------
    // Fill line of data for table

    std::vector<double> line;

    // output mass and radius
    line.push_back(mass);
    line.push_back(rad);

    // Gravitational potential
    if (calcgpot) line.push_back(gpot);

    // output baryon mass
    if (te->baryon_column) line.push_back(bmass);
    
    // output central pressure, energy density, and baryon density

    double ed, nb;
    te->get_eden(x[0]/pfactor,ed,nb);
    
    line.push_back(x[0]/pfactor);
    line.push_back(ed);
    line.push_back(nb);

    // output surface gravity and redshift

    if (rad!=0.0) {
      line.push_back(schwarz_km/2.0*mass/rad/rad/
		     sqrt(1.0-schwarz_km*mass/rad));
      line.push_back(1.0/sqrt(1.0-mass*schwarz_km/rad)-1.0);
    } else {
      line.push_back(0.0);
      line.push_back(0.0);
    }
    
    // output derivatives

    line.push_back(0.0);
    line.push_back(0.0);
    if (calcgpot) line.push_back(0.0);
    if (te->baryon_column) line.push_back(0.0);

    // output remaining columns

    ovector auxp;
    if (naux>0) auxp.allocate(naux);
    te->get_aux(x[0]/pfactor,naux,auxp);
    for(size_t ik=0;ik<naux;ik++) line.push_back(auxp[ik]);
    
    // --------------------------------------------------------------
    // Copy line of data to table

    out_table.line_of_data(line.size(),&(line[0]));
    if (line.size()!=out_table.get_ncolumns()) {
      O2SCL_ERR_RET("Table size problem in tov_solve::mvsr().",
		    gsl_esanity);
    }
    
    // --------------------------------------------------------------
    // Get next central pressure

    if (logmode) x[0]*=princ;
    else x[0]+=princ;

  }

  return gsl_success;
}

int tov_solve::max() {
  ovector x(1), y(1);
  int vpx=0;
  
  if (eos_set==false) {
    O2SCL_ERR_RET
      ("EOS not specified and/or memory not allocated in tov_solve::mvsr().",
       gsl_efailed);
  }

  // --------------------------------------------------------------
  // Profile of maximum mass star
    
  if (!generel) {
    if (verbose>0) {
      cout << "Maximum mass not supported for non-relativistic stars. " 
	   << "Exiting. " << endl;
    }
    O2SCL_ERR_RET("Maximum mass not supported for non-relativistic stars.",
		gsl_efailed);
  }
  if (verbose>0) cout << "Maximum mass mode." << endl;

  // --------------------------------------------------------------
  // Create a good initial guess for the minimizer using a
  // naive search over several central pressures

  double initmin=0.0, initval=1.0e3;
  int minret;
  for(x[0]=7.0e-5;x[0]<=5.0e-3;x[0]*=1.3) {
    double tmp=maxfun(x[0],vpx);
    if (tmp<initval) {
      initval=tmp;
      initmin=x[0];
    }
  }
  x[0]=initmin;

  // --------------------------------------------------------------
  // Minimizer
  
  double ddtemp;
  funct_mfptr_noerr<tov_solve,int> mm(this,&tov_solve::maxfun);
  min_ptr->min(x[0],ddtemp,vpx,mm);
  starfun(1,x,y,vpx);
  
  if (verbose>0) {
    cout << "Maximum gravitational mass is: " << mass << endl;
    if (te->baryon_column) {
      cout << "Corresponding baryon mass is: " << bmass << endl;
    }
    cout << "Corresponding radius is: " << rad << endl;
  }

  // Output stellar profile
  profile_out(x[0]);

  return gsl_success;
}

int tov_solve::fixed(double d_tmass) {
  ovector x(1), y(1);
  int vpx=0;
  
  if (eos_set==false) {
    O2SCL_ERR_RET
      ("EOS not specified and/or memory not allocated in tov_solve::mvsr().",
       gsl_efailed);
  }

  tmass=d_tmass;

  // --------------------------------------------------------------
  // Profile of given solar mass star
    
  if (verbose>0) cout << tmass << " solar mass mode." << endl;
      
  // --------------------------------------------------------------
  // Negative value given, so calculate maximum mass first,
  // and adjust target mass accordingly.
  
  if (tmass<0.0) {
    max();
    tmass+=mass;
  }
  
  /// Initial guess

  x[0]=prguess;

  mm_funct_mfptr<tov_solve,int > fmf(this,&tov_solve::starfun);
  mroot_ptr->msolve(1,x,vpx,fmf);
  
  // Calculate gravitational potential and enclosed baryon mass
  starfun(1,x,y,vpx);

  if (verbose>0) {
    cout << "Gravitational mass is: " << mass << endl;
    cout << "Radius is: " << rad << endl;
    if (te->baryon_column) {
      cout << "Baryon mass is: " << bmass << endl;
    }
  }

  // If requested, compute the angular velocity
  if (compute_ang_vel) {
    if (calcgpot) {
      ang_vel();
    } else {
      O2SCL_ERR2("Angular velocity requested w/o gravitational ",
		 "potential in tov_solve::fixed().",gsl_efailed);
    }
  }
      
  // Output stellar profile
  profile_out(x[0]);
  
  if (fabs(y[0])>1.0e-3) {
    if (verbose>0) {
      cout << "Solution failed for target mass=" << tmass 
	   << " and gave mass=" << mass << "." << endl;
      cout << "Continuing anyway." << endl;
    }
    O2SCL_ERR_RET("Failed to return correct mass in tov_solve::fixed().",
		  gsl_efailed);
  }
  
 return gsl_success;

}

int tov_solve::ang_vel() {

  double ax, h=1.0e-3;
  ovector ay(2);
  table tmpt;
  tmpt.line_of_names("x y0 y1");
  
  ax=0.0;
  ay[0]=0.0;
  // Begin at c/10
  ay[1]=0.1;
  
  size_t ilast=rkx.size()-1;
  double lastrad=rkx[ilast];
  int vp=0;
  
  cout << "Last radius: " << lastrad << endl;
  
  ode_funct_mfptr<tov_solve,int > ofm(this,&tov_solve::derivs_ang_vel);
  
  bool done=false;
  while(done==false) {
    
    // ---------------------------------------------------------------
    // Fix step size if too large or too small
    
    if (h>hmax) h=hmax;
    if (h<hmin) h=hmin;
    if (ax+h>maxradius) h=maxradius-ax;
    
    // ---------------------------------------------------------------
    // Take an adaptive step 
    
    ovector vprime(2), ve(2);
    int test=as_ptr->astep(ax,h,ax+h,2,ay,vprime,ve,vp,ofm);
    //int test=as_ptr->astep(ax,h,ax+h,2,ay,vp,ofm);
    if (test!=0) {
      cout << "Failed." << endl;
    }

    double linex[3]={ax,ay[0],ay[1]};
    tmpt.line_of_data(3,linex);
    if (3!=tmpt.get_ncolumns()) {
      O2SCL_ERR_RET("Table size problem in tov_solve::ang_vel().",
		    gsl_esanity);
    }
    
    if (ax+h>lastrad) done=true;
  }

  /// Make sure enough space is available
  rky[4].allocate(rkx.size());
  rky[5].allocate(rkx.size());
  for(size_t i=0;i<rkx.size();i++) {
    rky[4][i]=tmpt.interp("x",rkx[i],"y0");
    rky[5][i]=tmpt.interp("x",rkx[i],"y1");
  }

  double lastm=rky[0][ilast];
  double lastgp=rky[2][ilast];
  double lastj=(1.0-schwarz_km*lastm/lastrad)*exp(-lastgp);
  
  cap_omega=rky[5][ilast]-lastrad/3.0*(rky[4][ilast]/lastj/pow(lastrad,4.0));

  cout.precision(12);
  cout << "ilast: " << ilast << endl;
  cout << "lastm: " << lastm << endl;
  cout << "lastgp: " << lastgp << endl;
  cout << "lastj: " << lastj << endl;
  cout << "Omega: " << cap_omega << endl;

  return 0;
}
