/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_FIT_BASE_H
#define O2SCL_FIT_BASE_H

#include <string>
#include <o2scl/collection.h>
#include <o2scl/omatrix_tlate.h>
#include <o2scl/text_file.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Fitting function [abstract base]
  */
  template<class param_t, class vec_t=ovector_base> class fit_funct {
    public:  
    
    fit_funct() {}
    virtual ~fit_funct() {}
    
    /** \brief Using parameters in \c p, predict \c y given \c x
     */
    virtual int operator()(size_t np, vec_t &p, double x, double &y, 
			   param_t &pa)=0;
    
#ifndef DOXYGEN_INTERNAL
    
    private:
    
    fit_funct(const fit_funct &);
    fit_funct& operator=(const fit_funct&);
    
#endif

  };
  
  /** 
      \brief Function pointer fitting function
  */
  template<class param_t, class vec_t=ovector_base> class fit_funct_fptr : 
  public fit_funct<param_t,vec_t> {
    
    public:
    
    /** \brief Specify a fitting function by a function pointer
     */
    fit_funct_fptr(int (*fp)(size_t np, vec_t &p, double x, 
			     double &y, param_t &pa)) {
      fptr=fp;
    }
    
    virtual ~fit_funct_fptr() {}

    /** \brief Using parameters in \c p, predict \c y given \c x
     */
    virtual int operator()(size_t np, vec_t &p, double x, double &y, 
			   param_t &pa) {
      return fptr(np,p,x,y,pa);
    }

#ifndef DOXYGEN_INTERNAL

    protected:
  
    fit_funct_fptr() {};

    /// Storage for the user-specified function pointer
    int (*fptr)(size_t np, vec_t &p, double x, double &y, param_t &pa);

    fit_funct_fptr(const fit_funct_fptr &);
    fit_funct_fptr& operator=(const fit_funct_fptr&);

#endif

  };

  /** 
      \brief Member function pointer fitting function
  */
  template <class tclass, class param_t, class vec_t=ovector_base> 
    class fit_funct_mfptr : public fit_funct<param_t,vec_t> {
    public:
    
    /** \brief Specify the member function pointer
     */
    fit_funct_mfptr(tclass *tp, 
		    int (tclass::*fp)(size_t np, vec_t &p, double x, 
				      double &y, param_t &pa)) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~fit_funct_mfptr() {};
  
    /** \brief Using parameters in \c p, predict \c y given \c x
     */
    virtual int operator()(size_t np, vec_t &p, double x, double &y, 
			   param_t &pa) {
      return (*tptr.*fptr)(np,p,x,y,pa);
    }
  
#ifndef DOXYGEN_INTERNAL

    protected:
  
    /// Storage for the user-specified function pointer
    int (tclass::*fptr)(size_t np, vec_t &p, double x, double &y, param_t &pa);

    /// Storage for the class pointer
    tclass *tptr;
  
    private:

    fit_funct_mfptr(const fit_funct_mfptr &);
    fit_funct_mfptr& operator=(const fit_funct_mfptr&);

#endif

  };

  /** 
      \brief Const member function pointer fitting function
  */
  template <class tclass, class param_t, class vec_t=ovector_base> 
    class fit_funct_cmfptr : public fit_funct<param_t,vec_t> {
    public:
    
    /** \brief Specify the member function pointer
     */
    fit_funct_cmfptr(tclass *tp, 
		    int (tclass::*fp)(size_t np, vec_t &p, double x, 
				      double &y, param_t &pa) const) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~fit_funct_cmfptr() {};
  
    /** \brief Using parameters in \c p, predict \c y given \c x
     */
    virtual int operator()(size_t np, vec_t &p, double x, double &y, 
			   param_t &pa) {
      return (*tptr.*fptr)(np,p,x,y,pa);
    }
  
#ifndef DOXYGEN_INTERNAL

    protected:
  
    /// Storage for the user-specified function pointer
    int (tclass::*fptr)(size_t np, vec_t &p, double x, double &y, 
			param_t &pa) const;

    /// Storage for the class pointer
    tclass *tptr;
  
    private:

    fit_funct_cmfptr(const fit_funct_cmfptr &);
    fit_funct_cmfptr& operator=(const fit_funct_cmfptr&);

#endif

  };

  // ----------------------------------------------------------------
  // Array versions
  // ----------------------------------------------------------------

  /** 
      \brief Fitting function with arrays [abstract base]
  */
  template<class param_t, size_t nvar> class fit_vfunct {
  public:  

    fit_vfunct() {}
    virtual ~fit_vfunct() {}
    
    /** \brief Using parameters in \c p, predict \c y given \c x
     */
    virtual int operator()(size_t np, double p[], double x, double &y, 
			   param_t &pa)=0;

#ifndef DOXYGEN_INTERNAL

  private:

    fit_vfunct(const fit_vfunct &);
    fit_vfunct& operator=(const fit_vfunct&);

#endif
  };

  /** 
      \brief Function pointer fitting function with arrays
  */
  template<class param_t, size_t nvar> class fit_vfunct_fptr : 
  public fit_vfunct<param_t,nvar> {
    
  public:
    
    /** \brief Specify a fitting function by a function pointer
     */
    fit_vfunct_fptr(int (*fp)(size_t np, double p[], double x, 
			      double &y, param_t &pa)) {
      fptr=fp;
    }
    
    virtual ~fit_vfunct_fptr() {}

    /** \brief Using parameters in \c p, predict \c y given \c x
     */
    virtual int operator()(size_t np, double p[], double x, double &y, 
			   param_t &pa) {
      return fptr(np,p,x,y,pa);
    }

#ifndef DOXYGEN_INTERNAL

  protected:
  
    fit_vfunct_fptr() {};

    /// Storage for the user-specified function pointer
    int (*fptr)(size_t np, double p[], double x, double &y, param_t &pa);
    
  private:

    fit_vfunct_fptr(const fit_vfunct_fptr &);
    fit_vfunct_fptr& operator=(const fit_vfunct_fptr&);

#endif

  };

  /** 
      \brief Member function pointer fitting function with arrays
  */
  template <class tclass, class param_t, size_t nvar> 
    class fit_vfunct_mfptr : public fit_vfunct<param_t,nvar> {
  public:
    
    /** \brief Specify the member function pointer
     */
    fit_vfunct_mfptr(tclass *tp, 
		     int (tclass::*fp)(size_t np, double p[], double x, 
				       double &y, param_t &pa)) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~fit_vfunct_mfptr() {};
  
    /** \brief Using parameters in \c p, predict \c y given \c x
     */
    virtual int operator()(size_t np, double p[], double x, double &y, 
			   param_t &pa) {
      return (*tptr.*fptr)(np,p,x,y,pa);
    }
  
#ifndef DOXYGEN_INTERNAL

  protected:
  
    /// Storage for the user-specified function pointer
    int (tclass::*fptr)(size_t np, double p[], double x, double &y, 
			param_t &pa);

    /// Storage for the class pointer
    tclass *tptr;
  
  private:

    fit_vfunct_mfptr(const fit_vfunct_mfptr &);
    fit_vfunct_mfptr& operator=(const fit_vfunct_mfptr&);

#endif

  };

  /** 
      \brief Const member function pointer fitting function with arrays
  */
  template <class tclass, class param_t, size_t nvar> 
    class fit_vfunct_cmfptr : public fit_vfunct<param_t,nvar> {
  public:
    
    /** \brief Specify the member function pointer
     */
    fit_vfunct_cmfptr(tclass *tp, 
		     int (tclass::*fp)(size_t np, double p[], double x, 
				       double &y, param_t &pa) const) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~fit_vfunct_cmfptr() {};
  
    /** \brief Using parameters in \c p, predict \c y given \c x
     */
    virtual int operator()(size_t np, double p[], double x, double &y, 
			   param_t &pa) {
      return (*tptr.*fptr)(np,p,x,y,pa);
    }
  
#ifndef DOXYGEN_INTERNAL

  protected:
  
    /// Storage for the user-specified function pointer
    int (tclass::*fptr)(size_t np, double p[], double x, double &y, 
			param_t &pa) const;

    /// Storage for the class pointer
    tclass *tptr;
  
  private:

    fit_vfunct_cmfptr(const fit_vfunct_cmfptr &);
    fit_vfunct_cmfptr& operator=(const fit_vfunct_cmfptr&);

#endif

  };

  // ----------------------------------------------------------------
  // Fitter base
  // ----------------------------------------------------------------

  /** 
      \brief Non-linear least-squares fitting [abstract base]
  */
  template<class param_t, class func_t, 
    class vec_t=ovector_base, class mat_t=omatrix_base> class fit_base {
    public:

    fit_base() {
      verbose=0;
    }

    virtual ~fit_base() {}
    
    /** 
	\brief Print out iteration information.
	
	Depending on the value of the variable verbose, this prints out
	the iteration information. If verbose=0, then no information is
	printed, while if verbose>1, then after each iteration, the
	present values of x and y are output to std::cout along with the
	iteration number. If verbose>=2 then each iteration waits for a
	character.  
    */
    virtual int print_iter(size_t nv, vec_t &x, double y, int iter,
			   double value=0.0, double limit=0.0) {
      if (verbose<=0) return 0;
  
      size_t i;
      char ch;

      std::cout << "Iteration: " << iter << std::endl;
      text_out_file outs(&std::cout,79);
      outs.word_out("x:");
      for(i=0;i<nv;i++) outs.double_out(x[i]);
      outs.end_line();
      std::cout << "y: " << y << " Val: " << value << " Lim: " << limit 
      << std::endl;
      if (verbose>1) {
	std::cout << "Press a key and type enter to continue. ";
	std::cin >> ch;
      }

      return 0;
    }

    /** \brief Fit the data specified in (xdat,ydat) to
	the function \c fitfun with the parameters in \c par.

	The covariance matrix for the parameters is returned in \c covar
	and the value of \f$ \chi^2 \f$ is returned in \c chi2.

    */
    virtual int fit(size_t ndat, vec_t &xdat, vec_t &ydat, vec_t &yerr,
		    size_t npar, vec_t &par, mat_t &covar, double &chi2,
		    param_t &pa, func_t &fitfun)=0;
    
    /** \brief An integer describing the verbosity of the output 
     */
    int verbose;

    /// Return string denoting type ("fit_base")
    virtual const char *type() { return "fit_base"; }

    /// The number of data points
    size_t n_dat;

    /// The number of parameters
    size_t n_par;


  };

#ifndef DOXYGENP
}
#endif

#endif
