/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/*
 * Copyright (C) 1996, 1997, 1998, 1999, 2000 Gerard Jungman
 * Copyright (C) 2001, 2007 Brian Gough
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */
/** \file cblas_base.h
    \brief O2scl basic linear algebra function templates
    
    \future Convert to size_t and add float and complex versions
*/

#ifdef DOXYGENP
namespace o2scl_cblas {
#endif

  /// Matrix order, either column-major or row-major
  enum O2CBLAS_ORDER {O2cblasRowMajor=101, O2cblasColMajor=102};

  /// Transpose operations
  enum O2CBLAS_TRANSPOSE {O2cblasNoTrans=111, O2cblasTrans=112, 
			  O2cblasConjTrans=113};

  /// Upper- or lower-triangular
  enum O2CBLAS_UPLO {O2cblasUpper=121, O2cblasLower=122};

  /// Unit or generic diagonal
  enum O2CBLAS_DIAG {O2cblasNonUnit=131, O2cblasUnit=132};

  /// Left or right sided operation
  enum O2CBLAS_SIDE {O2cblasLeft=141, O2cblasRight=142};

  /// \name Standard BLAS functions
  //@{
  /// Compute \f$ y= \alpha x + y \f$
  template<class vec_t, class vec2_t> void 
    daxpy(const int N, const double alpha, const vec_t &X, 
	  vec2_t &Y) {
    
    int i;
    
    if (alpha == 0.0) {
      return;
    }
    
    const int m = N % 4;
    
    for (i = 0; i < m; i++) {
      O2SCL_IX(Y,i) += alpha * O2SCL_IX(X,i);
    }
    
    for (i = m; i + 3 < N; i += 4) {
      O2SCL_IX(Y,i) += alpha * O2SCL_IX(X,i);
      O2SCL_IX(Y,i+1) += alpha * O2SCL_IX(X,i+1);
      O2SCL_IX(Y,i+2) += alpha * O2SCL_IX(X,i+2);
      O2SCL_IX(Y,i+3) += alpha * O2SCL_IX(X,i+3);
    }
  }
  
  /// Compute \f$ r = x \cdot y \f$
  template<class vec_t, class vec2_t> double
    ddot(const int N, const vec_t &X, const vec2_t &Y) {
    double r = 0.0;
    int i;

    const int m = N % 4;
      
    for (i = 0; i < m; i++) {
      r += O2SCL_IX(X,i) * O2SCL_IX(Y,i);
    }
      
    for (i = m; i + 3 < N; i += 4) {
      r += O2SCL_IX(X,i) * O2SCL_IX(Y,i);
      r += O2SCL_IX(X,i+1) * O2SCL_IX(Y,i+1);
      r += O2SCL_IX(X,i+2) * O2SCL_IX(Y,i+2);
      r += O2SCL_IX(X,i+3) * O2SCL_IX(Y,i+3);
    }

    return r;
  }
  
  /// Compute \f$ x = \alpha x \f$
  template<class vec_t> void
    dscal(const int N, const double alpha, vec_t &X) {
    int i;
    const int m = N % 4;

    for (i = 0; i < m; i++) {
      O2SCL_IX(X,i)*=alpha;
    }
    
    for (i = m; i + 3 < N; i += 4) {
      O2SCL_IX(X,i)*=alpha;
      O2SCL_IX(X,i+1)*=alpha;
      O2SCL_IX(X,i+2)*=alpha;
      O2SCL_IX(X,i+3)*=alpha;
    }
  }
  
  /// Compute the squared norm of the vector \c X
  template<class vec_t> double dnrm2(const int N, const vec_t &X) {
    
    double scale = 0.0;
    double ssq = 1.0;
    int i;
    
    if (N <= 0) {
      return 0;
    } else if (N == 1) {
      return fabs(O2SCL_IX(X,0));
    }
    
    for (i = 0; i < N; i++) {
      const double x = O2SCL_IX(X,i);
      
      if (x != 0.0) {
	const double ax = fabs(x);
	
	if (scale < ax) {
	  ssq = 1.0 + ssq * (scale / ax) * (scale / ax);
	  scale = ax;
	} else {
	  ssq += (ax / scale) * (ax / scale);
	}
      }
      
    }
    
    return scale * sqrt(ssq);
  }

  /// Compute \f$ y = \alpha \mathrm{op}(A) x + \beta y \f$.
  template<class mat_t, class vec_t>
    int dgemv(const enum O2CBLAS_ORDER order, 
	      const enum O2CBLAS_TRANSPOSE TransA,
	      const int M, const int N, const double alpha, const mat_t &A,
	      const vec_t &X, const double beta, vec_t &Y) {
    
    int i, j;
    int lenX, lenY;
    
    const int Trans = (TransA != O2cblasConjTrans) ? TransA : O2cblasTrans;

    if (M == 0 || N == 0) {
      return 0;
    }

    if (alpha == 0.0 && beta == 1.0) {
      return 0;
    }

    if (Trans == O2cblasNoTrans) {
      lenX = N;
      lenY = M;
    } else {
      lenX = M;
      lenY = N;
    }

    /* form  y := beta*y */
    if (beta == 0.0) {
      int iy = 0;
      for (i = 0; i < lenY; i++) {
	O2SCL_IX(Y,iy) = 0.0;
	iy++;
      }
    } else if (beta != 1.0) {
      int iy = 0;
      for (i = 0; i < lenY; i++) {
	O2SCL_IX(Y,iy) *= beta;
	iy++;
      }
    }

    if (alpha == 0.0) {
      return 0;
    }

    if ((order == O2cblasRowMajor && Trans == O2cblasNoTrans)
	|| (order == O2cblasColMajor && Trans == O2cblasTrans)) {

      /* form  y := alpha*A*x + y */
      int iy = 0;
      for (i = 0; i < lenY; i++) {
	double temp = 0.0;
	int ix = 0;
	for (j = 0; j < lenX; j++) {
	  temp += O2SCL_IX(X,ix) * O2SCL_IX2(A,i,j);
	  ix++;
	}
	O2SCL_IX(Y,iy) += alpha * temp;
	iy++;
      }

    } else if ((order == O2cblasRowMajor && Trans == O2cblasTrans)
	       || (order == O2cblasColMajor && Trans == O2cblasNoTrans)) {

      /* form  y := alpha*A'*x + y */
      int ix = 0;
      for (j = 0; j < lenX; j++) {
	const double temp = alpha * O2SCL_IX(X,ix);
	if (temp != 0.0) {
	  int iy = 0;
	  for (i = 0; i < lenY; i++) {
	    O2SCL_IX(Y,iy) += temp * O2SCL_IX2(A,j,i);
	    iy++;
	  }
	}
	ix++;
      }

    } else {
      O2SCL_ERR_RET("Unrecognized operation in dgemv().",o2scl::gsl_einval);
    }
    return 0;
  }
  
  /// Compute \f$ x = \mathrm{op} (A)^{-1} x \f$
  template<class mat_t, class vec_t> 
    int dtrsv(const enum O2CBLAS_ORDER order, 
	      const enum O2CBLAS_UPLO Uplo,
	      const enum O2CBLAS_TRANSPOSE TransA, 
	      const enum O2CBLAS_DIAG Diag,
	      const int M, const int N, const mat_t &A, vec_t &X) {
    const int nonunit = (Diag == O2cblasNonUnit);
    int ix, jx;
    int i, j;
    const int Trans = (TransA != O2cblasConjTrans) ? TransA : O2cblasTrans;

    if (N == 0) {
      return 0;
    }

    /* form  x := inv( A )*x */

    if ((order == O2cblasRowMajor && Trans == O2cblasNoTrans && 
	 Uplo == O2cblasUpper)
	|| (order == O2cblasColMajor && Trans == O2cblasTrans && 
	    Uplo == O2cblasLower)) {

      /* backsubstitution */
      ix = (N - 1);
      if (nonunit) {
	O2SCL_IX(X,ix) = O2SCL_IX(X,ix) / O2SCL_IX2(A,N-1,N-1);
      }
      ix--;
      for (i = N - 1; i > 0 && i--;) {
	double tmp = O2SCL_IX(X,ix);
	jx = ix +1;
	for (j = i + 1; j < N; j++) {
	  const double Aij = O2SCL_IX2(A,i,j);
	  tmp -= Aij * O2SCL_IX(X,jx);
	  jx++;
	}
	if (nonunit) {
	  O2SCL_IX(X,ix) = tmp / O2SCL_IX2(A,i,i);
	} else {
	  O2SCL_IX(X,ix) = tmp;
	}
	ix--;
      }

    } else if ((order == O2cblasRowMajor && Trans == O2cblasNoTrans && 
		Uplo == O2cblasLower) || 
	       (order == O2cblasColMajor && Trans == O2cblasTrans && 
		Uplo == O2cblasUpper)) {
	
      /* forward substitution */
      ix = 0;
      if (nonunit) {
	O2SCL_IX(X,ix) = O2SCL_IX(X,ix) / O2SCL_IX2(A,0,0);
      }
      ix++;
      for (i = 1; i < N; i++) {
	double tmp = O2SCL_IX(X,ix);
	jx = 0;
	for (j = 0; j < i; j++) {
	  const double Aij = O2SCL_IX2(A,i,j);
	  tmp -= Aij * O2SCL_IX(X,jx);
	  jx++;
	}
	if (nonunit) {
	  O2SCL_IX(X,ix) = tmp / O2SCL_IX2(A,i,i);
	} else {
	  O2SCL_IX(X,ix) = tmp;
	}
	ix++;
      }

    } else if ((order == O2cblasRowMajor && Trans == O2cblasTrans && 
		Uplo == O2cblasUpper)
	       || (order == O2cblasColMajor && Trans == O2cblasNoTrans && 
		   Uplo == O2cblasLower)) {
	
      /* form  x := inv( A' )*x */
	
      /* forward substitution */
      ix = 0;
      if (nonunit) {
	O2SCL_IX(X,ix) = O2SCL_IX(X,ix) / O2SCL_IX2(A,0,0);
      }
      ix++;
      for (i = 1; i < N; i++) {
	double tmp = O2SCL_IX(X,ix);
	jx = 0;
	for (j = 0; j < i; j++) {
	  const double Aji = O2SCL_IX2(A,j,i);
	  tmp -= Aji * O2SCL_IX(X,jx);
	  jx++;
	}
	if (nonunit) {
	  O2SCL_IX(X,ix) = tmp / O2SCL_IX2(A,i,i);
	} else {
	  O2SCL_IX(X,ix) = tmp;
	}
	ix++;
      }

    } else if ((order == O2cblasRowMajor && Trans == O2cblasTrans && 
		Uplo == O2cblasLower)
	       || (order == O2cblasColMajor && Trans == O2cblasNoTrans && 
		   Uplo == O2cblasUpper)) {
	
      /* backsubstitution */
      ix = (N - 1);
      if (nonunit) {
	O2SCL_IX(X,ix) = O2SCL_IX(X,ix) / O2SCL_IX2(A,N-1,N-1);
      }
      ix--;
      for (i = N - 1; i > 0 && i--;) {
	double tmp = O2SCL_IX(X,ix);
	jx = ix +1;
	for (j = i + 1; j < N; j++) {
	  const double Aji = O2SCL_IX2(A,j,i);
	  tmp -= Aji * O2SCL_IX(X,jx);
	  jx++;
	}
	if (nonunit) {
	  O2SCL_IX(X,ix) = tmp / O2SCL_IX2(A,i,i);
	} else {
	  O2SCL_IX(X,ix) = tmp;
	}
	ix--;
      }

    } else {
      O2SCL_ERR_RET("Unrecognized operation in dtrsv().",o2scl::gsl_einval);
    }
    return 0;
  }

  //@}

  /// \name Helper BLAS functions 
  //@{
  /**
     \brief Compute \f$ x = \alpha x \f$ beginning with index \c ie and
     ending with index \c N-1

     Used in \ref householder_hv().
  */
  template<class vec_t, class vec2_t> 
    void daxpy_subvec(const int N, const double alpha, const vec_t &X,
		      vec2_t &Y, const int ie) {
    
    int i;

    if (alpha == 0.0) {
      return;
    }

    const int m = (N-ie) % 4;
    
    for (i = ie; i < ie+m; i++) {
      O2SCL_IX(Y,i) += alpha * O2SCL_IX(X,i);
    }

    for (i = ie+m; i + 3 < N; i += 4) {
      O2SCL_IX(Y,i) += alpha * O2SCL_IX(X,i);
      O2SCL_IX(Y,i+1) += alpha * O2SCL_IX(X,i+1);
      O2SCL_IX(Y,i+2) += alpha * O2SCL_IX(X,i+2);
      O2SCL_IX(Y,i+3) += alpha * O2SCL_IX(X,i+3);
    }
  }

  /**
     \brief Compute \f$ r = x \cdot y \f$ beginning with index \c ie and
     ending with index \c N-1

     Used in \ref householder_hv().
  */
  template<class vec_t, class vec2_t> 
    double ddot_subvec(const int N, const vec_t &X, const vec2_t &Y,
		       const int ie) {
    double r = 0.0;
    int i;
    
    const int m = (N-ie) % 4;

    for (i = ie; i < ie+m; i++) {
      r += O2SCL_IX(X,i) * O2SCL_IX(Y,i);
    }

    for (i = ie+m; i + 3 < N; i += 4) {
      r += O2SCL_IX(X,i) * O2SCL_IX(Y,i);
      r += O2SCL_IX(X,i+1) * O2SCL_IX(Y,i+1);
      r += O2SCL_IX(X,i+2) * O2SCL_IX(Y,i+2);
      r += O2SCL_IX(X,i+3) * O2SCL_IX(Y,i+3);
    }

    return r;
  }

  /**
     \brief Compute \f$ x = \alpha x \f$ beginning with index \c ie and
     ending with index \c N-1

     Used in \ref householder_transform().
  */
  template<class vec_t> 
    void dscal_subvec(const int N, const double alpha, vec_t &X,
		      const int ie) {
    int i;
    const int m = (N-ie) % 4;

    for (i = ie; i < ie+m; i++) {
      O2SCL_IX(X,i)*=alpha;
    }
    
    for (i = ie+m; i + 3 < N; i += 4) {
      O2SCL_IX(X,i)*=alpha;
      O2SCL_IX(X,i+1)*=alpha;
      O2SCL_IX(X,i+2)*=alpha;
      O2SCL_IX(X,i+3)*=alpha;
    }
  }
  
  /**
     \brief Compute the squared norm of the vector \c X  beginning with 
     index \c ie and ending with index \c N-1

     Used in \ref householder_transform().
  */
  template<class vec_t> 
    double dnrm2_subvec(const int N, const vec_t &X, const int ie) {
    
    double scale = 0.0;
    double ssq = 1.0;
    int i;
    
    if (N <= 0) {
      return 0;
    } else if (N == 1) {
      return fabs(O2SCL_IX(X,ie));
    }
    
    for (i = ie; i < N; i++) {
      const double x = O2SCL_IX(X,i);
      
      if (x != 0.0) {
	const double ax = fabs(x);
	
	if (scale < ax) {
	  ssq = 1.0 + ssq * (scale / ax) * (scale / ax);
	  scale = ax;
	} else {
	  ssq += (ax / scale) * (ax / scale);
	}
      }
      
    }
    
    return scale * sqrt(ssq);
  }
  
  /**
     \brief Compute the squared norm of the last \c N rows
     of a column of a matrix

     Given matrix \c M, this computes the norm of the last \c N rows
     of the column with index \c ic, beginning with the element with
     index \c ir. If the matrix \c M has \c r rows, and \c c columns,
     then the parameter \c N should be \c r-ir.

     Used in householder_transform_subcol().
  */
  template<class mat_t> 
    double dnrm2_subcol(const mat_t &M, const size_t ir, const size_t ic,
			const size_t N) {
    
    double scale = 0.0;
    double ssq = 1.0;
    size_t i;
    
    if (N <= 0) {
      return 0;
    } else if (N == 1) {
      return fabs(O2SCL_IX2(M,ir,ic));
    }
    
    size_t tot=ir+N;
    for (i = ir; i < tot; i++) {
      const double x = O2SCL_IX2(M,i,ic);
      
      if (x != 0.0) {
	const double ax = fabs(x);
	
	if (scale < ax) {
	  ssq = 1.0 + ssq * (scale / ax) * (scale / ax);
	  scale = ax;
	} else {
	  ssq += (ax / scale) * (ax / scale);
	}
      }
      
    }
    
    return scale * sqrt(ssq);
  }

  /**
     \brief Compute \f$ x = \alpha x \f$
     
     Used in householder_transform_subcol().
  */
  template<class mat_t> 
    void dscal_subcol(mat_t &A, const size_t ir, const size_t ic,
		      const size_t n, const double alpha) {

    size_t tot=ir+n;
    size_t i;
    const size_t m = (tot-ir) % 4;
    
    for (i = ir; i < m+ir; i++) {
      O2SCL_IX2(A,i,ic)*=alpha;
    }
    
    for (i = m+ir; i + 3 < tot; i += 4) {
      O2SCL_IX2(A,i,ic)*=alpha;
      O2SCL_IX2(A,i+1,ic)*=alpha;
      O2SCL_IX2(A,i+2,ic)*=alpha;
      O2SCL_IX2(A,i+3,ic)*=alpha;
    }
    
  }

  /**
     \brief Compute \f$ x = \alpha x \f$ for \ref householder_hv_sub()

     Used in householder_hv_sub().

     \future Implement explicit loop unrolling
  */
  template<class mat_t, class vec_t> 
    void daxpy_hv_sub(const int N, const double alpha, const mat_t &X,
		      vec_t &Y, const int ie) {
    
    if (alpha == 0.0) {
      return;
    }

    for(int i=ie+1;i<N;i++) {
      O2SCL_IX(Y,i)+=alpha*O2SCL_IX2(X,i,ie);
    }

    /*
      const int m = N % 4;
      
      for (i = 0; i < m; i++) {
      O2SCL_IX(X,i)*=alpha;
      }
      
      for (i = m; i + 3 < N; i += 4) {
      O2SCL_IX(X,i)*=alpha;
      O2SCL_IX(X,i+1)*=alpha;
      O2SCL_IX(X,i+2)*=alpha;
      O2SCL_IX(X,i+3)*=alpha;
      }
    */
  }

  /**
     \brief Compute \f$ r = x \cdot y \f$ for \ref householder_hv_sub()

     Used in householder_hv_sub().

     \future Implement explicit loop unrolling
  */
  template<class mat_t, class vec_t> 
    double ddot_hv_sub(const int N, const mat_t &X, const vec_t &Y,
		       const int ie) {
    double r = 0.0;
    int i;
    
    for(i=ie+1;i<N;i++) {
      r+=O2SCL_IX2(X,i,ie)*O2SCL_IX(Y,i);
    }
    return r;
  }

  //@}

  /// Compute \f$ y = \alpha \mathrm{op}(A) x + \beta y \f$.
  template<class mat_t>
    int dgemm(const enum O2CBLAS_ORDER Order, 
	      const enum O2CBLAS_TRANSPOSE TransA,
	      const enum O2CBLAS_TRANSPOSE TransB, const int M, const int N,
	      const int K, const double alpha, const mat_t &A, 
	      const mat_t &B, const double beta, mat_t &C) {
    
    int i, j, k;
    int n1, n2;
    int ldf, ldg;
    int TransF, TransG;

    if (alpha == 0.0 && beta == 1.0) {
      return 0;
    }

    /*
      This is a little more complicated than the original, which
      assigned the matrices A and B to variables *F and *G which then
      allowed some code duplication. We can't really do that here,
      since we don't have that kind of type info on A and B, so we
      just handle the two cases separately.
    */
    
    if (Order == O2cblasRowMajor) {

      n1 = M;
      n2 = N;

      /* form  y := beta*y */
      if (beta == 0.0) {
	for (i = 0; i < n1; i++) {
	  for (j = 0; j < n2; j++) {
	    O2SCL_IX2(C,i,j)=0.0;
	  }
	}
      } else if (beta != 1.0) {
	for (i = 0; i < n1; i++) {
	  for (j = 0; j < n2; j++) {
	    O2SCL_IX2(C,i,j)*=beta;
	  }
	}
      }

      if (alpha == 0.0) {
	return 0;
      }

      TransF = (TransA == O2cblasConjTrans) ? O2cblasTrans : TransA;
      TransG = (TransB == O2cblasConjTrans) ? O2cblasTrans : TransB;
      
      if (TransF == O2cblasNoTrans && TransG == O2cblasNoTrans) {

	/* form  C := alpha*A*B + C */

	for (k = 0; k < K; k++) {
	  for (i = 0; i < n1; i++) {
	    const double temp = alpha * O2SCL_IX2(A,i,k);
	    if (temp != 0.0) {
	      for (j = 0; j < n2; j++) {
		O2SCL_IX2(C,i,j) += temp * O2SCL_IX2(B,k,j);
	      }
	    }
	  }
	}

      } else if (TransF == O2cblasNoTrans && TransG == O2cblasTrans) {

	/* form  C := alpha*A*B' + C */

	for (i = 0; i < n1; i++) {
	  for (j = 0; j < n2; j++) {
	    double temp = 0.0;
	    for (k = 0; k < K; k++) {
	      temp += O2SCL_IX2(A,i,k) * O2SCL_IX2(B,j,k);
	    }
	    O2SCL_IX2(C,i,j) += alpha * temp;
	  }
	}

      } else if (TransF == O2cblasTrans && TransG == O2cblasNoTrans) {

	for (k = 0; k < K; k++) {
	  for (i = 0; i < n1; i++) {
	    const double temp = alpha * O2SCL_IX2(A,k,i);
	    if (temp != 0.0) {
	      for (j = 0; j < n2; j++) {
		O2SCL_IX2(C,i,j) += temp * O2SCL_IX2(B,k,j);
	      }
	    }
	  }
	}

      } else if (TransF == O2cblasTrans && TransG == O2cblasTrans) {
	
	for (i = 0; i < n1; i++) {
	  for (j = 0; j < n2; j++) {
	    double temp = 0.0;
	    for (k = 0; k < K; k++) {
	      temp += O2SCL_IX2(A,k,i) * O2SCL_IX2(B,j,k);
	    }
	    O2SCL_IX2(C,i,j) += alpha * temp;
	  }
	}

      } else {
	O2SCL_ERR_RET("Unrecognized operation in dgemm().",o2scl::gsl_einval);
      }

    } else {

      n1 = N;
      n2 = M;

      /* form  y := beta*y */
      if (beta == 0.0) {
	for (i = 0; i < n1; i++) {
	  for (j = 0; j < n2; j++) {
	    O2SCL_IX2(C,i,j)=0.0;
	  }
	}
      } else if (beta != 1.0) {
	for (i = 0; i < n1; i++) {
	  for (j = 0; j < n2; j++) {
	    O2SCL_IX2(C,i,j)*=beta;
	  }
	}
      }

      if (alpha == 0.0) {
	return 0;
      }

      TransF = (TransB == O2cblasConjTrans) ? O2cblasTrans : TransB;
      TransG = (TransA == O2cblasConjTrans) ? O2cblasTrans : TransA;

      if (TransF == O2cblasNoTrans && TransG == O2cblasNoTrans) {

	/* form  C := alpha*A*B + C */

	for (k = 0; k < K; k++) {
	  for (i = 0; i < n1; i++) {
	    const double temp = alpha * O2SCL_IX2(B,i,k);
	    if (temp != 0.0) {
	      for (j = 0; j < n2; j++) {
		O2SCL_IX2(C,i,j) += temp * O2SCL_IX2(A,k,j);
	      }
	    }
	  }
	}

      } else if (TransF == O2cblasNoTrans && TransG == O2cblasTrans) {

	/* form  C := alpha*A*B' + C */

	for (i = 0; i < n1; i++) {
	  for (j = 0; j < n2; j++) {
	    double temp = 0.0;
	    for (k = 0; k < K; k++) {
	      temp += O2SCL_IX2(B,i,k) * O2SCL_IX2(A,j,k);
	    }
	    O2SCL_IX2(C,i,j) += alpha * temp;
	  }
	}

      } else if (TransF == O2cblasTrans && TransG == O2cblasNoTrans) {

	for (k = 0; k < K; k++) {
	  for (i = 0; i < n1; i++) {
	    const double temp = alpha * O2SCL_IX2(B,k,i);
	    if (temp != 0.0) {
	      for (j = 0; j < n2; j++) {
		O2SCL_IX2(C,i,j) += temp * O2SCL_IX2(A,k,j);
	      }
	    }
	  }
	}

      } else if (TransF == O2cblasTrans && TransG == O2cblasTrans) {

	for (i = 0; i < n1; i++) {
	  for (j = 0; j < n2; j++) {
	    double temp = 0.0;
	    for (k = 0; k < K; k++) {
	      temp += O2SCL_IX2(B,k,i) * O2SCL_IX2(A,j,k);
	    }
	    O2SCL_IX2(C,i,j) += alpha * temp;
	  }
	}

      } else {
	O2SCL_ERR_RET("Unrecognized operation in dgemm().",o2scl::gsl_einval);
      }
    }

    return 0;
  }
  
#ifdef DOXYGENP
}
#endif

