/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.
  
  -------------------------------------------------------------------
*/
/* linalg/householder.c
 * 
 * Copyright (C) 1996, 1997, 1998, 1999, 2000, 2004, 
 * 2007 Gerard Jungman, Brian Gough
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */
/** \file householder_base.h
    \brief File for Householder transformations
*/

#include <o2scl/err_hnd.h>
#include <o2scl/cblas.h>
#include <o2scl/permutation.h>
#include <o2scl/vec_arith.h>

#ifdef DOXYGENP
namespace o2scl_linalg {
#endif
  
  /** 
      \brief Replace the vector \c v with a householder vector and a
      coefficient tau that annihilates the last \c n-1 elements of 
      \c v.
  */
  template<class vec_t> 
    double householder_transform(const size_t n, vec_t &v) {

    if (n == 1) {
      /* tau = 0 */
      return 0.0; 
    } else { 
      double alpha, beta, tau ;
    
      double xnorm = o2scl_cblas::dnrm2_subvec(n,v,1);
    
      if (xnorm == 0) {
	/* tau = 0 */
	return 0.0; 
      }
    
      alpha = O2SCL_IX(v,0);
      beta = - (alpha >= 0.0 ? +1.0 : -1.0) * hypot(alpha, xnorm) ;
      tau = (beta - alpha) / beta ;
    
      o2scl_cblas::dscal_subvec(n,1.0/(alpha-beta),v,1);
      O2SCL_IX(v,0)=beta;
    
      return tau;
    }
  }

  /** 
      \brief Compute the householder transform of a vector
      formed with the last \c n rows of a column of a matrix
      
      Used in QR_decomp().
  */
  template<class mat_t> 
    double householder_transform_subcol(mat_t &A, const size_t ir,
					const size_t ic, const size_t n) {
    
    if (n == 1) {
      /* tau = 0 */
      return 0.0; 
    } else { 
      double alpha, beta, tau ;
      
      double xnorm = o2scl_cblas::dnrm2_subcol(A,ir+1,ic,n-1);
      
      if (xnorm == 0) {
	/* tau = 0 */
	return 0.0; 
      }
    
      alpha=O2SCL_IX2(A,ir,ic);
      beta = - (alpha >= 0.0 ? +1.0 : -1.0) * hypot(alpha, xnorm) ;
      tau = (beta - alpha) / beta ;
      
      o2scl_cblas::dscal_subcol(A,ir+1,ic,n-1,1.0/(alpha-beta));
      O2SCL_IX2(A,ir,ic)=beta;
    
      return tau;
    }
  }

  /** \brief Apply a householder transformation v,tau to matrix m */
  template<class vec_t, class mat_t>
    int householder_hm(const size_t M, const size_t N, 
		       double tau, const vec_t &v, mat_t &A) {

    if (tau == 0.0) {
      return o2scl::gsl_success;
    }
    
    size_t i, j;
    
    for (j = 0; j < N; j++) {
      
      /* Compute wj = Akj vk */
      double wj = O2SCL_IX2(A,0,j);
      
      for (i = 1; i < M; i++) {
	wj += O2SCL_IX2(A,i,j)*O2SCL_IX(v,i);
      }
      
      /* Aij = Aij - tau vi wj */
      
      /* i = 0 */
      O2SCL_IX2(A,0,j)-=tau*wj;
      
      /* i = 1 .. M-1 */
      for (i = 1; i < M; i++) {
	double Aij = O2SCL_IX2(A,i,j);
	double vi = O2SCL_IX(v,i);
	O2SCL_IX2(A,i,j)=Aij - tau * vi * wj;
      }
    }
    return o2scl::gsl_success;
  }

  /** \brief Apply a householder transformation \c v, \c tau 
      to submatrix of m 

      Used in QR_decomp().
  */
  template<class mat_t>
    int householder_hm_sub(mat_t &M, const size_t ir,
			   const size_t ic, const size_t nr,
			   const size_t nc, const mat_t &M2,
			   const size_t ir2, const size_t ic2,
			   double tau) {

    if (tau == 0.0) {
      return o2scl::gsl_success;
    }
    
    size_t i, j, last=ir+nr;
    
    for (j = ic; j < ic+nc; j++) {
      
      /* Compute wj = Akj vk */
      double wj = O2SCL_IX2(M,ir,j);
      
      for (i = ir+1; i < last; i++) {
	wj += O2SCL_IX2(M,i,j)*O2SCL_IX2(M2,i-ir+ir2,ic2);
      }
      
      /* Aij = Aij - tau vi wj */
      
      /* i = 0 */
      O2SCL_IX2(M,ir,j)-=tau*wj;
      
      /* i = 1 .. M-1 */
      for (i = ir+1; i < last; i++) {
	double Aij = O2SCL_IX2(M,i,j);
	double vi = O2SCL_IX2(M2,i-ir+ir2,ic2);
	O2SCL_IX2(M,i,j)=Aij - tau * vi * wj;
      }
    }
    return o2scl::gsl_success;
  }

  /** 
      \brief Apply a householder transformation \c v to vector \c w 
  */
  template<class vec_t>
    int householder_hv(const size_t N, double tau, const vec_t &v, vec_t &w) {
    
    if (tau==0) return o2scl::gsl_success;
    
    // compute d=v'w
    
    double d0=O2SCL_IX(w,0);
    double d1, d;
      
    d1=o2scl_cblas::ddot_subvec(N,v,w,1);
      
    d=d0+d1;
      
    // compute w=w-tau(v)(v'w)
    double w0=O2SCL_IX(w,0);
    O2SCL_IX(w,0)=w0-tau*d;
    
    o2scl_cblas::daxpy_subvec(N,-tau*d,v,w,1);

    return o2scl::gsl_success;
  }

  /** 
      \brief Apply a householder transformation \c v to vector \c w 

      Used in QR_QTvec().
  */
  template<class mat_t, class vec_t>
    int householder_hv_sub(const mat_t &M, vec_t &w, double tau,
			   const size_t ie, const size_t N) {
    
    if (tau==0) return o2scl::gsl_success;
    
    // compute d=v'w
    
    double d0=O2SCL_IX(w,ie);
    double d1, d;
      
    d1=o2scl_cblas::ddot_hv_sub(N,M,w,ie);
      
    d=d0+d1;
      
    // compute w=w-tau(v)(v'w)
    double w0=O2SCL_IX(w,ie);
    O2SCL_IX(w,ie)=w0-tau*d;
    
    o2scl_cblas::daxpy_hv_sub(N,-tau*d,M,w,ie);

    return o2scl::gsl_success;
  }

  /** 
      \brief Special version of householder transformation for
      QR_unpack()
  */
  template<class mat1_t, class mat2_t>
    int householder_hm_sub2(const size_t M, const size_t ic, 
			    double tau, const mat1_t &mv, mat2_t &A) {
    
    if (tau == 0.0) {
      return o2scl::gsl_success;
    }
    
    size_t i, j;
    
    for (j = ic; j < M; j++) {
      
      /* Compute wj = Akj vk */
      double wj = O2SCL_IX2(A,ic,j);
      
      for (i = ic+1; i < M; i++) {
	wj += O2SCL_IX2(A,i,j)*O2SCL_IX2(mv,i,ic);
      }
      
      /* Aij = Aij - tau vi wj */
      
      /* i = 0 */
      O2SCL_IX2(A,ic,j)-=tau*wj;
      
      /* i = 1 .. M-1 */
      for (i = ic+1; i < M; i++) {
	double Aij = O2SCL_IX2(A,i,j);
	double vi = O2SCL_IX2(mv,i,ic);
	O2SCL_IX2(A,i,j)=Aij - tau * vi * wj;
      }
    }
    return o2scl::gsl_success;
  }

#ifdef DOXYGENP
}
#endif
