/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/* linalg/tridiag.c
 * 
 * Copyright (C) 1996, 1997, 1998, 1999, 2000, 2002, 2004, 
 * 2007 Gerard Jungman, Brian Gough, David Necas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 .
*/
/** \file tridiag_base.h
    \brief File for solving tridiagonal systems
*/

#ifdef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Solve a symmetric tridiagonal linear system

      \future Convert into class for memory managment and combine
      with other functions below
    
      For description of method see [Engeln-Mullges + Uhlig, p. 92]
      \verbatim
      *
      *     diag[0]  offdiag[0]             0   .....
      *  offdiag[0]     diag[1]    offdiag[1]   .....
      *           0  offdiag[1]       diag[2]
      *           0           0    offdiag[2]   .....
      \endverbatim

  */
  template<class vec_t, class vec2_t> 
    int solve_tridiag_sym(const vec_t &diag, const vec2_t &offdiag, 
			  const vec_t &b, vec_t &x, size_t N) {
    int status;
    double *gamma = (double *) malloc (N * sizeof (double));
    double *alpha = (double *) malloc (N * sizeof (double));
    double *c = (double *) malloc (N * sizeof (double));
    double *z = (double *) malloc (N * sizeof (double));

    if (gamma == 0 || alpha == 0 || c == 0 || z == 0) {
      status = GSL_ENOMEM;
    } else {
      size_t i, j;

      /* Cholesky decomposition
	 A = L.D.L^t
	 lower_diag(L) = gamma
	 diag(D) = alpha
      */
      alpha[0] = O2SCL_IX(diag,0);
      gamma[0] = O2SCL_IX(offdiag,0) / alpha[0];

      for (i = 1; i < N - 1; i++) {
	alpha[i] = O2SCL_IX(diag,i) - O2SCL_IX(offdiag,i-1) * gamma[i - 1];
	gamma[i] = O2SCL_IX(offdiag,i) / alpha[i];
      }

      if (N > 1) {
	alpha[N-1]=O2SCL_IX(diag,N-1)-O2SCL_IX(offdiag,N-2)*gamma[N-2];
      }
	
      /* update RHS */
      z[0] = b[0];
      for (i = 1; i < N; i++) {
	z[i] = O2SCL_IX(b,i) - gamma[i - 1] * z[i - 1];
      } for (i = 0; i < N; i++) {
	c[i] = z[i] / alpha[i];
      }

      /* backsubstitution */
      O2SCL_IX(x,N-1)=c[N-1];
      if (N >= 2) {
	for (i = N - 2, j = 0; j <= N - 2; j++, i--) {
	  O2SCL_IX(x,i) = c[i] - gamma[i] * O2SCL_IX(x,i+1);
	}
      }

      status = GSL_SUCCESS;
    }

    if (z != 0)
      free (z);
    if (c != 0)
      free (c);
    if (alpha != 0)
      free (alpha);
    if (gamma != 0)
      free (gamma);

    return status;
  }

  /** 
      \brief Solve an asymmetric tridiagonal linear system
    
      This function uses plain gauss elimination, 
      only not bothering with the zeroes. 

      \verbatim
      * 
      *       diag[0]  abovediag[0]             0   .....
      *  belowdiag[0]       diag[1]  abovediag[1]   .....
      *             0  belowdiag[1]       diag[2]
      *             0             0  belowdiag[2]   .....
      \endverbatim
    
  */
  template<class vec_t, class vec2_t> 
    int solve_tridiag_nonsym(const vec_t &diag, const vec2_t &abovediag, 
			     const vec2_t &belowdiag, const vec_t &rhs, 
			     vec_t &x, size_t N) {
    int status;
    double *alpha = (double *) malloc (N * sizeof (double));
    double *z = (double *) malloc (N * sizeof (double));

    if (alpha == 0 || z == 0) {
      status = GSL_ENOMEM;
    } else {
      size_t i, j;

      /* Bidiagonalization (eliminating belowdiag)
	 & rhs update
	 diag' = alpha
	 rhs' = z
      */
      alpha[0] = O2SCL_IX(diag,0);
      z[0] = O2SCL_IX(rhs,0);

      for (i = 1; i < N; i++) {
	const double t = O2SCL_IX(belowdiag,i-1)/alpha[i-1];
	alpha[i] = O2SCL_IX(diag,i) - t*O2SCL_IX(abovediag,i-1);
	z[i] = O2SCL_IX(rhs,i) - t*z[i-1];
	/* FIXME!!! */
	if (alpha[i] == 0) {
	  status = GSL_EZERODIV;
	  goto solve_tridiag_nonsym_END;
	}
      }
      
      /* backsubstitution */
      O2SCL_IX(x,N-1) = z[N - 1]/alpha[N - 1];
      if (N >= 2) {
	for (i = N - 2, j = 0; j <= N - 2; j++, i--) {
	  O2SCL_IX(x,i) = (z[i] - O2SCL_IX(abovediag,i) * 
			   O2SCL_IX(x,i+1))/alpha[i];
	}
      }
      
      status = GSL_SUCCESS;
    }

  solve_tridiag_nonsym_END:

    if (z != 0)
      free (z);
    if (alpha != 0)
      free (alpha);

    return status;
  }

  /** 
      \brief Solve a symmetric cyclic tridiagonal linear system
    
      \todo Put reference in correctly.

      For a description of the method see [Engeln-Mullges + Uhlig, p. 96]
      \verbatim
      *
      *      diag[0]  offdiag[0]             0   .....  offdiag[N-1]
      *   offdiag[0]     diag[1]    offdiag[1]   .....
      *            0  offdiag[1]       diag[2]
      *            0           0    offdiag[2]   .....
      *          ...         ...
      * offdiag[N-1]         ...
      \endverbatim
    
  */
  template<class vec_t> 
    int solve_cyc_tridiag_sym(const vec_t &diag, const vec_t &offdiag, 
			      const vec_t &b, vec_t &x, size_t N) {
    int status;
    double * delta = (double *) malloc (N * sizeof (double));
    double * gamma = (double *) malloc (N * sizeof (double));
    double * alpha = (double *) malloc (N * sizeof (double));
    double * c = (double *) malloc (N * sizeof (double));
    double * z = (double *) malloc (N * sizeof (double));

    if (delta == 0 || gamma == 0 || alpha == 0 || c == 0 || z == 0) {
      status = GSL_ENOMEM;
    } else {
      size_t i, j;
      double sum = 0.0;
      
      /* factor */

      if (N == 1)  {
	x[0] = b[0] / O2SCL_IX(diag,0);
	return GSL_SUCCESS;
      }

      alpha[0] = O2SCL_IX(diag,0);
      gamma[0] = O2SCL_IX(offdiag,0) / alpha[0];
      delta[0] = O2SCL_IX(offdiag,N-1) / alpha[0];
      
      for (i = 1; i < N - 2; i++) {
	alpha[i] = O2SCL_IX(diag,i) - O2SCL_IX(offdiag,i-1) * gamma[i - 1];
	gamma[i] = O2SCL_IX(offdiag,i) / alpha[i];
	delta[i] = -delta[i - 1] * O2SCL_IX(offdiag,i-1) / alpha[i];
      }

      for (i = 0; i < N - 2; i++) {
	sum += alpha[i] * delta[i] * delta[i];
      }

      alpha[N - 2] = diag[ (N - 2)] - O2SCL_IX(offdiag,N-3) * gamma[N - 3];

      gamma[N - 2] = (offdiag[(N - 2)] - offdiag[(N - 3)] * 
		      delta[N - 3]) / alpha[N - 2];
      
      alpha[N - 1] = diag[(N - 1)] - sum - alpha[(N - 2)] * 
	gamma[N - 2] * gamma[N - 2];
      
      /* update */
      z[0] = b[0];
      for (i = 1; i < N - 1; i++) {
	z[i] = O2SCL_IX(b,i) - z[i - 1] * gamma[i - 1];
      }
      sum = 0.0;
      for (i = 0; i < N - 2; i++) {
	sum += delta[i] * z[i];
      }
      z[N - 1] = b[(N - 1)] - sum - gamma[N - 2] * z[N - 2];
      for (i = 0; i < N; i++) {
	c[i] = z[i] / alpha[i];
      }

      /* backsubstitution */
      O2SCL_IX(x,N-1) = c[N - 1];
      x[(N - 2)] = c[N - 2] - gamma[N - 2] * O2SCL_IX(x,N-1);
      if (N >= 3) {
	for (i = N - 3, j = 0; j <= N - 3; j++, i--) {
	  O2SCL_IX(x,i) = c[i] - gamma[i] * x[(i + 1)] - 
	    delta[i] * O2SCL_IX(x,N-1);
	}
      }

      status = GSL_SUCCESS;
    }

    if (z != 0)
      free (z);
    if (c != 0)
      free (c);
    if (alpha != 0)
      free (alpha);
    if (gamma != 0)
      free (gamma);
    if (delta != 0)
      free (delta);

    return status;
  }

  /** 
      \brief Solve an asymmetric cyclic tridiagonal linear system
    
      This function solves the following system w/o the corner
      elements and then use Sherman-Morrison formula to compensate for
      them
      \verbatim
      *
      *        diag[0]  abovediag[0]             0   .....  belowdiag[N-1]
      *   belowdiag[0]       diag[1]  abovediag[1]   .....
      *              0  belowdiag[1]       diag[2]
      *              0             0  belowdiag[2]   .....
      *            ...           ...
      * abovediag[N-1]           ...
      \endverbatim
      
  */
  template<class vec_t> 
    int solve_cyc_tridiag_nonsym(const vec_t &diag, const vec_t &abovediag, 
				 const vec_t &belowdiag, const vec_t &rhs, 
				 vec_t &x, size_t N) {
    int status;
    double *alpha = (double *) malloc (N * sizeof (double));
    double *zb = (double *) malloc (N * sizeof (double));
    double *zu = (double *) malloc (N * sizeof (double));
    double *w = (double *) malloc (N * sizeof (double));
    double beta;

    if (alpha == 0 || zb == 0 || zu == 0 || w == 0) {
      status = GSL_ENOMEM;
    } else {
      /* Bidiagonalization (eliminating belowdiag)
	 & rhs update
	 diag' = alpha
	 rhs' = zb
	 rhs' for Aq=u is zu
      */
      zb[0] = O2SCL_IX(rhs,0);
      if (O2SCL_IX(diag,0) != 0) {
	beta = -O2SCL_IX(diag,0); 
      } else {
	beta = 1;
      }
      const double q = 1 - O2SCL_IX(abovediag,0)*O2SCL_IX(belowdiag,0)/
	(O2SCL_IX(diag,0)*diag[1]);
      if (fabs(q/beta) > 0.5 && fabs(q/beta) < 2) {
	beta *= (fabs(q/beta) < 1) ? 0.5 : 2;
      }
      zu[0] = beta;
      alpha[0] = O2SCL_IX(diag,0) - beta;

      {
	size_t i;
	for (i = 1; i+1 < N; i++) {
	  const double t = O2SCL_IX(belowdiag,i-1)/alpha[i-1];
	  alpha[i] = O2SCL_IX(diag,i) - t*O2SCL_IX(abovediag,i-1);
	  zb[i] = O2SCL_IX(rhs,i) - t*zb[i-1];
	  zu[i] = -t*zu[i-1];
	  /* FIXME!!! */
	  if (alpha[i] == 0) {
	    status = GSL_EZERODIV;
	    goto solve_cyc_tridiag_nonsym_END;
	  }
	}
      }

      {
	const size_t i = N-1;
	const double t = O2SCL_IX(belowdiag,i-1)/alpha[i-1];
	alpha[i]=O2SCL_IX(diag,i)-O2SCL_IX(abovediag,i)*
	  O2SCL_IX(belowdiag,i)/beta-
	  t*O2SCL_IX(abovediag,i-1);
	zb[i] = O2SCL_IX(rhs,i) - t*zb[i-1];
	zu[i] = O2SCL_IX(abovediag,i) - t*zu[i-1];

	/* FIXME!!! */
	if (alpha[i] == 0) {
	  status = GSL_EZERODIV;
	  goto solve_cyc_tridiag_nonsym_END;
	}
      }

      {
	/* backsubstitution */
	size_t i, j;
	w[N-1] = zu[N-1]/alpha[N-1];
	O2SCL_IX(x,N-1) = zb[N-1]/alpha[N-1];
	for (i = N - 2, j = 0; j <= N - 2; j++, i--) {
	  w[i] = (zu[i] - O2SCL_IX(abovediag,i) * w[i+1])/alpha[i];
	  O2SCL_IX(x,i) = (zb[i] - O2SCL_IX(abovediag,i) * 
			   O2SCL_IX(x,i + 1))/alpha[i];
	}
      }
      
      /* Sherman-Morrison */
      const double vw = w[0] + O2SCL_IX(belowdiag,N-1)/beta * w[N-1];
      const double vx = O2SCL_IX(x,0) + 
	O2SCL_IX(belowdiag,N-1)/beta * O2SCL_IX(x,N-1);

      /* FIXME!!! */
      if (vw + 1 == 0) {
	status = GSL_EZERODIV;
	goto solve_cyc_tridiag_nonsym_END;
      }

      {
	size_t i;
	for (i = 0; i < N; i++)
	  O2SCL_IX(x,i) -= vx/(1 + vw)*w[i];
      }

      status = GSL_SUCCESS;
    }

  solve_cyc_tridiag_nonsym_END:

    if (zb != 0)
      free (zb);
    if (zu != 0)
      free (zu);
    if (w != 0)
      free (w);
    if (alpha != 0)
      free (alpha);

    return status;
  }

#ifdef DOXYGENP
}
#endif
