/*
  -------------------------------------------------------------------

  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner

  This file is part of O2scl.

  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_MMIN_BFGS2_H
#define O2SCL_GSL_MMIN_BFGS2_H

#include <gsl/gsl_blas.h>
#include <gsl/gsl_poly.h>
#include <gsl/gsl_multimin.h>
#include <o2scl/multi_min.h>
#include <o2scl/cblas.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Virtual base for the gsl_mmin_bfgs2 wrapper

      This class is useful so that the gsl_mmin_linmin class doesn't
      need to depend on any template parameters, even though it will
      need a wrapping object as an argument for the
      gsl_mmin_linmin::minimize() function.
  */
  class gsl_mmin_wrap_base {
  public:
    virtual ~gsl_mmin_wrap_base() {}
    /// Function
    virtual double wrap_f(double alpha, int params)=0;
    /// Derivative
    virtual double wrap_df(double alpha, int params)=0;
    /// Function and derivative
    virtual void wrap_fdf(double alpha, int params, double *f, double *df)=0;
  };

  /**
     \brief Wrapper class for the gsl_mmin_bfgs2 minimizer

     This is a reimplmentation of the internal GSL wrapper for
     function calls in the BFGS minimizer.
     
     \future There's a bit of extra vector copying here which 
     could potentially be avoided.
  */
  template<class param_t, class func_t, class vec_t=ovector_base, 
    class alloc_vec_t=ovector, class alloc_t=ovector_alloc,
    class dfunc_t=grad_funct<param_t,ovector_base>, 
    class auto_grad_t=gradient<param_t,func_t,ovector_base> > 
    class gsl_mmin_wrapper : public gsl_mmin_wrap_base {
    
#ifndef DOXYGEN_INTERNAL
    
    protected:

    /// Function
    func_t *func;

    /// Derivative
    dfunc_t *dfunc;

    /// The automatic gradient object
    auto_grad_t *agrad;
    
    /// Parameters
    param_t *pa;

    /// True if the gradient was given by the user
    bool grad_given;
    
    /** \name fixed values 
     */
    //@{
    gsl_vector *x;
    gsl_vector *g;
    gsl_vector *p;
    //@}
    
    /** \name cached values, for x(alpha) = x + alpha * p 
     */
    //@{
    double f_alpha;
    double df_alpha;
    //@}
    
    /** \name cache keys
     */
    //@{
    double f_cache_key;
    double df_cache_key;
    double x_cache_key;
    double g_cache_key;
    //@}
    
    /// Move to a new point, using the cached value if possible
    void moveto(double alpha) {

      if (alpha == x_cache_key) {
	/* using previously cached position */
	return;
      }
      
      /* set x_alpha = x + alpha * p */
      
      gsl_vector *tmp=gsl_vector_alloc(dim);
      for(size_t i=0;i<dim;i++) {
	av_x_alpha[i]=gsl_vector_get(x,i);
	gsl_vector_set(tmp,i,av_x_alpha[i]);
      }

      gsl_blas_daxpy(alpha,p,tmp);
      for(size_t i=0;i<dim;i++) {
	av_x_alpha[i]=gsl_vector_get(tmp,i);
      }
      gsl_vector_free(tmp);

      x_cache_key = alpha;
    }

    /// Compute the slope
    double slope() {
      double df;
      gsl_vector *tmp=gsl_vector_alloc(dim);
      for(size_t i=0;i<dim;i++) {
	gsl_vector_set(tmp,i,av_g_alpha[i]);
      }
      gsl_blas_ddot(tmp,p,&df);
      for(size_t i=0;i<dim;i++) {
	av_g_alpha[i]=gsl_vector_get(tmp,i);
      }
      gsl_vector_free(tmp);
      return df;
    }

    /// Evaluate the function
    virtual double wrap_f(double alpha, int params) {
      if (alpha==f_cache_key) {
	return f_alpha;
      }
      moveto(alpha);
      (*func)(dim,av_x_alpha,f_alpha,*pa);

      f_cache_key = alpha;
      return f_alpha;
    }

    /// Evaluate the derivative
    virtual double wrap_df(double alpha, int params) {

      /* using previously cached df(alpha) */
      if (alpha==df_cache_key) return df_alpha;

      moveto(alpha);
      if (alpha!=g_cache_key) {
	if (grad_given) {
	  (*dfunc)(dim,av_x_alpha,av_g_alpha,*pa);
	} else {
	  (*agrad)(dim,av_x_alpha,av_g_alpha,*pa);
	}
	g_cache_key=alpha;
      }
      df_alpha=slope();
      df_cache_key=alpha;

      return df_alpha;
    }

    /// Evaluate the function and the derivative
    virtual void wrap_fdf(double alpha, int params, double *f, double *df) {

      /* Check for previously cached values */
      if (alpha ==  f_cache_key && alpha ==  df_cache_key) {
	*f =  f_alpha;
	*df =  df_alpha;
	return;
      }
      if (alpha ==  f_cache_key || alpha ==  df_cache_key) {
	*f = wrap_f (alpha, params);
	*df = wrap_df (alpha, params);
	return;
      }
      
      moveto(alpha);
      (*func)(dim,av_x_alpha,f_alpha,*pa);
      if (grad_given) {
	(*dfunc)(dim,av_x_alpha,av_g_alpha,*pa);
      } else {
	(*agrad)(dim,av_x_alpha,av_g_alpha,*pa);
      }
      f_cache_key = alpha;
      g_cache_key = alpha;
      
      df_alpha = slope();
      df_cache_key = alpha;
      
      *f =  f_alpha;
      *df =  df_alpha;
      
      return;
    }
    
#endif

    public:

    /// Temporary storage
    alloc_vec_t av_x_alpha;

    /// Temporary storage
    alloc_vec_t av_g_alpha;

    /// Number of minimization dimensions
    size_t dim;
    
    /// Initialize wrapper
    void prepare_wrapper(func_t &ufunc, dfunc_t *udfunc,
			 param_t &upa, gsl_vector *t_x,
			 double f, gsl_vector *t_g, gsl_vector *t_p,
			 auto_grad_t *ag) {
      
      func=&ufunc;
      dfunc=udfunc;
      if (dfunc!=0) grad_given=true;
      else grad_given=false;
      agrad=ag;
      pa=&upa;
      
      x=t_x;
      g=t_g;
      p=t_p;
      
      x_cache_key=0.0;
      f_cache_key=0.0;
      g_cache_key=0.0;
      df_cache_key=0.0;

      for(size_t i=0;i<dim;i++) {
	av_x_alpha[i]=gsl_vector_get(x,i);
	av_g_alpha[i]=gsl_vector_get(g,i);
      }
      
      f_alpha=f;
      df_alpha=slope();

      return;
    }
    
    /// Update position
    void update_position(double alpha, vec_t &t_x, double *t_f, 
			 vec_t &t_g) {
      
      /* ensure that everything is fully cached */
      { 
	double t_f_alpha, t_df_alpha; 
	wrap_fdf(alpha,0,&t_f_alpha,&t_df_alpha); 
      }
      
      *t_f = f_alpha;
      for(size_t i=0;i<dim;i++) {
	t_x[i]=av_x_alpha[i];
	t_g[i]=av_g_alpha[i];
      }
    }
    
    /** 
	\brief Convert cache values to the new minimizer direction

	Convert the cache values from the end of the current minimisation
	to those needed for the start of the next minimisation, alpha=0
    */
    void change_direction() {
      
      /* The new x_alpha for alpha=0 is the current position and the
	 new g_alpha for alpha=0 is the current gradient at the
	 endpoint
      */
      for(size_t i=0;i<dim;i++) {
	av_x_alpha[i]=gsl_vector_get(x,i);
	av_g_alpha[i]=gsl_vector_get(g,i);
      }
      x_cache_key = 0.0;
      g_cache_key = 0.0;
      
      /* The function value does not change */
      f_cache_key = 0.0;
      
      /* Calculate the slope along the new direction vector, p */
      df_alpha = slope ();
      df_cache_key = 0.0;
      
      return;
    }
    
  };

  /**
     \brief The line minimizer for gsl_mmin_bfgs2
  */
  class gsl_mmin_linmin {

#ifndef DOXYGEN_INTERNAL

  protected:

    /**
       \brief Minimize the interpolating quadratic

       Find a minimum in x=[0,1] of the interpolating quadratic through
       (0,f0) (1,f1) with derivative fp0 at x=0.  The interpolating
       polynomial is q(x) = f0 + fp0 * z + (f1-f0-fp0) * z^2
    */
    double interp_quad(double f0, double fp0, double f1, double zl, 
		       double zh);
    
    /**
       \brief Minimize the interpolating cubic
       
       Find a minimum in x=[0,1] of the interpolating cubic through
       (0,f0) (1,f1) with derivatives fp0 at x=0 and fp1 at x=1.
       
       The interpolating polynomial is:
       
       c(x) = f0 + fp0 * z + eta * z^2 + xi * z^3
       
       where eta=3*(f1-f0)-2*fp0-fp1, xi=fp0+fp1-2*(f1-f0).
    */
    double cubic(double c0, double c1, double c2, double c3, double z);
    
    /// Test to see curvature is positive
    void check_extremum(double c0, double c1, double c2, double c3, double z,
			double *zmin, double *fmin);
    
    /// Interpolate using a cubic
    double interp_cubic(double f0, double fp0, double f1, 
			double fp1, double zl, double zh);
    
    /// Perform the interpolation
    double interpolate(double a, double fa, double fpa, double b, 
		       double fb, double fpb, double xmin, double xmax,
		       int order);
#endif
    
  public:

    /** 
	\brief The line minimization
	
	Recommended values from \ref Fletcher87 are
	<tt>rho = 0.01, sigma = 0.1, tau1 = 9, tau2 = 0.05, 
	tau3 = 0.5 </tt>
    */
    int minimize(gsl_mmin_wrap_base &wrap, double rho, double sigma,
		 double tau1, double tau2, double tau3,
		 int order, double alpha1, double *alpha_new);
  };

  /** \brief Multidimensional minimization by the BFGS
      algorithm (GSL)
      
      The functions mmin() and mmin_de() minimize a given function
      until the gradientis smaller than the value of multi_min::tolf
      (which defaults to \f$ 10^{-4} \f$ ).

      See an example for the usage of this class
      in \ref ex_mmin_sect .
      
      This class includes the optimizations from the GSL minimizer \c
      vector_bfgs2.

      \comment
      Doxygen has had trouble parsing this template, so I have
      simplified it slightly for the documentation below.
      \endcomment
  */
#ifdef DOXYGENP
  template<class param_t, class func_t=multi_funct<param_t>, 
    class vec_t=ovector_base, class alloc_vec_t=ovector,
    class alloc_t=ovector_alloc,
    class dfunc_t=grad_funct<param_t,ovector_base>, 
    class auto_grad_t=gradient<param_t,func_t,ovector_base>,
    class def_auto_grad_t=simple_grad<param_t,func_t,ovector_base> > 
    class gsl_mmin_bfgs2 : public multi_min
#else
  template<class param_t, class func_t=multi_funct<param_t>, 
    class vec_t=ovector_base, class alloc_vec_t=ovector,
    class alloc_t=ovector_alloc,
    class dfunc_t=grad_funct<param_t,ovector_base>, 
    class auto_grad_t=gradient<param_t,func_t,ovector_base>,
    class def_auto_grad_t=simple_grad<param_t,func_t,ovector_base> > 
    class gsl_mmin_bfgs2 : public multi_min<param_t,func_t,func_t,vec_t> 
#endif
    {
    
#ifndef DOXYGEN_INTERNAL

    protected:

    /// \name The original variables from the GSL state structure
    //@{
    int iter;
    double step;
    double g0norm;
    double pnorm;
    double delta_f;
    /* f'(0) for f(x-alpha*p) */
    double fp0;                   
    gsl_vector *x0;
    gsl_vector *g0;
    gsl_vector *p;
    /* work space */
    gsl_vector *dx0;
    gsl_vector *dg0;
    /* wrapper function */
    gsl_mmin_wrapper<param_t,func_t,vec_t,alloc_vec_t,alloc_t,
    dfunc_t,auto_grad_t> wrap;
    /* minimization parameters */
    double rho;
    double sigma;
    double tau1;
    double tau2;
    double tau3;
    int order;
    //@}

    /// The line minimizer
    gsl_mmin_linmin lm;
      
    /// \name Store the arguments to set() so we can use them for iterate()
    //@{
    vec_t *st_x;
    gsl_vector *st_dx;
    alloc_vec_t st_grad;
    double st_f;
    //@}

    /// Memory size
    size_t dim;

    /// Memory allocation
    alloc_t ao;

    /// Automatic gradient object
    auto_grad_t *agrad;
    
#endif
    
    public:
     
    gsl_mmin_bfgs2() {
      lmin_tol=1.0e-4;
      this->tolf=1.0e-6;
      step_size=0.01;
      agrad=&def_grad;
    }
    
    virtual ~gsl_mmin_bfgs2() {}

    /// Perform an iteration
    virtual int iterate() {

    this->last_conv=0;

      double alpha = 0.0, alpha1;

      double pg, dir;
      int status;
      
      double f0 = st_f;
      
      if (pnorm == 0.0 || g0norm == 0.0 || fp0 == 0) {
	gsl_vector_set_zero(st_dx);
	this->last_conv=gsl_enoprog;
	O2SCL_CONV2_RET("Either pnorm, g0norm, or fp0 vanished in ",
			"in gsl_mmin_bfgs2::iterate().",
			gsl_enoprog,this->err_nonconv);
      }
      
      if (delta_f < 0) {
	double del = GSL_MAX_DBL (-delta_f, 10 * GSL_DBL_EPSILON * fabs(f0));
	alpha1 = GSL_MIN_DBL (1.0, 2.0 * del / (- fp0));
      } else {
	alpha1 = fabs(step);
      }
	
      /* line minimisation, with cubic interpolation (order = 3) */
      
      status=lm.minimize(wrap,rho,sigma,tau1,tau2,tau3,order,
			 alpha1,&alpha);
      
      if (status != gsl_success) {
	this->last_conv=gsl_enoprog;
	O2SCL_CONV2_RET("Variable stepb vanished in ",
			"in gsl_mmin_bfgs2::iterate().",
			gsl_enoprog,this->err_nonconv);
      }
      
      wrap.update_position(alpha,*st_x,&st_f,st_grad);

      delta_f = st_f - f0;

      /* Choose a new direction for the next step */

      {
	/* This is the BFGS update: */
	/* p' = g1 - A dx - B dg */
	/* A = - (1+ dg.dg/dx.dg) B + dg.g/dx.dg */
	/* B = dx.g/dx.dg */

	double dxg, dgg, dxdg, dgnorm, A, B;

	/* dx0 = x - x0 */
	for(size_t i=0;i<dim;i++) gsl_vector_set(dx0,i,(*st_x)[i]);
	gsl_blas_daxpy (-1.0, x0, dx0);

	/* keep a copy */
	gsl_vector_memcpy (st_dx, dx0);  

	/* dg0 = g - g0 */
	for(size_t i=0;i<dim;i++) gsl_vector_set(dg0,i,st_grad[i]);
	gsl_blas_daxpy (-1.0, g0, dg0);

	ovector *odx0=(ovector *)dx0;
	ovector *odg0=(ovector *)dg0;
	dxg=o2scl_cblas::ddot(dim,*odx0,st_grad);
	dgg=o2scl_cblas::ddot(dim,*odg0,st_grad);
	gsl_blas_ddot (dx0, dg0, &dxdg);

	dgnorm = gsl_blas_dnrm2 (dg0);

	if (dxdg != 0) {
	  B = dxg / dxdg;
	  A = -(1.0 + dgnorm * dgnorm / dxdg) * B + dgg / dxdg;
	} else {
	  B = 0;
	  A = 0;
	}
	
	for(size_t i=0;i<dim;i++) gsl_vector_set(p,i,st_grad[i]);
	gsl_blas_daxpy (-A, dx0, p);
	gsl_blas_daxpy (-B, dg0, p);
      }
      
      for(size_t i=0;i<dim;i++) gsl_vector_set(g0,i,st_grad[i]);
      for(size_t i=0;i<dim;i++) gsl_vector_set(x0,i,(*st_x)[i]);
      
      g0norm = gsl_blas_dnrm2 (g0);
      pnorm = gsl_blas_dnrm2 (p);

      /* update direction and fp0 */

      ovector *op=(ovector *)p;
      pg=o2scl_cblas::ddot(dim,st_grad,*op);
      dir = (pg >= 0.0) ? -1.0 : +1.0;
      gsl_blas_dscal (dir /  pnorm, p);
      pnorm = gsl_blas_dnrm2 (p);
      gsl_blas_ddot (p, g0, & fp0);

      wrap.change_direction();

      return gsl_success;
    }
    
    /// Return string denoting type("gsl_mmin_bfgs2")
    virtual const char *type() { return "gsl_mmin_bfgs2";}

    /// Allocate the memory
    virtual int allocate(size_t n) {

      p=gsl_vector_calloc(n);
      if (p == 0) {
	O2SCL_ERR2_RET("Failed to allocate p in ",
		      "gsl_mmin_bfgs2::allocate().",gsl_enomem);
      }

      x0=gsl_vector_calloc(n);
      if (x0 == 0) {
	gsl_vector_free(p);
	O2SCL_ERR2_RET("Failed to allocate x0 in ",
		      "gsl_mmin_bfgs2::allocate().",gsl_enomem);
      }

      g0=gsl_vector_calloc(n);
      if (g0 == 0) {
	gsl_vector_free(x0);
	gsl_vector_free(p);
	O2SCL_ERR2_RET("Failed to allocate g0 in ",
		       "gsl_mmin_bfgs2::allocate().",gsl_enomem);
      }
      
      dx0=gsl_vector_calloc(n);
      if (dx0 == 0) {
	gsl_vector_free(g0);
	gsl_vector_free(x0);
	gsl_vector_free(p);
	O2SCL_ERR_RET
	  ("Failed to allocate dx0 in gsl_mmin_bfgs2::allocate().",
	   gsl_enomem);
      }

      dg0=gsl_vector_calloc(n);
      if (dg0 == 0) {
	gsl_vector_free(dx0);
	gsl_vector_free(g0);
	gsl_vector_free(x0);
	gsl_vector_free(p);
	O2SCL_ERR_RET
	  ("Failed to allocate dg0 in gsl_mmin_bfgs2::allocate().",
	   gsl_enomem);
      }

      st_dx=gsl_vector_alloc(n);
      ao.allocate(st_grad,n);

      ao.allocate(wrap.av_x_alpha,n);
      ao.allocate(wrap.av_g_alpha,n);
      wrap.dim=n;
      dim=n;

      return gsl_success;
    }
    
    /// Free the allocated memory
    virtual int free() {
      ao.free(wrap.av_x_alpha);
      ao.free(wrap.av_g_alpha);
      ao.free(st_grad);
      gsl_vector_free(dg0);
      gsl_vector_free(dx0);
      gsl_vector_free(g0);
      gsl_vector_free(x0);
      gsl_vector_free(p);
      gsl_vector_free(st_dx);
      wrap.dim=0;
      dim=0;
      return 0;
    }
      
    /// Reset the minimizer to use the current point as a new starting point
    int restart() {
      iter=0;
      return gsl_success;
    }

    /// Set the function and initial guess
    virtual int set(vec_t &x, double u_step_size, double tol_u, 
		    func_t &ufunc, param_t &upa) {
      
      iter=0;
      step=u_step_size;
      delta_f=0;
      
      st_x=&x;
      
      ufunc(dim,x,st_f,upa);
      agrad->set_function(ufunc);
      (*agrad)(dim,x,st_grad,upa);
      
      /* Use the gradient as the initial direction */

      for(size_t i=0;i<dim;i++) {
	gsl_vector_set(x0,i,x[i]);
	gsl_vector_set(g0,i,st_grad[i]);
      }
      g0norm=gsl_blas_dnrm2(g0);
      for(size_t i=0;i<dim;i++) {
	gsl_vector_set(p,i,st_grad[i]);
      }
      gsl_blas_dscal(-1/g0norm,p);

      // pnorm should be 1
      pnorm=gsl_blas_dnrm2(p);     
      fp0=- g0norm;

      /* Prepare the wrapper */
      
      wrap.prepare_wrapper(ufunc,0,upa,x0,st_f,g0,p,agrad);
      
      /* Prepare 1d minimisation parameters */

      rho=0.01;
      sigma=tol_u;
      tau1=9;
      tau2=0.05;
      tau3=0.5;
      // use cubic interpolation where possible 
      order=3;  

      return gsl_success;
 
    }

    /// Set the function, the gradient, and the initial guess
    virtual int set_de(vec_t &x, double u_step_size, double tol_u, 
		       func_t &ufunc, dfunc_t &udfunc, param_t &upa) {
      
      iter=0;
      step=u_step_size;
      delta_f=0;
      
      st_x=&x;
      
      ufunc(dim,x,st_f,upa);
      udfunc(dim,x,st_grad,upa);
      
      /* Use the gradient as the initial direction */

      for(size_t i=0;i<dim;i++) {
	gsl_vector_set(x0,i,x[i]);
	gsl_vector_set(g0,i,st_grad[i]);
      }
      g0norm=gsl_blas_dnrm2(g0);
      for(size_t i=0;i<dim;i++) {
	gsl_vector_set(p,i,st_grad[i]);
      }
      gsl_blas_dscal(-1/g0norm,p);

      // pnorm should be 1
      pnorm=gsl_blas_dnrm2(p);     
      fp0=-g0norm;

      /* Prepare the wrapper */
      
      wrap.prepare_wrapper(ufunc,&udfunc,upa,x0,st_f,g0,p,agrad);
      
      /* Prepare 1d minimisation parameters */

      rho=0.01;
      sigma=tol_u;
      tau1=9;
      tau2=0.05;
      tau3=0.5;
      // use cubic interpolation where possible 
      order=3;  

      return gsl_success;
 
    }

    /// The size of the first trial step
    double step_size;
    
    /// The tolerance for the 1-dimensional minimizer
    double lmin_tol;

    /// Default automatic gradient object
    def_auto_grad_t def_grad;

    /** \brief Calculate the minimum \c min of \c func w.r.t the
	array \c x of size \c nn.
    */
    virtual int mmin(size_t nn, vec_t &xx, double &fmin, param_t &pa,
		     func_t &ufunc) {

      if (nn==0) {
	O2SCL_ERR2_RET("Tried to minimize over zero variables ",
		       " in gsl_mmin_bfgs2::mmin().",gsl_einval);
      }

      int xiter=0, status;

      allocate(nn);

      set(xx,step_size,lmin_tol,ufunc,pa);

      do {

	xiter++;

	status=iterate();
	
	if (status) {
	  break;
	}

	// Equivalent to gsl_multimin_test_gradient with
	// additional code to print out present iteration
	
	double norm=o2scl_cblas::dnrm2(nn,st_grad);
	
	if(this->verbose>0) {
	  this->print_iter(nn,*st_x,st_f,xiter,
			   norm,this->tolf,"gsl_mmin_bfgs2");
	}

	if (norm<this->tolf) {
	  status=gsl_success;
	} else {
	  status=gsl_continue;
	}
	
      } while (status == gsl_continue && xiter < this->ntrial);
      
      for(size_t i=0;i<nn;i++) xx[i]=(*st_x)[i];
      fmin=st_f;
      
      free();
      this->last_ntrial=xiter;
      
      if (status==gsl_continue && xiter==this->ntrial) {
	this->last_conv=gsl_emaxiter;
	O2SCL_CONV_RET("Too many iterations in gsl_mmin_bfgs2::mmin().",
		       gsl_emaxiter,this->err_nonconv);
      }
      return status;
    }

    /** \brief Calculate the minimum \c min of \c func w.r.t the
	array \c x of size \c nn.
    */
    virtual int mmin_de(size_t nn, vec_t &xx, double &fmin, param_t &pa,
			func_t &ufunc, dfunc_t &udfunc) {

      if (nn==0) {
	O2SCL_ERR2_RET("Tried to minimize over zero variables ",
		       " in gsl_mmin_bfgs2::mmin().",gsl_einval);
      }

      int xiter=0, status;

      allocate(nn);

      set_de(xx,step_size,lmin_tol,ufunc,udfunc,pa);

      do {
	xiter++;

	status=iterate();
	
	if (status) {
	  break;
	}

	// Equivalent to gsl_multimin_test_gradient with
	// additional code to print out present iteration
	
	double norm=o2scl_cblas::dnrm2(nn,st_grad);
	
	if(this->verbose>0) {
	  this->print_iter(nn,*st_x,st_f,xiter,
			   norm,this->tolf,"gsl_mmin_bfgs2");
	}

	if (norm<this->tolf) status=gsl_success;
	else status=gsl_continue;

      } while (status == gsl_continue && xiter < this->ntrial);
      
      for(size_t i=0;i<nn;i++) xx[i]=(*st_x)[i];
      fmin=st_f;
      
      free();
      this->last_ntrial=xiter;

      if (status==gsl_continue && xiter==this->ntrial) {
	this->last_conv=gsl_emaxiter;
	O2SCL_CONV_RET("Too many iterations in gsl_mmin_bfgs2::mmin_de().",
		       gsl_emaxiter,this->err_nonconv);
      }

      return status;
    }

  };

  /**
     \brief An array version of gsl_mmin_bfgs2
     
     \comment
     Doxygen has had trouble parsing this template, so I have
     simplified it slightly for the documentation below.
     \endcomment
  */
#ifdef DOXYGENP
  template<class param_t, size_t nv> class gsl_mmin_bfgs2_array :
  public gsl_mmin_bfgs2<param_t,func_t,vec_t,alloc_vec_t,
    alloc_t,dfunc_t,auto_grad_t,def_auto_grad_t> {
  };
#else
  template<class param_t, size_t nv> class gsl_mmin_bfgs2_array :
  public gsl_mmin_bfgs2<param_t,multi_vfunct<param_t,nv>,
    double[nv],double[nv],array_alloc<double[nv]>,grad_vfunct<param_t,nv>,
    gradient_array<param_t,multi_vfunct<param_t,nv>,nv>,
    simple_grad_array<param_t,multi_vfunct<param_t,nv>,nv> > {
  };
#endif

#ifndef DOXYGENP
}
#endif

#endif
