/*
  -------------------------------------------------------------------

  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner

  This file is part of O2scl.

  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_MMIN_CONF_H
#define O2SCL_GSL_MMIN_CONF_H

#include <gsl/gsl_blas.h>
#include <gsl/gsl_multimin.h>
#include <o2scl/multi_min.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief Base minimization routines for gsl_mmin_conf and gsl_mmin_conp

      This class is used by the gsl_mmin_conf and gsl_mmin_conp 
      minimizers to perform the line minimization along a
      specified direction. It is not intended for a casual end-user. 
      
      Default template arguments
      - \c param_t - no default
      - \c func_t - \ref multi_funct\<param_t\>
      - \c vec_t - \ref ovector_base
      - \c alloc_vec_t - \ref ovector
      - \c alloc_t - \ref ovector_alloc
      - \c dfunc_t - \ref grad_funct\<param_t,ovector_base\>
      - \c auto_grad_t - \ref gradient\<param_t,func_t,ovector_base\>
      - \c def_auto_grad_t - \ref simple_grad\<param_t,func_t,ovector_base\>
      
      \comment
      Doxygen has had trouble parsing this template, so I have
      simplified it slightly for the documentation below.
      \endcomment
  */
#ifdef DOXYGENP
  template<class param_t, class func_t = multi_funct<param_t>, 
    class vec_t = ovector_base, class alloc_vec_t = ovector,
    class alloc_t = ovector_alloc, 
    class dfunc_t = grad_funct<param_t,ovector_base>,
    class auto_grad_t = gradient<param_t,func_t,ovector_base>,
    class def_auto_grad_t = simple_grad<param_t,func_t,ovector_base> > 
    class gsl_mmin_base : public multi_min
#else
    template<class param_t, class func_t = multi_funct<param_t>, 
    class vec_t = ovector_base, class alloc_vec_t = ovector,
    class alloc_t = ovector_alloc, 
    class dfunc_t = grad_funct<param_t,ovector_base>,
    class auto_grad_t = gradient<param_t,func_t,ovector_base>,
    class def_auto_grad_t = simple_grad<param_t,func_t,ovector_base> > 
    class gsl_mmin_base : public multi_min<param_t,func_t,func_t,vec_t>
#endif
    {
      
#ifndef DOXYGEN_INTERNAL

    protected:
    
    /// User-specified function
    func_t *func;

    /// User-specified gradient
    dfunc_t *grad;

    /// Automatic gradient object
    auto_grad_t *agrad;
      
    /// If true, a gradient has been specified
    bool grad_given;

    /// User-specified parameter
    param_t *params;
    /// Memory size
    size_t dim;
    /// Memory allocation
    alloc_t ao;
    /// Temporary vector
    alloc_vec_t avt;
    /// Temporary vector
    alloc_vec_t avt2;
      
    /// Take a step
    void take_step(const gsl_vector *x, const gsl_vector *px,
		   double stepx, double lambda, gsl_vector *x1x, 
		   gsl_vector *dx) {
	
      gsl_vector_set_zero(dx);
      gsl_blas_daxpy(-stepx*lambda,px,dx);
	
      gsl_vector_memcpy(x1x,x);
      gsl_blas_daxpy(1.0,dx,x1x);
	
      return;
    }

    /** 
	\brief Line minimization
	  
	Do a line minimisation in the region (xa,fa) (xc,fc) to find
	an intermediate (xb,fb) satisifying fa > fb < fc.  Choose an
	initial xb based on parabolic interpolation
    */
    void intermediate_point(const gsl_vector *x, const gsl_vector *px,
			    double lambda, double pg, double stepa, 
			    double stepc, double fa, double fc,
			    gsl_vector *x1x, gsl_vector *dx, 
			    gsl_vector *gradient, double *stepx, double *f) {
      double stepb, fb;

      bool trial_failed;

      // This looks like an infinite loop, but will eventually 
      // stop because stepb will eventually become zero. This
      // error condition is handled later in iterate().
      do {

	double u = fabs (pg * lambda * stepc);
	stepb = 0.5 * stepc * u / ((fc - fa) + u);
	  
	take_step (x, px, stepb, lambda, x1x, dx);

	// Evaluate the function
	for(size_t i=0;i<dim;i++) avt[i]=gsl_vector_get(x1x,i);
	(*func)(dim,avt,fb,*params);
	  
	trial_failed=false;
	if (fb >= fa  && stepb > 0.0) {
	  /* downhill step failed, reduce step-size and try again */
	  fc = fb;
	  stepc = stepb;
	  trial_failed=true;
	}

      } while (trial_failed);

      *stepx = stepb;
      *f = fb;

      // Evaluate the gradient
      if (grad_given) {
	(*grad)(dim,avt,avt2,*params);
      } else {
	(*agrad)(dim,avt,avt2,*params);
      }
      for(size_t i=0;i<dim;i++) gsl_vector_set(gradient,i,avt2[i]);

      return;
    }
      
    /** 
	\brief Perform the minimization

	Starting at (x0, f0) move along the direction p to find a
	minimum f(x0 - lambda * p), returning the new point x1 =
	x0-lambda*p, f1=f(x1) and g1 = grad(f) at x1.
	
	\future Remove goto statements
    */
    void minimize(const gsl_vector *x, const gsl_vector *xp, double lambda,
		  double stepa, double stepb, double stepc, double fa, 
		  double fb, double fc, double xtol, gsl_vector *x1x, 
		  gsl_vector *dx1x, gsl_vector *x2x, gsl_vector *dx2x, 
		  gsl_vector *gradient, double *xstep, double *f, 
		  double *gnorm_u) {

      double u = stepb;
      double v = stepa;
      double w = stepc;
      double fu = fb;
      double fv = fa;
      double fw = fc;

      double old2 = fabs(w - v);
      double old1 = fabs(v - u);

      double stepm, fm, pg, gnorm1;

      int xiter = 0;

      gsl_vector_memcpy (x2x, x1x);
      gsl_vector_memcpy (dx2x, dx1x);

      *f = fb;
      *xstep = stepb;
      *gnorm_u = gsl_blas_dnrm2(gradient);

    mid_trial:

      xiter++;

      if (xiter > nmaxiter) {
	// Exceeded maximum number of iterations
	return;  
      }
	
      {
	double dw = w - u;
	double dv = v - u;
	double du = 0.0;
	double e1 = ((fv - fu) * dw * dw + (fu - fw) * dv * dv);
	double e2 = 2.0 * ((fv - fu) * dw + (fu - fw) * dv);

	if (e2 != 0.0) {
	  du = e1 / e2;
	}

	if (du > 0.0 && du < (stepc - stepb) && fabs(du) < 0.5 * old2) {
	  stepm = u + du;
	} else if (du < 0.0 && du > (stepa - stepb) && 
		   fabs(du) < 0.5 * old2) {
	  stepm = u + du;
	} else if ((stepc - stepb) > (stepb - stepa)) {
	  stepm = 0.38 * (stepc - stepb) + stepb;
	} else {
	  stepm = stepb - 0.38 * (stepb - stepa);
	}
      }

      take_step (x, xp, stepm, lambda, x1x, dx1x);

      for(size_t i=0;i<dim;i++) avt[i]=gsl_vector_get(x1x,i);
      (*func)(dim,avt,fm,*params);
	
      if (fm > fb) {
	if (fm < fv) {
	  w = v;
	  v = stepm;
	  fw = fv;
	  fv = fm;
	} else if (fm < fw) {
	  w = stepm;
	  fw = fm;
	}
	if (stepm < stepb) {
	  stepa = stepm;
	  fa = fm;
	} else {
	  stepc = stepm;
	  fc = fm;
	}
	goto mid_trial;

      } else if (fm <= fb) {

	old2 = old1;
	old1 = fabs(u - stepm);
	w = v;
	v = u;
	u = stepm;
	fw = fv;
	fv = fu;
	fu = fm;

	gsl_vector_memcpy (x2x, x1x);
	gsl_vector_memcpy (dx2x, dx1x);

	for(size_t i=0;i<dim;i++) avt[i]=gsl_vector_get(x1x,i);
	if (grad_given) {
	  (*grad)(dim,avt,avt2,*params);
	} else {
	  (*agrad)(dim,avt,avt2,*params);
	}
	for(size_t i=0;i<dim;i++) gsl_vector_set(gradient,i,avt2[i]);

	gsl_blas_ddot (xp, gradient, &pg);
	gnorm1 = gsl_blas_dnrm2 (gradient);

	*f = fm;
	*xstep = stepm;
	*gnorm_u = gnorm1;

	if (fabs (pg * lambda / gnorm1) < xtol) {
	  /* SUCCESS */
	  return; 
	}
	    
	if (stepm < stepb) {
	  stepc = stepb;
	  fc = fb;
	  stepb = stepm;
	  fb = fm;
	} else {
	  stepa = stepb;
	  fa = fb;
	  stepb = stepm;
	  fb = fm;
	}
	goto mid_trial;
      }

      return;
    }

#endif

    public:

    /// Stepsize for finite-differencing ( default \f$ 10^{-4} \f$ )
    double deriv_h;

    /// Maximum iterations for line minimization (default 10)
    int nmaxiter;
      
    /// Default automatic \gradient object
    def_auto_grad_t def_grad;
      
    gsl_mmin_base() {
      deriv_h=1.0e-4;
      nmaxiter=10;
      grad_given=false;
      agrad=&def_grad;
    }
      
    /// Set the function
    int base_set(func_t &ufunc, param_t &pa, auto_grad_t &u_def_grad) {
      func=&ufunc;
      params=&pa;
      agrad=&u_def_grad;
      grad_given=false;
      return 0;
    }

    /// Set the function and the \gradient
    int base_set_de(func_t &ufunc, dfunc_t &udfunc, param_t &pa) {
      func=&ufunc;
      params=&pa;
      grad=&udfunc;
      grad_given=true;
      return 0;
    }

    /// Allocate memory
    int base_allocate(size_t nn) {
      ao.allocate(avt,nn);
      ao.allocate(avt2,nn);
      dim=nn;
      return 0;
    }

    /// Clear allocated memory
    int base_free() {
      ao.free(avt);
      ao.free(avt2);
      dim=0;
      return 0;
    }


    };
  
  /** \brief Multidimensional minimization by the Fletcher-Reeves 
      conjugate \gradient algorithm (GSL)

      The functions mmin() and mmin_de() \minimize a given function
      until the \gradient is smaller than the value of multi_min::tolf
      (which defaults to \f$ 10^{-4} \f$ ).

      See an example for the usage of this class
      in \ref ex_mmin_sect .
      
      \future A bit of needless copying is required in the function
      wrapper to convert from \c gsl_vector to the templated vector
      type. This can be fixed, probably by rewriting take_step
      to produce a vec_t &x1 rather than a gsl_vector *x1;

      Note that the state variable \c max_iter has not been included
      here, because it was not really used in the original GSL code
      for these minimizers.
      
      Default template arguments
      - \c param_t - no default
      - \c func_t - \ref multi_funct\<param_t\>
      - \c vec_t - \ref ovector_base
      - \c alloc_vec_t - \ref ovector
      - \c alloc_t - \ref ovector_alloc
      - \c dfunc_t - \ref grad_funct\<param_t,ovector_base\>
      - \c auto_grad_t - \ref gradient\<param_t,func_t,ovector_base\>
      - \c def_auto_grad_t - \ref simple_grad\<param_t,func_t,ovector_base\>

      \comment
      \bug The examples in ex_mmin.cpp fail because they always
      iterate until ntrial is maxxed.  (1/18/09 - They don't fail
      right now, and I think this was either fixed by properly ensuring
      that the derivative evaluation was sufficiently accurate, or
      by adding the it_info variable and ensuring iterate() doesn't
      unreasonably return a non-zero value. )
      \endcomment
  */
  template<class param_t, class func_t=multi_funct<param_t>, 
    class vec_t=ovector_base, class alloc_vec_t=ovector,
    class alloc_t=ovector_alloc, 
    class dfunc_t=grad_funct<param_t,ovector_base>, 
    class auto_grad_t=gradient<param_t,func_t,ovector_base>,
    class def_auto_grad_t=simple_grad<param_t,func_t,ovector_base> > 
    class gsl_mmin_conf : 
  public gsl_mmin_base<param_t,func_t,vec_t,alloc_vec_t,alloc_t,dfunc_t,
    auto_grad_t,def_auto_grad_t> {
    
#ifndef DOXYGEN_INTERNAL

  protected:

  /// \name The original variables from the GSL state structure
  //@{
  /// Iteration number
  int iter;
  /// Stepsize
  double step;
  /// Tolerance
  double tol;
  /// Desc
  gsl_vector *x1;
  /// Desc
  gsl_vector *dx1;
  /// Desc
  gsl_vector *x2;
  /// Desc
  double pnorm;
  /// Desc
  gsl_vector *p;
  /// Desc
  double g0norm;
  /// Desc
  gsl_vector *g0;
  //@}

  /// \name Store the arguments to set() so we can use them for iterate()
  //@{
  /// Proposed minimum
  gsl_vector *ugx;
  /// Gradient
  gsl_vector *ugg;
  /// Proposed step
  gsl_vector *udx;
  /// Desc
  double it_min;
  //@}

  /// Temporary vector
  alloc_vec_t avt5; 
  /// Temporary vector
  alloc_vec_t avt6;
  /// Temporary vector
  alloc_vec_t avt7; 
  /// Temporary vector
  alloc_vec_t avt8;
      
#endif
    
  public:
      
  gsl_mmin_conf() {
    lmin_tol=1.0e-4;
    this->tolf=1.0e-4;
    step_size=0.01;
  }
    
  virtual ~gsl_mmin_conf() {}

  /// Tolerance for the line minimization (default \f$ 10^{-4} \f$)
  double lmin_tol;

  /// Size of the initial step (default 0.01)
  double step_size;
    
  /// Perform an iteration
  virtual int iterate() {
	
    this->last_conv=0;
	
    if (this->dim==0) {
      O2SCL_ERR2_RET("Memory not allocated in ",
		     "gsl_mmin_conf::iterate().",gsl_efailed);
    }
	
    gsl_vector *x=ugx;
    gsl_vector *gradient=ugg;
    gsl_vector *dx=udx;

    double fa = it_min, fb, fc;
    double dir;
    double stepa = 0.0, stepb, stepc = step;

    double g1norm;
    double pg;

    if (pnorm == 0.0 || g0norm == 0.0) {
      gsl_vector_set_zero(dx);
      this->last_conv=gsl_enoprog;
      O2SCL_CONV2_RET("Either pnorm or g0norm vanished in ",
		      "in gsl_mmin_conf::iterate().",
		      gsl_enoprog,this->err_nonconv);
    }

    /* Determine which direction is downhill, +p or -p */
	
    gsl_blas_ddot(p, gradient, &pg);

    dir = (pg >= 0.0) ? +1.0 : -1.0;

    /* Compute new trial point at x_c= x - step * p, where p is the
       current direction */

    this->take_step (x, p, stepc, dir / pnorm, x1, dx);

    /* Evaluate function and gradient at new point xc */
	
    for(size_t i=0;i<this->dim;i++) avt5[i]=gsl_vector_get(x1,i);
    (*this->func)(this->dim,avt5,fc,*this->params);

    if (fc < fa) {

      /* Success, reduced the function value */
      step = stepc * 2.0;
      it_min = fc;
      gsl_vector_memcpy(x,x1);
	  
      for(size_t i=0;i<this->dim;i++) avt6[i]=gsl_vector_get(x1,i);
      if (this->grad_given) {
	(*this->grad)(this->dim,avt6,avt7,*this->params);
      } else {
	(*this->agrad)(this->dim,avt6,avt7,*this->params);
      }
      for(size_t i=0;i<this->dim;i++) gsl_vector_set(gradient,i,avt7[i]);

      return gsl_success;
    }

    /* Do a line minimization in the region (xa,fa) (xc,fc) to find an
       intermediate (xb,fb) satisifying fa > fb < fc.  Choose an initial
       xb based on parabolic interpolation 
    */
    this->intermediate_point(x,p,dir / pnorm,pg,stepa,stepc,fa,fc,x1,
			     dx1,gradient,&stepb,&fb);
	
    if (stepb == 0.0) {
      this->last_conv=gsl_enoprog;
      O2SCL_CONV2_RET("Variable stepb vanished in ",
		      "in gsl_mmin_conf::iterate().",
		      gsl_enoprog,this->err_nonconv);
    }
	
    this->minimize(x,p,dir / pnorm,stepa,stepb,stepc,fa,fb,fc,tol,
		   x1,dx1,x2,dx,gradient,&step,&it_min,&g1norm);
	
    gsl_vector_memcpy (x, x2);

    /* Choose a new conjugate direction for the next step */
    iter=(iter+1) % x->size;

    if (iter==0) {

      gsl_vector_memcpy (p, gradient);
      pnorm=g1norm;

    } else {

      /* p' = g1 - beta * p */
      double beta = -pow (g1norm / g0norm, 2.0);
      gsl_blas_dscal (-beta, p);
      gsl_blas_daxpy (1.0, gradient, p);
      pnorm=gsl_blas_dnrm2 (p);
    }
	
    g0norm=g1norm;
    gsl_vector_memcpy (g0, gradient);

    return gsl_success;
  }
    
  /// Return string denoting type("gsl_mmin_conf")
  virtual const char *type() { return "gsl_mmin_conf";}

  /** 
      \brief Allocate the memory

      \future Use a gsl_alloc_arrays() like function for this (but
      keep in mind these are calloc, not malloc statements
  */
  virtual int allocate(size_t n) {

    x1=gsl_vector_calloc(n);
    if (x1==0) {
      O2SCL_ERR_RET("Failed to allocate x1 in gsl_mmin_conf::allocate().",
		    gsl_enomem);
    }
	
    dx1=gsl_vector_calloc(n);
    if (dx1==0) {
      gsl_vector_free(x1);
      O2SCL_ERR_RET("Failed to allocate dx1 in gsl_mmin_conf::allocate().",
		    gsl_enomem);
    }
	
    x2=gsl_vector_calloc(n);
    if (x2==0) {
      gsl_vector_free(dx1);
      gsl_vector_free(x1);
      O2SCL_ERR_RET("Failed to allocate x2 in gsl_mmin_conf::allocate().",
		    gsl_enomem);
    }

    p=gsl_vector_calloc(n);
    if (p==0) {
      gsl_vector_free(x2);
      gsl_vector_free(dx1);
      gsl_vector_free(x1);
      O2SCL_ERR_RET("Failed to allocate p in gsl_mmin_conf::allocate().",
		    gsl_enomem);
    }

    g0=gsl_vector_calloc(n);
    if (g0==0) {
      gsl_vector_free(p);
      gsl_vector_free(x2);
      gsl_vector_free(dx1);
      gsl_vector_free(x1);
      O2SCL_ERR_RET("Failed to allocate g0 in gsl_mmin_conf::allocate().",
		    gsl_enomem);
    }

    udx=gsl_vector_calloc(n);
    if (udx==0) {
      gsl_vector_free(g0);
      gsl_vector_free(p);
      gsl_vector_free(x2);
      gsl_vector_free(dx1);
      gsl_vector_free(x1);
      O2SCL_ERR_RET("Failed to allocate udx in gsl_mmin_conf::allocate().",
		    gsl_enomem);
    }

    ugg=gsl_vector_calloc(n);
    if (ugg==0) {
      gsl_vector_free(udx);
      gsl_vector_free(g0);
      gsl_vector_free(p);
      gsl_vector_free(x2);
      gsl_vector_free(dx1);
      gsl_vector_free(x1);
      O2SCL_ERR_RET("Failed to allocate ugg in gsl_mmin_conf::allocate().",
		    gsl_enomem);
    }

    ugx=gsl_vector_calloc(n);
    if (ugx==0) {
      gsl_vector_free(ugg);
      gsl_vector_free(udx);
      gsl_vector_free(g0);
      gsl_vector_free(p);
      gsl_vector_free(x2);
      gsl_vector_free(dx1);
      gsl_vector_free(x1);
      O2SCL_ERR_RET("Failed to allocate ugx in gsl_mmin_conf::allocate().",
		    gsl_enomem);
    }

    this->ao.allocate(avt5,n);
    this->ao.allocate(avt6,n);
    this->ao.allocate(avt7,n);
    this->ao.allocate(avt8,n);
	
    this->base_allocate(n);

    return gsl_success;

  }
    
  /// Free the allocated memory
  virtual int free() {
    gsl_vector_free(ugx);
    gsl_vector_free(ugg);
    gsl_vector_free(udx);
    gsl_vector_free(g0);
    gsl_vector_free(p);
    gsl_vector_free(x2);
    gsl_vector_free(dx1);
    gsl_vector_free(x1);

    this->ao.free(avt5);
    this->ao.free(avt6);
    this->ao.free(avt7);
    this->ao.free(avt8);
	
    this->base_free();

    return 0;
  }
      
  /// Reset the minimizer to use the current point as a new starting point
  int restart() {
    iter=0;
    return gsl_success;
  }

  /// Set the function and initial guess
  virtual int set(vec_t &x, double u_step_size, double tol_u,
		  func_t &ufunc, param_t &pa) {

    if (this->dim==0) {
      O2SCL_ERR_RET("Memory not allocated in set().",gsl_efailed);
    }
    
    this->base_set(ufunc,pa,*this->agrad);
    for(size_t i=0;i<this->dim;i++) gsl_vector_set(ugx,i,x[i]);
	
    iter=0;
    step=u_step_size;
    tol=tol_u;

    /// Evaluate the function and its gradient
    ufunc(this->dim,x,it_min,pa);
    this->agrad->set_function(ufunc);
    (*this->agrad)(this->dim,x,avt8,*this->params);
    for(size_t i=0;i<this->dim;i++) {
      gsl_vector_set(ugg,i,avt8[i]);
    }
	
    /* Use the gradient as the initial direction */
	
    gsl_vector_memcpy(p,ugg);
    gsl_vector_memcpy(g0,ugg);
	
    double gnorm=gsl_blas_dnrm2(ugg);
	
    pnorm=gnorm;
    g0norm=gnorm;
	
    return gsl_success;
  }

  /// Set the function and initial guess
  virtual int set_de(vec_t &x, double u_step_size, double tol_u,
		     func_t &ufunc, dfunc_t &udfunc, param_t &pa) {

    if (this->dim==0) {
      O2SCL_ERR_RET("Memory not allocated in set().",gsl_efailed);
    }
	
    this->base_set_de(ufunc,udfunc,pa);
    for(size_t i=0;i<this->dim;i++) gsl_vector_set(ugx,i,x[i]);
	
    iter=0;
    step=u_step_size;
    tol=tol_u;

    // Evaluate the function and its gradient
    ufunc(this->dim,x,it_min,pa);
    udfunc(this->dim,x,avt8,*this->params);
    for(size_t i=0;i<this->dim;i++) {
      gsl_vector_set(ugg,i,avt8[i]);
    }
	
    /* Use the gradient as the initial direction */
	
    gsl_vector_memcpy(p,ugg);
    gsl_vector_memcpy(g0,ugg);
	
    double gnorm=gsl_blas_dnrm2(ugg);
	
    pnorm=gnorm;
    g0norm=gnorm;

    return gsl_success;
  }

  /** \brief Calculate the minimum \c min of \c func w.r.t the
      array \c x of size \c nvar.
  */
  virtual int mmin(size_t nn, vec_t &xx, double &fmin, param_t &pa,
		   func_t &ufunc) {

    if (nn==0) {
      O2SCL_ERR2_RET("Tried to minimize over zero variables ",
		     " in gsl_mmin_conf::mmin().",gsl_einval);
    }

    int xiter=0, status;
	
    allocate(nn);
	
    set(xx,step_size,lmin_tol,ufunc,pa);
	
    do {

      xiter++;
	  
      status=iterate();
	  
      if (status) {
	break;
      }
	  
      // Equivalent to gsl_multimin_test_gradient with
      // additional code to print out present iteration

      double norm=gsl_blas_dnrm2(ugg);
	  
      if(this->verbose>0) {
	this->print_iter(nn,*((ovector *)ugx),it_min,xiter,
			 norm,this->tolf,type());
      }
    
      if (norm<this->tolf) status=gsl_success;
      else status=gsl_continue;

    } while (status==gsl_continue && xiter < this->ntrial);

    for(size_t i=0;i<nn;i++) xx[i]=gsl_vector_get(ugx,i);
    fmin=it_min;
	
    free();
    this->last_ntrial=xiter;
	
    if (status==gsl_continue && xiter==this->ntrial) {
      this->last_conv=gsl_emaxiter;
      O2SCL_CONV_RET("Too many iterations in gsl_mmin_conf::mmin().",
		     gsl_emaxiter,this->err_nonconv);
    }

    return status;
  }

  /** \brief Calculate the minimum \c min of \c func w.r.t the
      array \c x of size \c nvar.
  */
  virtual int mmin_de(size_t nn, vec_t &xx, double &fmin, param_t &pa,
		      func_t &ufunc, dfunc_t &udfunc) {

    if (nn==0) {
      O2SCL_ERR2_RET("Tried to minimize over zero variables ",
		     " in gsl_mmin_conf::mmin_de().",gsl_einval);
    }

    int xiter=0, status;
	
    allocate(nn);
	
    set_de(xx,step_size,lmin_tol,ufunc,udfunc,pa);
	
    do {
      xiter++;
	  
      status=iterate();
	  
      if (status) {
	break;
      }
	  
      // Equivalent to gsl_multimin_test_gradient with
      // additional code to print out present iteration

      double norm=gsl_blas_dnrm2(ugg);
	  
      if(this->verbose>0) {
	this->print_iter(nn,*((ovector *)ugx),it_min,xiter,
			 norm,this->tolf,type());
      }
    
      if (norm<this->tolf) status=gsl_success;
      else status=gsl_continue;

    }
    while (status==gsl_continue && xiter < this->ntrial);
     
    for(size_t i=0;i<nn;i++) xx[i]=gsl_vector_get(ugx,i);
    fmin=it_min;
	
    free();
    this->last_ntrial=xiter;

    if (status==gsl_continue && xiter==this->ntrial) {
      this->last_conv=gsl_emaxiter;
      O2SCL_CONV_RET("Too many iterations in gsl_mmin_conf::mmin().",
		     gsl_emaxiter,this->err_nonconv);
    }

    return status;
  }

  };

  /**
     \brief An array version of gsl_mmin_conf
      
      Default template arguments
      - \c param_t - no default
      - \c func_t - \ref multi_vfunct\<param_t\>
      - \c vec_t - <tt>double [nv]</tt>
      - \c alloc_vec_t - <tt>double [nv]</tt>
      - \c alloc_t - \ref array_alloc\<double[nv]\>
      - \c dfunc_t - \ref grad_vfunct\<param_t,nv\>
      - \c auto_grad_t - \ref gradient_array
      \<param_t,multi_vfunct\<param_t,nv\>,nv\>
      - \c def_auto_grad_t - \ref simple_grad
      \<param_t,multi_vfunct\<param_t,nv\>,nv\>
      
     \comment
     Doxygen has had trouble parsing this template, so I have
     simplified it slightly for the documentation below.
     \endcomment
  */
#ifdef DOXYGENP
  template<class param_t, size_t nv> class gsl_mmin_conf_array :
  public gsl_mmin_conf<param_t,func_t,vec_t,alloc_vec_t,
    alloc_t,dfunc_t,auto_grad_t,def_auto_grad_t> {
  };
#else
  template<class param_t, size_t nv> class gsl_mmin_conf_array :
  public gsl_mmin_conf<param_t,multi_vfunct<param_t,nv>,
    double[nv],double[nv],array_alloc<double[nv]>,grad_vfunct<param_t,nv>,
    gradient_array<param_t,multi_vfunct<param_t,nv>,nv>,
    simple_grad_array<param_t,multi_vfunct<param_t,nv>,nv> > {
  };
#endif
  

#ifndef DOXYGENP
}
#endif

#endif
