/*
  -------------------------------------------------------------------

  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner

  This file is part of O2scl.

  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_CONTOUR_H
#define O2SCL_CONTOUR_H

#include <cmath>
#include <gsl/gsl_math.h>
#include <o2scl/smart_interp.h>
#include <o2scl/omatrix_tlate.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief A contour line 

      The contour lines generated by the \ref contour class are given
      as objects of this type.

      \future Write an I/O object for contour lines
      \future Make this a subclass of \ref contour .
   */
  class contour_line {
  public:
    /// The contour level
    double level;
    /// The line x coordinates
    ovector x;
    /// The line y coordinates
    ovector y;

    /// Create an empty line
    contour_line() {
    }

    /// Copy constructor for STL allocator
    contour_line(const contour_line &c) {
      level=c.level;
      x=c.x;
      y=c.y;
    }
  };

  /** 
      \brief Edges for the contour class
      
      The edge crossings generated by the \ref contour class are given
      as objects of this type.

      \future Write an I/O object for edge crossings
      \future Make this a subclass of \ref contour .
  */
  class edge_crossings {

  public:

    /// Edge status
    omatrix_int status;
    /// Edge values
    omatrix values;
  };

  /** 
      \brief Calculate contour lines from a two-dimensional data set

      \b Basic \b Usage
      
      - Specify the data as a two-dimensional square grid of "heights"
      with set_data().
      - Specify the contour levels with set_levels().
      - Compute the contours with calc_contours()

      The contours are generated as a series of x- and y-coordinates,
      defining a line. If the contour is closed, then the first and
      the last set of coordinates will be equal. 

      The storage of the matrix to be specified in the function
      set_data() and this function is designed to follow the format:
      \f[
      \begin{array}{cccc}
      & x_0 & x_1 & x_2 \\
      y_0 & M_{00} & M_{01} & M_{02} \\
      y_1 & M_{10} & M_{11} & M_{12} \\
      y_2 & M_{20} & M_{21} & M_{22}
      \end{array}
      \f]
      thus the matrix should be \c M[i][j] where \c i is the y index
      and \c j is the row index. (See also the discussion in the
      User's guide in the section called \ref rowcol_subsect.) 

      The data is copied by set_data(), so changing the data will not
      change the contours unless set_data() is called again. The
      functions set_levels() and calc() can be called several times
      for the same data without calling set_data() again.

      Note that in order to simplify the algorithm for computing
      contour lines, the calc_contours() function will adjust the
      user-specified contour levels slightly in order to ensure that
      no contour line passes exactly through any data point on the
      grid. The contours are adjusted by multiplying the original
      contour level by 1 plus a small number (\f$ 10^{-8} \f$ by
      default), which is specified in \ref lev_adjust.

      Linear interpolation is used to decide whether or not a line
      segment and a contour cross. This choice is intentional, since
      (in addition to making the algorithm much simpler) it is the
      user (and not the class) which is likely best able to refine the
      data. In case a simple refinement scheme is desired, the method
      regrid_data() is provided which refines the data for any
      interpolation type.

      Since linear interpolation is used, the contour calculation
      implicitly assumes that there is not more than one intersection
      of any contour level with any line segment. For contours which
      do not close inside the region of interest, the results will
      always end at either the minimum or maximum values of the x or y
      grid points (no extrapolation is ever done). Note also that the
      points defining the contour are not necessarily equally spaced,
      but two neighboring points will never be farther apart than the
      distance across opposite corners of one cell in the grid.

      \b The \b Algorithm:

      This works by viewing the data as defining a square
      two-dimensional grid. The function calc_contours() exhaustively
      enumerates every line segment in the grid which involves a level
      crossing and then organizes the points defined by the
      intersection of a line segment with a level curve into a full
      contour.

      \future Rewrite the code which adjusts the contour levels 
      to ensure contours don't go through the data to adjust the
      internal copy of the data instead. 

      \future It would be nice to have a function which creates a set
      of closed regions to fill which represent the data. However,
      this likely requires a completely new algorithm, because it's
      not easy to simply close the contours already generated by the
      calc_contours() function. There are, for example, several cases
      which are difficult to handle, such as filling a region in
      between several closed contours.

  */
  class contour {
    
  public:
    
    contour();

    ~contour();

    /// \name Basic usage
    //@{

    /** 
	\brief Set the data 

	The types \c vec_t and \c mat_t can be any types which have \c
	operator[] and \c operator[][] for array and matrix indexing.
	
	Note that this method copies all of the user-specified data to
	local storage so that changes in the data after calling this
	function will not be reflected in the contours that are
	generated.
    */
    template<class vec_t, class mat_t>
      int set_data(size_t sizex, size_t sizey, const vec_t &x_fun, 
		   const vec_t &y_fun, const mat_t &udata) {
      
      if (sizex<2 || sizey<2) { 
	O2SCL_ERR_RET("Not enough data (must be at least 2x2) in set_data().",
		    gsl_einval);
      }
      
      nx=sizex;
      ny=sizey;
      xfun.allocate(nx);
      yfun.allocate(ny);
      data.allocate(ny,nx);
      for(int i=0;i<nx;i++) xfun[i]=x_fun[i];
      for(int i=0;i<ny;i++) yfun[i]=y_fun[i];
      for(int i=0;i<nx;i++) {
	for(int j=0;j<ny;j++) {
	  data[j][i]=udata[j][i];
	}
      }

      return check_data();
    }

    /** 
	\brief Set the contour levels 
	
	This is separate from the function calc_contours() so that
	the user can compute the contours for different data sets using
	the same levels
    */
    template<class vec_t> int set_levels(size_t nlevels, vec_t &ulevels) {
      nlev=nlevels;
      levels.allocate(nlevels);
      for(size_t i=0;i<nlevels;i++) {
	levels[i]=ulevels[i];
      }
      levels_set=true;
      return 0;
    }

    /** 
	\brief Calculate the contours 

	The function calc_contours() returns the total number of
	contours found.  Since there may be more than one disconnected
	contours for the same contour level, or no contours for a
	given level, the total number of contours may be less than or
	greater than the number of levels given by set_levels().

	If an error occurs, zero is returned.
    */
    int calc_contours(std::vector<contour_line> &clines,
		      bool debug=false);

    //@}

    /// \name Regrid function
    //@{
    /** 
	\brief Regrid the data 
	
	Use interpolation to refine the data set. This can be called
	before calc_contours() in order to make the contour levels
	smoother by providing a smaller grid size. If the original
	number of data points is \f$ (\mathrm{nx},\mathrm{ny}) \f$,
	then the new number of data points is
	\f[
	(\mathrm{xfact}~(\mathrm{nx}-1)+1,
	\mathrm{yfact}~(\mathrm{ny}-1)+1)
	\f]
	The parameters \c xfact and \c yfact must both be larger than 
	zero and they cannot both be 1.
    */
    int regrid_data(size_t xfact, size_t yfact,
		    base_interp_mgr<ovector_const_view> &bim1,
		    base_interp_mgr<ovector_const_subvector> &bim2);
    //@}

    /// \name Obtain internal data
    //@{
    /** 
	\brief Get the data 
	
	This is useful to see how the data has changed after
	a call to regrid_data().

	\future There is probably a better way than returning
	pointers to the internal data.
    */
    int get_data(size_t &sizex, size_t &sizey, ovector *&x_fun, 
		 ovector *&y_fun, omatrix *&udata) {
      if (nx==0) {
	O2SCL_ERR_RET("Data not set in calc().",gsl_einval);
      }
      sizex=nx;
      sizey=ny;
      x_fun=&xfun;
      y_fun=&yfun;
      udata=&data;
      return 0;
    }

    /// Return the edges
    int get_edges(std::vector<edge_crossings> &rt_edges,
		  std::vector<edge_crossings> &bm_edges) {
      rt_edges=red;
      bm_edges=bed;
      return 0;
    }

    /// Print out the edges to cout
    int print_edges(edge_crossings &right,
		    edge_crossings &bottom);

    //@}

    /// Verbosity parameter
    int verbose;

    /// (default \f$ 10^{-8} \f$)
    double lev_adjust;
    
#ifndef DOXYGEN_INTERNAL

  protected:

    /// \name Edge direction
    //@{
    static const int dright=0;
    static const int dbottom=1;
    //@}

    /// \name Edge status
    //@{
    static const int empty=0;
    static const int edge=1;
    static const int contourp=2;
    static const int endpoint=3;
    //@}

    /// \name Edge found or not found
    //@{
    static const int efound=1;
    static const int enot_found=0;
    //@}

    /// \name User-specified data
    //@{
    int nx, ny;
    ovector xfun, yfun;
    omatrix data;
    //@}

    /// \name User-specified contour levels
    //@{
    int nlev;
    ovector levels;
    bool levels_set;
    //@}

    /// Right edge list
    std::vector<edge_crossings> red;
    /// Bottom edge list
    std::vector<edge_crossings> bed;

    /// Find next point starting from a point on a right edge
    int find_next_point_right(int j, int k, int &jnext, int &knext, 
			      int &dir_next, int nsw,
			      edge_crossings &right,
			      edge_crossings &bottom);
    
    /// Find next point starting from a point on a bottom edge
    int find_next_point_bottom(int j, int k, int &jnext, int &knext, 
			       int &dir_next, int nsw,
			       edge_crossings &right,
			       edge_crossings &bottom);

    /// Find all of the intersections of the edges with the contour level
    int find_intersections(size_t ilev, double &level,
			   edge_crossings &right, edge_crossings &bottom);

    /// Interpolate all right edge crossings 
    int right_edges(double level, o2scl::sm_interp *si,
		    edge_crossings &right);
    
    /// Interpolate all bottom edge crossings
    int bottom_edges(double level, o2scl::sm_interp *si,
		     edge_crossings &bottom);

    /// Create a contour line from a starting edge
    int process_line(int j, int k, int dir, ovector &x, ovector &y, 
		     bool first, edge_crossings &right,
		     edge_crossings &bottom);

    /// Check to ensure the x- and y-arrays are monotonic
    int check_data();

#endif

  };
  
#ifndef DOXYGENP
}
#endif

#endif


