/*
  -------------------------------------------------------------------
  
  Copyright (C) 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_CONVERT_UNITS_H
#define O2SCL_CONVERT_UNITS_H

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <map>
#include <o2scl/err_hnd.h>
#include <o2scl/file_format.h>
#include <o2scl/misc.h>
#include <o2scl/constants.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Convert units

      Allow the user to convert between two different units after
      specifying a conversion factor. This class will also
      automatically combine two conversion factors to create a new
      unit conversion.

      Conversions are performed by the \ref convert() function and
      the conversion factors must be specified beforehand using the 
      \ref insert_cache() function.

      Example:
      \code
      convert_units cu;
      cu.insert_cache("in","cm",2.54);
      cout << "12 in is " << cu.convert("in","cm",12.0) << " cm. " << endl;
      \endcode
      
      \future Ideally, a real C++ API for the GNU units command,
      or some interface to the Boost units library would
      probably be better.
  */
  class convert_units {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// The type for caching unit conversions
    typedef struct {
      /// The input unit
      std::string f;
      /// The output unit
      std::string t;
      /// The conversion factor
      double c;
    } unit_t;

    /// The cache where unit conversions are stored
    std::map<std::string,unit_t,string_comp> mcache;
    
    /// The iterator type
    typedef std::map<std::string,unit_t,string_comp>::iterator miter;
      
#endif

  public:

    /// Verbosity (default 0)
    int verbose;

    convert_units() {
      verbose=0;
    }

    virtual ~convert_units() {}

    /** \brief Return the value \c val after converting using units \c
	from and \c to
    */
    virtual double convert(std::string from, std::string to, double val) {

      // Look in cache for conversion
      std::string both=from+","+to;
      miter m3=mcache.find(both);
      if (m3!=mcache.end()) {
	return val*m3->second.c;
      }

      // Look in cache for reverse conversion
      std::string both2=to+","+from;
      m3=mcache.find(both2);
      if (m3!=mcache.end()) {
	return val/m3->second.c;
      }

      // Look for combined conversions
      for(miter m=mcache.begin();m!=mcache.end();m++) {
	if (m->second.f==from) {
	  std::string b=m->second.t+","+to;
	  miter m2=mcache.find(b);
	  if (m2!=mcache.end()) {
	    if (verbose>0) {
	      std::cout << "Using conversions: " << m->second.f << " , "
			<< m->second.t << std::endl;
	      std::cout << " (1)          and: " << m2->second.f << " , "
			<< m2->second.t << std::endl;
	    }
	    return val*m->second.c*m2->second.c;
	  }
	  std::string b2=to+","+m->second.t;
	  miter m4=mcache.find(b2);
	  if (m4!=mcache.end()) {
	    if (verbose>0) {
	      std::cout << "Using conversions: " << m->second.f << " , "
			<< m->second.t << std::endl;
	      std::cout << " (2)          and: " << m4->second.f << " , "
			<< m4->second.t << std::endl;
	    }
	    return val*m->second.c/m4->second.c;
	  }
	} else if (m->second.t==from) {
	  std::string b=m->second.f+","+to;
	  miter m2=mcache.find(b);
	  if (m2!=mcache.end()) {
	    if (verbose>0) {
	      std::cout << "Using conversions: " << m->second.f << " , "
			<< m->second.t << std::endl;
	      std::cout << " (3)          and: " << m2->second.f << " , "
			<< m2->second.t << std::endl;
	    }
	    return val/m->second.c*m2->second.c;
	  }
	} else if (m->second.f==to) {
	  std::string b=m->second.t+","+from;
	  miter m2=mcache.find(b);
	  if (m2!=mcache.end()) {
	    if (verbose>0) {
	      std::cout << "Using conversions: " << m->second.f << " , "
			<< m->second.t << std::endl;
	      std::cout << " (4)          and: " << m2->second.f << " , "
			<< m2->second.t << std::endl;
	    }
	    return val/m->second.c/m2->second.c;
	  }
	  std::string b2=from+","+m->second.t;
	  miter m4=mcache.find(b2);
	  if (m4!=mcache.end()) {
	    if (verbose>0) {
	      std::cout << "Using conversions: " << m->second.f << " , "
			<< m->second.t << std::endl;
	      std::cout << " (5)          and: " << m4->second.f << " , "
			<< m4->second.t << std::endl;
	    }
	    return val/m->second.c*m4->second.c;
	  }
	} else if (m->second.t==to) {
	  std::string b=m->second.f+","+from;
	  miter m2=mcache.find(b);
	  if (m2!=mcache.end()) {
	    if (verbose>0) {
	      std::cout << "Using conversions: " << m->second.f << " , "
			<< m->second.t << std::endl;
	      std::cout << " (6)          and: " << m2->second.f << " , "
			<< m2->second.t << std::endl;
	    }
	    return val*m->second.c/m2->second.c;
	  }
	}
      }
      
      O2SCL_ERR("Conversion not found in convert_units::convert().",
		gsl_efailed);

      return 0.0;
    }

    /// Manually insert a unit conversion into the cache
    int insert_cache(std::string from, std::string to, double conv) {
      unit_t ut;
      ut.f=from;
      ut.t=to;
      ut.c=conv;
      std::string both=from+","+to;
      mcache.insert(make_pair(both,ut));
      return 0;
    }
    
    /// Print the present unit cache
    int print_cache() {
      miter m;
      std::cout << "Unit cache: " << std::endl;
      for (m=mcache.begin();m!=mcache.end();m++) {
	std::cout.setf(std::ios::left);
	std::cout.width(20);
	std::cout << m->second.f << " ";
	std::cout.width(20);
	std::cout << m->second.t << " ";
	std::cout.unsetf(std::ios::left);
	std::cout.precision(10);
	std::cout << m->second.c << std::endl;
      }
      return 0;
    }

    /// Add conversion factors for energy equivalents
    int energy_conv() {
      
      // Entries derived from http://physics.nist.gov/cuu/Constants.
      // The NIST website uses 'h' to convert instead of hbar,
      // so we have to add a lot of factors of 2 pi
      
      insert_cache("J","m^-1",5.03411747e24*2.0*o2scl_const::pi);
      insert_cache("J","km^-1",5.03411747e27*2.0*o2scl_const::pi);
      insert_cache("J","cm^-1",5.03411747e22*2.0*o2scl_const::pi);
      insert_cache("J","fm^-1",5.03411747e9*2.0*o2scl_const::pi);
      insert_cache("J","s^-1",1.509190450e33*2.0*o2scl_const::pi);
      insert_cache("J","K",1.0e7/gsl_cgs::boltzmann);
      insert_cache("J","eV",6.24150965e18);
      insert_cache("J","MeV",6.24150965e12);
      insert_cache("J","u",6.70053641e9);
      insert_cache("J","erg",1.0e7);
      insert_cache("J","kg",1.0/gsl_mks::speed_of_light/
		   gsl_mks::speed_of_light);
      insert_cache("J","g",1.0e3/gsl_mks::speed_of_light/
		   gsl_mks::speed_of_light);
      
      return 0;
    }
      
  };

#ifndef DOXYGENP
}
#endif

#endif
