/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_BOSON_H
#define O2SCL_BOSON_H

#include <string>
#include <iostream>
#include <fstream>
#include <math.h>
#include <o2scl/constants.h>
#include <o2scl/mroot.h>
#include <o2scl/inte.h>
#include <o2scl/part.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Boson class [abstract base]

      For bosons: \n
      - if either nu or mu is greater than ms, then they are taken to
      be equal to ms \n
      - All contributions from any type of condensate are ignored.

      \htmlonly
      This Mathematica notebook contains the series expansions
      for the bosonic integrals.
      functions.
      <a href="../extras/boson.nb">
      boson.nb</a>, and
      <a href="../extras/boson.pdf">
      boson.pdf</a>.
      \endhtmlonly
      \latexonly
      This Mathematica notebook contains the series expansions
      for the bosonic integrals.
      functions.
      \begin{verbatim}
      doc/o2scl/extras/boson.nb
      doc/o2scl/extras/boson.pdf
      \end{verbatim}
      \endlatexonly
  */
  class boson : virtual public part {

  public:

    /** 
	\brief The condensate

	The condensate variable is mostly ignored by class boson
	and its descendants, and is provided for user storage.
    */
    double co;

    /// Create a boson with mass \c m and degeneracy \c g 
    boson(double m=0.0, double g=0.0);

    virtual int calc_mu(const double temper)=0;
    virtual int calc_density(const double temper)=0;
    virtual int pair_mu(const double temper)=0;
    virtual int pair_density(const double temper)=0;

    /** 
	\brief Calculate properties of massless bosons 
	
	The expressions used are exact. The chemical
	potentials are ignored and the scalar density is
	set to zero
    */
    virtual int massless_calc_mu(const double temper);
    
    /// Return string denoting type ("boson")
    virtual const char *type() { return "boson"; }

#ifndef DOXYGEN_INTERNAL

    //friend class io_tlate<boson>;

#endif

  protected:

  };

  /*
    template<> int io_tlate<boson>::input
    (cinput *co, in_file_format *ins, boson *b);
    template<> int io_tlate<boson>::output
    (coutput *co, out_file_format *outs, boson *b);
    template<> const char *io_tlate<boson>::type();
    
    typedef io_tlate<boson> boson_io_type;
  */

#ifndef DOXYGENP
}
#endif

#endif
