/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_DERIV_PART_H
#define O2SCL_DERIV_PART_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <o2scl/boson.h>
#include <o2scl/fermion.h>
#include <o2scl/classical.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Storage for deriviatives wrt \f$ \mu \f$ and T.

      The variables \c dndmu, \c dndT, and \c dsdT correspond
      to 
      \f[
      \left(\frac{d n}{d \mu}\right)_{T}, \quad
      \left(\frac{d n}{d T}\right)_{\mu}, \quad \mathrm{and} \quad
      \left(\frac{d s}{d T}\right)_{\mu}
      \f]
      respectively.

      All other derivatives can be expressed simply in terms of these
      three. 

      \hline

      <b>Derivatives wrt to chemical potential and temperature:</b>

      There is a Maxwell relation
      \f[
      \left(\frac{d s}{d \mu}\right)_T =
      \left(\frac{d n}{d T}\right)_{\mu}
      \f]
      The pressure derivatives are trivial
      \f[
      \left(\frac{d P}{d \mu}\right)_{T}=n, \quad
      \left(\frac{d P}{d T}\right)_{\mu}=s
      \f]
      The energy density derivatives are related through the 
      thermodynamic identity:
      \f[
      \left(\frac{d \varepsilon}{d \mu}\right)_{T}=
      \mu \left(\frac{d n}{d \mu}\right)_{T}+
      T \left(\frac{d s}{d \mu}\right)_{T}
      \f]
      \f[
      \left(\frac{d \varepsilon}{d T}\right)_{\mu}=
      \mu \left(\frac{d n}{d T}\right)_{\mu}+
      T \left(\frac{d s}{d T}\right)_{\mu}
      \f]

      \hline

      <b>Other derivatives:</b>

      Note that the derivative of the entropy with respect to the
      temperature above is not the specific heat, \f$ c_V \f$.
      The specific heat is
      \f[
      C_V = \frac{T}{N} \left( \frac{\partial S}{\partial T} \right)_{V,N}
      = \frac{T}{n} \left( \frac{\partial s}{\partial T} \right)_{V,n}
      \f] 
      To compute the specific heat in terms of the derivatives above,
      note that the descendants of deriv_part provide all of the
      thermodynamic functions in terms of \f$ \mu, V \f$ and \f$ T
      \f$, so we have
      \f[
      s=s(\mu,V,T) \quad \mathrm{and} \quad n=n(\mu,V,T) \, .
      \f]
      We can then construct a function
      \f[
      s=s[\mu(n,V,T),V,T]
      \f]
      and then write the required derivative directly
      \f[
      \left(\frac{\partial s}{\partial T}\right)_{n,V} =
      \left(\frac{\partial s}{\partial \mu}\right)_{T,V}
      \left(\frac{\partial \mu}{\partial T}\right)_{n,V} +
      \left(\frac{\partial s}{\partial T}\right)_{\mu,V} \, .
      \f]
      Now we use the identity
      \f[
      \left(\frac{\partial \mu}{\partial T}\right)_{n,V} = -
      \left(\frac{\partial n}{\partial T}\right)_{\mu,V} 
      \left(\frac{\partial n}{\partial \mu}\right)_{T,V}^{-1} \, ,
      \f]
      and the Maxwell relation above to give
      \f[
      C_V = \frac{T}{n}
      \left[ 
      \left(\frac{\partial s}{\partial T}\right)_{\mu,V}
      -\left(\frac{\partial n}{\partial T}\right)_{\mu,V}^2
      \left(\frac{\partial n}{\partial \mu}\right)_{T,V}^{-1}
      \right]
      \f]
      which expresses the specific heat in terms of the three
      derivatives which are given.

      Note that this is the specific heat per particle, and 
      has no units. If specific heat per unit volume is required,
      you must multiply by the number density. 

      No derivative with respect to the bare mass is given, since
      classes cannot know how to relate the effective mass to the
      bare mass. 

  */
  class deriv_part {
    
  public:
    
    /// Derivative of number density with respect to chemical potential
    double dndmu;

    /// Derivative of number density with respect to temperature
    double dndT;

    /// Derivative of entropy density with respect to temperature
    double dsdT;

    /// Derivative of number density with respect to the effective mass
    double dndm;

  };

  template<> int io_tlate<deriv_part>::input
    (cinput *co, in_file_format *ins, deriv_part *f);
  template<> int io_tlate<deriv_part>::output
    (coutput *co, out_file_format *ins, deriv_part *f);
  template<> const char *io_tlate<deriv_part>::type();

  typedef io_tlate<deriv_part> deriv_part_io_type;

#ifndef DOXYGENP
}
#endif

#endif
