/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_EFF_BOSON_H
#define O2SCL_EFF_BOSON_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/funct.h>
#include <o2scl/mm_funct.h>
#include <o2scl/root.h>
#include <o2scl/mroot.h>
#include <o2scl/inte.h>
#include <o2scl/boson.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/gsl_mroot_hybrids.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /**
     \brief Boson class from fitting method
     
     \todo Better documentation (see eff_fermion)
     \todo Remove the 'meth2' stuff
     \todo Remove static variables fix_density and stat_temper
  */
  class eff_boson : public boson {

  public:
    
    /// Create a boson with mass \c m and degeneracy \c g 
    eff_boson(double m=0.0, double g=0.0);

    virtual ~eff_boson();
  
    /** 
	\brief Load coefficients for finite-temperature approximation
	
	Presently acceptable values of \c fn are: \c boselat3 from
	Lattimer's notes \c bosejel21, \c bosejel22, \c bosejel34, and
	\c bosejel34cons from \ref Johns96.
    */
    static int load_coefficients(int ctype);
    /// A set of coefficients from Jim Lattimer 
    static const int cf_boselat3=1;
    /// A set of coefficients from \ref Johns96
    static const int cf_bosejel21=2;
    /// A set of coefficients from \ref Johns96
    static const int cf_bosejel22=3;
    /// A set of coefficients from \ref Johns96
    static const int cf_bosejel34=4;
    /** \brief The set of coefficients from \ref Johns96 which retains 
	better thermodynamic consistency
     */
    static const int cf_bosejel34cons=5;
  
    /** 
	\brief Calculate thermodynamic 
	properties as function of chemical potential
    */
    virtual int calc_mu(const double temper);

    /** 
	\brief Calculate thermodynamic 
	properties as function of density
    */
    virtual int calc_density(const double temper);

    /** \brief Calculate thermodynamic properties with antiparticles
	as function of chemical potential
    */
    virtual int pair_mu(const double temper);

    /** \brief Calculate thermodynamic properties with antiparticles
	as function of density
    */
    virtual int pair_density(const double temper);

    /** \brief Set the solver for use in calculating \f$ \psi \f$ 
     */
    int set_psi_root(root<double,funct<double> > &rp) {
      psi_root=&rp;
      return 0;
    }

    /** \brief Set the solver for use in calculating the chemical
	potential from the density 
    */
    int set_density_mroot(mroot<int ,mm_funct<int> > &rp) {
      density_mroot=&rp;
      return 0;
    }

    /** \brief Set the solver for use in calculating the chemical
	potential from the density (meth2=true) 
    */
    int set_meth2_root(root<int ,funct<int> > &rp) {
      meth2_root=&rp;
      return 0;
    }

    /** \brief The default solver for calc_density() and pair_density()
     */
    gsl_mroot_hybrids<int ,mm_funct<int> > def_density_mroot;

    /** \brief The default solver for \f$ \psi \f$
     */
    cern_mroot_root<double,funct<double> > def_psi_root;

    /** \brief The default solver for calc_density() and pair_density()
     */
    cern_mroot_root<int ,funct<int> > def_meth2_root;

    virtual const char *type() { return "eff_boson"; }

#ifndef DOXYGEN_INTERNAL

    friend class io_tlate<eff_boson>;

  protected:
  
    /// The coefficients
    static double **Pmnb;
    /// The number of coefficient rows
    static int sizem;
    /// The number of coefficient columns
    static int sizen;
    /// The parameter, \f$ a \f$
    static double parma;
    /// Temporary storage
    static double fix_density;
    /// Temporary storage
    static double stat_temper;

    /// The solver for calc_density()
    mroot<int ,mm_funct<int> > *density_mroot;

    /// The solver to compute \f$ h \f$ from \f$ \psi \f$.
    root<double,funct<double> > *psi_root;

    /// The solver for calc_density()
    root<int ,funct<int> > *meth2_root;

    /// The function which solves for \f$ h \f$ from \f$ \psi \f$.
    int solve_fun(double x, double &y, double &psi);

    /// Fix density for calc_density()
    int density_fun(size_t nv, const ovector_base &x, ovector_base &y, 
		    int &pa);

    /// Fix density for pair_density()
    int pair_density_fun(size_t nv, const ovector_base &x, ovector_base &y, 
			 int &pa);

#endif
  };

  template<> int io_tlate<eff_boson>::input
    (cinput *co, in_file_format *ins, eff_boson *b);
  template<> int io_tlate<eff_boson>::output
    (coutput *co, out_file_format *outs, eff_boson *b);
  template<> const char *io_tlate<eff_boson>::type();
  
  typedef io_tlate<eff_boson> eff_boson_io_type;

#ifndef DOXYGENP
}
#endif

#endif
