/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/nonrel_fermion.h>
#include <o2scl/classical.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

//--------------------------------------------
// nonrel_fermion class

nonrel_fermion::nonrel_fermion(double mass, double dof) : 
  part(mass,dof), fermion_T(mass,dof) {
  density_root=&def_density_root;
}

nonrel_fermion::~nonrel_fermion() {
}

int nonrel_fermion::calc_mu_zerot() {
  if (non_interacting) { nu=mu; ms=m; }
  if (inc_rest_mass) {
    kf=sqrt(2.0*ms*(nu-m));
  } else {
    kf=sqrt(2.0*ms*nu);
  }
  n=kf*kf*kf*g/6.0/pi2;
  ed=g*pow(kf,5.0)/20.0/pi2/ms;
  if (inc_rest_mass) ed+=n*m;
  pr=-ed+n*nu;
  en=0.0;
  return gsl_success;
}

int nonrel_fermion::calc_density_zerot() {
  if (non_interacting) { ms=m; }
  kf_from_density();
  nu=kf*kf/2.0/ms;
  ed=g*pow(kf,5.0)/20.0/pi2/ms;
  if (inc_rest_mass) {
    ed+=n*m;
    nu+=m;
  }
  pr=-ed+n*nu;
  en=0.0;
  
  if (non_interacting) { mu=nu; }
  return gsl_success;
}

int nonrel_fermion::calc_mu(const double temper) {
  double y, sy, spi, ey, int1, int2;

  if (non_interacting==true) { nu=mu; ms=m; }
  if (temper<=0.0) {
    return calc_mu_zerot();
  }

  if (inc_rest_mass) {
    y=(nu-m)/temper;
  } else {
    y=nu/temper;
  }

  // Number density
  n=gsl_sf_fermi_dirac_half(y)*sqrt(pi)/2.0;
  n*=g*pow(2.0*ms*temper,1.5)/4.0/pi2;

  // Energy density:
  ed=gsl_sf_fermi_dirac_3half(y)*0.75*sqrt(pi);

  if (inc_rest_mass) {
    
    // Finish energy density
    ed*=g*pow(2.0*ms*temper,2.5)/8.0/pi2/ms;
    ed+=n*m;
    
    // entropy density
    en=(5.0*(ed-n*m)/3.0-(nu-m)*n)/temper;
    
    // pressure
    pr=2.0*(ed-n*m)/3.0;
    
  } else {

    // Finish energy density
    ed*=g*pow(2.0*ms*temper,2.5)/8.0/pi2/ms;

    // entropy density
    en=(5.0*ed/3.0-nu*n)/temper;
    
    // pressure
    pr=2.0*ed/3.0;
    
  }

  if (!finite(nu) || !finite(n)) {
    O2SCL_ERR2_RET("Chemical potential or density in ",
		   "nonrel_fermion::calc_mu().",
		   gsl_efailed);
  }
  
  return gsl_success;
}

int nonrel_fermion::nu_from_n(const double temper) {

  double T=temper, nex;

  // Use initial value of nu for initial guess
  if (inc_rest_mass) {
    nex=-(nu-m)/temper;
  } else {
    nex=-nu/temper;
  } 
  
  // Make a correction if nex is too small and negative
  // (Note GSL_LOG_DBL_MIN is about -708)
  if (nex>-GSL_LOG_DBL_MIN*0.9) nex=-GSL_LOG_DBL_MIN/2.0;
  
  funct_mfptr<nonrel_fermion,double> mf(this,&nonrel_fermion::solve_fun);
  
  // Turn off convergence errors temporarily, since we'll
  // try again if it fails
  bool enc=density_root->err_nonconv;
  density_root->err_nonconv=false;
  int ret=density_root->solve(nex,T,mf);
  density_root->err_nonconv=enc;

  if (ret!=0 || density_root->last_conv!=0) {

    // If it failed, try to get a guess from classical particle
    
    classical c(m,g);
    c.inc_rest_mass=inc_rest_mass;
    c.non_interacting=non_interacting;
    c.n=n;
    c.calc_density(temper);
    nu=c.nu;
    if (inc_rest_mass) {
      nex=-(nu-m)/temper;
    } else {
      nex=-nu/temper;
    } 
    ret=density_root->solve(nex,T,mf);

    // If it failed again, add error information
    if (ret!=0) {
      O2SCL_ERR("Solver failed in nonrel_fermion::nu_from_n().",ret);
    }
  }

  if (inc_rest_mass) {
    nu=-nex*temper+m;
  } else {
    nu=-nex*temper;
  }
  
  return ret;
}

int nonrel_fermion::calc_density(const double temper) {
  double y, spi, ey, sy;

  if (n<=0.0) {
    nu=0.0;
    mu=0.0;
    ed=0.0;
    en=0.0;
    pr=0.0;
    return 0;
  }
  
  if (non_interacting==true) { nu=mu; ms=m; }
  if (temper<=0.0) {
    return calc_density_zerot();
  }

  double guess=nu;
  int ret=nu_from_n(temper);

  if (ret!=0) {
    O2SCL_ERR2("Get chemical potential failed in ",
		   "nonrel_fermion::calc_density().",
		   gsl_efailed);
  }

  if (non_interacting) { mu=nu; }

  if (inc_rest_mass) {
    y=-(nu-m)/temper;
  } else {
    y=-nu/temper;
  }

  // energy density
  ed=gsl_sf_fermi_dirac_3half(-y)*sqrt(pi)*0.75;

  if (inc_rest_mass) {
    
    // Finish energy density
    ed*=g*pow(2.0*ms*temper,2.5)/8.0/pi2/ms;
    ed+=n*m;
    
    // entropy density
    en=(5.0*(ed-n*m)/3.0-(nu-m)*n)/temper;
    
    // pressure
    pr=2.0*(ed-n*m)/3.0;
    
  } else {

    // Finish energy density
    ed*=g*pow(2.0*ms*temper,2.5)/8.0/pi2/ms;

    // entropy density
    en=(5.0*ed/3.0-nu*n)/temper;
    
    // pressure
    pr=2.0*ed/3.0;
    
  }
  
  if (non_interacting==true) { mu=nu; }
  
  if (ret!=0) return gsl_efailed;
  return gsl_success;

}

int nonrel_fermion::solve_fun(double x, double &y, double &T) {
  double nden;
  
  // If the argument to gsl_sf_fermi_dirac_half() is less
  // than GSL_LOG_DBL_MIN (which is about -708), then 
  // an underflow occurs. We just set nden to zero in this 
  // case
  
  if (x>fabs(GSL_LOG_DBL_MIN)) nden=0.0;
  else nden=gsl_sf_fermi_dirac_half(-x)*sqrt(pi)/2.0;

  nden*=g*pow(2.0*ms*T,1.5)/4.0/pi2;
  y=nden/n-1.0;
  
  return 0;
}

template<> int io_tlate<nonrel_fermion>::input
(cinput *co, in_file_format *ins, nonrel_fermion *f) {
  std::string t;
  co->object_in("fermion",ins,f,t);
  return 0;
}

template<> int io_tlate<nonrel_fermion>::output
(coutput *co, out_file_format *outs, nonrel_fermion *f) {
  std::string t;
  co->object_out("fermion",outs,f,0,0,t);
  return 0;
}

template<> const char *io_tlate<nonrel_fermion>::type() 
{ return "nonrel_fermion"; }

int nonrel_fermion_zerot::calc_mu_zerot() {
  if (non_interacting) { nu=mu; ms=m; }
  if (inc_rest_mass) {
    kf=sqrt(2.0*ms*(nu-m));
  } else {
    kf=sqrt(2.0*ms*nu);
  }
  n=kf*kf*kf*g/6.0/pi2;
  ed=g*pow(kf,5.0)/20.0/pi2/ms;
  if (inc_rest_mass) ed+=n*m;
  pr=-ed+n*nu;
  en=0.0;
  return gsl_success;
}

int nonrel_fermion_zerot::calc_density_zerot() {
  if (non_interacting) { ms=m; }
  kf_from_density();
  nu=kf*kf/2.0/ms;
  ed=g*pow(kf,5.0)/20.0/pi2/ms;
  if (inc_rest_mass) {
    ed+=n*m;
    nu+=m;
  }
  pr=-ed+n*nu;
  en=0.0;
  
  if (non_interacting) { mu=nu; }
  return gsl_success;
}

