/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_REL_FERMION_H
#define O2SCL_REL_FERMION_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/mroot.h>
#include <o2scl/inte.h>
#include <o2scl/fermion.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/gsl_inte_qagiu.h>
#include <o2scl/gsl_inte_qag.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Equation of state for a relativistic fermion

      This implements an equation of state for a relativistic fermion
      using direct integration. Define the degeneracy parameter 
      \f[
      \psi=(\nu-m^{*})/T 
      \f] 
      where \f$ \nu \f$ is the effective chemical potential and \f$
      m^{*} \f$ is the effective mass. For \f$ \psi \f$ greater than
      \ref deg_limit (degenerate regime), a finite interval integrator
      is used and for \f$ \psi \f$ less than \ref deg_limit
      (non-degenerate regime), an integrator over the interval from
      \f$ [0,\infty) \f$ is used. The upper limit on the degenerate
      integration is given by
      \f[
      \sqrt{(20 T+{\nu})^2-m^{*,2}}
      \f]
      
      The default integrators are gsl_inte_qag (for degenerate particles)
      and gsl_inte_qagiu (for non-degenerate particles).

      When the integrators provide numerical uncertainties, these
      uncertainties are stored in \ref unc. In the case of
      calc_density() and pair_density(), the uncertainty from the
      numerical accuracy of the solver is not included. (There is also
      a relatively small inaccuracy due to the mathematical evaluation
      of the integrands which is not included in \ref unc.)
     
      One way to improve the accuracy of the computation is just to 
      decrease the tolerances on the default integration objects. This 
      can be done, using, for example
      \code
      rel_fermion rf(1.0,2.0);
      rf.def_dit.tolx/=1.0e2;
      rf.def_dit.tolf/=1.0e2;
      rf.def_nit.tolx/=1.0e2;
      rf.def_nit.tolf/=1.0e2;
      \endcode
      which decreases the both the relative and absolute tolerances
      for both the degenerate and non-degenerate integrators. If one
      is using either the calc_density() or pair_density() functions,
      one may also have to improve the accuracy of the solver which
      determines the chemical potential from the density. For
      the default solver, this could be done with
      \code
      rf.def_density_root.tolx/=1.0e2;
      rf.def_density_root.tolf/=1.0e2;
      \endcode
      Of course if these tolerances are too small, the calculation
      may fail.

      \note This does not work with inc_rest_mass=false (3/30/09:
      There is a significant amount of testing code to ensure that
      it does work with inc_rest_mass=false, so this may have
      been fixed already.)

      \future Allow the user to change the upper limit on the
      degenerate integration and the hard-coded value of 200 in the
      integrands.
      \future It appears this doesn't compute the uncertainty in the
      chemical potential or density with calc_density(). This could
      be fixed.
      
  */
  class rel_fermion : public fermion_T {

  public:

    /// The critical degeneracy at which to switch integration techniques
    double deg_limit;

    /// Storage for the uncertainty
    fermion unc;

    /** \brief If true, use the present value of the chemical potential as
	a guess for the new chemical potential
    */
    bool guess_from_nu;
    
    /// Create a fermion with mass \c m and degeneracy \c g
    rel_fermion(double m=0.0, double g=0.0);
    virtual ~rel_fermion();
    
    /** 
	\brief Calculate properties as function of chemical potential
    */
    virtual int calc_mu(const double temper);

    /** \brief Calculate properties as function of density
     */
    virtual int calc_density(const double temper);

    /** \brief Calculate properties with antiparticles as function of
	chemical potential
    */
    virtual int pair_mu(const double temper);

    /** \brief Calculate properties with antiparticles as function of
	density
     */
    virtual int pair_density(const double temper);

    /// Calculate effective chemical potential from density
    virtual int nu_from_n(const double temper);
    
    /// Set integrators
    int set_inte(inte<double,funct<double> > &non_it, 
		 inte<double,funct<double> > &deg_it);
    
    /** \brief Set the solver for use in calculating the chemical
	potential from the density */
    int set_density_root(root<double,funct<double> > &rp) {
      density_root=&rp;
      return 0;
    }
    
    /// The default solver for calc_density().
    cern_mroot_root<double,funct<double> > def_density_root;

    /// The default integrator for degenerate fermions
    gsl_inte_qag<double,funct<double> > def_dit;

    /// The default integrator for non-degenerate fermions
    gsl_inte_qagiu<double,funct<double> > def_nit;

    /// Return string denoting type ("rel_fermion")
    virtual const char *type() { return "rel_fermion"; }

  protected:
    
#ifndef DOXYGEN_INTERNAL
    
    friend class io_tlate<rel_fermion>;

    /// The non-degenerate integrator
    inte<double,funct<double> > *nit;
    /// The degenerate integrator
    inte<double,funct<double> > *dit;
    /// The solver for calc_density()
    root<double,funct<double> > *density_root;

    /// The integrand for the density for non-degenerate fermions
    double density_fun(double u, double &pa);

    /// The integrand for the energy density for non-degenerate fermions
    double energy_fun(double u, double &pa);

    /// The integrand for the entropy density for non-degenerate fermions
    double entropy_fun(double u, double &pa);

    /// The integrand for the density for degenerate fermions
    double deg_density_fun(double u, double &pa);

    /// The integrand for the energy density for degenerate fermions
    double deg_energy_fun(double u, double &pa);

    /// The integrand for the entropy density for degenerate fermions
    double deg_entropy_fun(double u, double &pa);

    /// Solve for the chemical potential given the density
    int solve_fun(double x, double &yy, double &pa);

    /// Solve for the chemical potential given the density with antiparticles
    int pair_fun(double x, double &yy, double &pa);
    
#endif

  };

  template<> int io_tlate<rel_fermion>::input
    (cinput *co, in_file_format *ins, rel_fermion *f);
  template<> int io_tlate<rel_fermion>::output
    (coutput *co, out_file_format *outs, rel_fermion *f);
  template<> const char *io_tlate<rel_fermion>::type();
  template<> int io_tlate<rel_fermion>::stat_input
    (cinput *co, in_file_format *ins, rel_fermion *f);
  template<> int io_tlate<rel_fermion>::stat_output
    (coutput *co, out_file_format *outs, rel_fermion *f);

  typedef io_tlate<rel_fermion> rel_fermion_io_type;

#ifndef DOXYGENP
}
#endif

#endif
