/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/sn_fermion.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

//--------------------------------------------
// sn_fermion class

sn_fermion::sn_fermion(double mass, double dof) : 
  part(mass,dof), unc(mass,dof) {
  
  deg_limit=2.0;
  flimit=20.0;

  density_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::density_fun);
  energy_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::energy_fun);
  entropy_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::entropy_fun);
  density_T_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::density_T_fun);
  density_mu_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::density_mu_fun);
  entropy_T_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::entropy_T_fun);
  density_ms_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::density_ms_fun);

  deg_density_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::deg_density_fun);
  deg_energy_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::deg_energy_fun);
  deg_entropy_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::deg_entropy_fun);
  deg_density_T_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::deg_density_T_fun);
  deg_density_mu_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::deg_density_mu_fun);
  deg_entropy_T_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::deg_entropy_T_fun);
  deg_density_ms_fun_f=new funct_mfptr_noerr<sn_fermion,const double>
    (this,&sn_fermion::deg_density_ms_fun);

  density_root=&def_density_root;
  nit=&def_nit;
  dit=&def_dit;
  
  method=byparts;
}

sn_fermion::~sn_fermion() {

  delete density_fun_f;
  delete energy_fun_f;
  delete entropy_fun_f;
  
  delete density_mu_fun_f;
  delete density_T_fun_f;
  delete entropy_T_fun_f;
  delete density_ms_fun_f;
  
  delete deg_density_fun_f;
  delete deg_energy_fun_f;
  delete deg_entropy_fun_f;
  
  delete deg_density_mu_fun_f;
  delete deg_density_T_fun_f;
  delete deg_entropy_T_fun_f;
  delete deg_density_ms_fun_f;
  
}

int sn_fermion::set_inte(inte<const double,funct<const double> > &l_nit, 
			 inte<const double,funct<const double> > &l_dit) {
  nit=&l_nit;
  dit=&l_dit;
  return gsl_success;
}

int sn_fermion::calc_mu(const double temper) {
  int ret=gsl_success, iret;
  
  if (temper<=0.0) return fermion::calc_mu_zerot();
  if (non_interacting==true) { nu=mu; ms=m; }

  double T=temper, prefac=g/2.0/pi2;

  bool deg=false;
  if (inc_rest_mass && (nu-ms)/temper>deg_limit) deg=true;
  else if (!inc_rest_mass && (nu+m-ms)/temper>deg_limit) deg=true;

  if (deg==false) {
    
    iret=nit->integ_err(*density_fun_f,0.0,0.0,temper,n,unc.n);
    if (iret!=0) {
      O2SCL_ERR2("Density integration (ndeg) failed in ",
		     "sn_fermion::calc_mu().",gsl_efailed);
      ret=gsl_efailed;
    }
    n*=prefac;
    unc.n*=prefac;
    
    iret=nit->integ_err(*density_T_fun_f,0.0,0.0,temper,dndT,dunc.dndT);
    if (iret!=0) {
      O2SCL_ERR("dndT integration (ndeg) failed in sn_fermion::calc_mu().",
		    gsl_efailed);
      ret=gsl_efailed;
    }
    dndT*=prefac;
    dunc.dndT*=prefac;

    iret=nit->integ_err(*density_mu_fun_f,0.0,0.0,temper,dndmu,dunc.dndmu);
    if (iret!=0) {
      O2SCL_ERR("dndmu integration (ndeg) failed in sn_fermion::calc_mu().",
		    gsl_efailed);
      ret=gsl_efailed;
    }
    dndmu*=prefac;
    dunc.dndmu*=prefac;
    
    iret=nit->integ_err(*energy_fun_f,0.0,0.0,temper,ed,unc.ed);
    if (iret!=0) {
      O2SCL_ERR2("Energy integration (ndeg) failed in ",
		     "sn_fermion::calc_mu().",gsl_efailed);
      ret=gsl_efailed;
    }
    ed*=prefac;
    unc.ed*=prefac;
    
    iret=nit->integ_err(*entropy_fun_f,0.0,0.0,temper,en,unc.en);
    if (iret!=0) {
      O2SCL_ERR2("Entropy integration (ndeg) failed in ",
		     "sn_fermion::calc_mu().",gsl_efailed);
      ret=gsl_efailed;
    }
    en*=prefac;
    unc.en*=prefac;
    
    iret=nit->integ_err(*entropy_T_fun_f,0.0,0.0,temper,dsdT,dunc.dsdT);
    if (iret!=0) {
      O2SCL_ERR("dsdT integration (ndeg) failed in sn_fermion::calc_mu().",
		    gsl_efailed);
      ret=gsl_efailed;
    }
    dsdT*=prefac;
    dunc.dsdT*=prefac;

    iret=nit->integ_err(*density_ms_fun_f,0.0,0.0,temper,dndm,dunc.dndm);
    if (iret!=0) {
      O2SCL_ERR("dndm integration (ndeg) failed in sn_fermion::calc_mu().",
		    gsl_efailed);
      ret=gsl_efailed;
    }
    dndm*=prefac;
    dunc.dndm*=prefac;

  } else {

    double ulimit=sqrt(pow(flimit*temper+nu+m,2.0)-ms*ms);
    double llimit=2.0*sqrt(nu*nu+2.0*nu*m)-ulimit;
    
    if (llimit<0.0) llimit=0.0;
    if (!finite(ulimit)) {
      O2SCL_ERR("Variable ulimit not finite in sn_fermion:::calc_mu().",
		gsl_efailed);
      ret=gsl_efailed;
    }
    
    iret=dit->integ_err(*deg_density_fun_f,0.0,ulimit,temper,n,unc.n);
    if (iret!=0) {
      O2SCL_ERR2("Density integration (deg) failed in ",
		     "sn_fermion::calc_mu().",gsl_efailed);
      ret=gsl_efailed;
    }
    n*=prefac;
    unc.n*=prefac;
    
    iret=dit->integ_err(*deg_density_mu_fun_f,0.0,ulimit,temper,
			dndmu,dunc.dndmu);
    if (iret!=0) {
      O2SCL_ERR("dndmu integration (deg) failed in sn_fermion::calc_mu().",
		    gsl_efailed);
      ret=gsl_efailed;
    }
    dndmu*=prefac;
    dunc.dndmu*=prefac;
    
    iret=dit->integ_err(*deg_density_T_fun_f,0.0,ulimit,temper,dndT,dunc.dndT);
    if (iret!=0) {
      O2SCL_ERR("dndT integration (deg) failed in sn_fermion::calc_mu().",
		    gsl_efailed);
      ret=gsl_efailed;
    }
    dndT*=prefac;
    dunc.dndT*=prefac;

    iret=dit->integ_err(*deg_energy_fun_f,0.0,ulimit,temper,ed,unc.ed);
    if (iret!=0) {
      O2SCL_ERR("Energy integration (deg) failed in sn_fermion::calc_mu().",
		    gsl_efailed);
      ret=gsl_efailed;
    }
    ed*=prefac;
    unc.ed*=prefac;
    
    iret=dit->integ_err(*deg_entropy_fun_f,llimit,ulimit,temper,en,unc.en);
    if (iret!=0) {
      O2SCL_ERR2("Entropy integration (deg) failed in ",
		    "sn_fermion::calc_mu().",gsl_efailed);
      ret=gsl_efailed;
    }
    en*=prefac;
    unc.en*=prefac;
    
    iret=dit->integ_err(*deg_entropy_T_fun_f,0.0,ulimit,temper,dsdT,dunc.dsdT);
    if (iret!=0) {
      O2SCL_ERR("dsdT integration (deg) failed in sn_fermion::calc_mu().",
		    gsl_efailed);
      ret=gsl_efailed;
    }
    dsdT*=prefac;
    dunc.dsdT*=prefac;

    iret=dit->integ_err(*deg_density_ms_fun_f,0.0,ulimit,temper,dndm,dunc.dndm);
    if (iret!=0) {
      O2SCL_ERR("dndm integration (deg) failed in sn_fermion::calc_mu().",
		    gsl_efailed);
      ret=gsl_efailed;
    }
    dndm*=prefac;
    dunc.dndm*=prefac;
    
  }
  
  if (!finite(en)) {
    O2SCL_ERR("Entropy not finite in sn_fermion::calc_mu().",gsl_efailed);
    ret=gsl_efailed;
  }
  pr=-ed+temper*en+nu*n;
  
  return ret;
}

int sn_fermion::nu_from_n(const double temper) {
  double nex;
  double T=temper;

  nex=nu/temper;
  funct_mfptr<sn_fermion,const double> mf(this,&sn_fermion::solve_fun);
  int ret=density_root->solve(nex,temper,mf);
  nu=nex*temper;

  if (ret!=0) {
    O2SCL_ERR_RET("Solver failed in sn_fermion::nu_from_n().",gsl_efailed);
  }
  return 0;
}

int sn_fermion::calc_density(const double temper) {
  int mret, nret;
  double T=temper;
  
  if (non_interacting==true) { ms=m; nu=mu; }
  
  mret=nu_from_n(temper);
  if (mret!=0) {
    O2SCL_ERR("nu_from_n() failed in sn_fermion::calc_density().",
		  gsl_efailed);
  }
  
  if (non_interacting) { mu=nu; }
  
  nret=calc_mu(temper);
  if (mret!=0) {
    O2SCL_ERR("Calc_mu() failed in sn_fermion::calc_density().",
		  gsl_efailed);
  }
  
  if (nret!=0 || mret!=0) return gsl_efailed;
  return gsl_success;
}

double sn_fermion::deg_density_fun(double k, const double &temper) {
  double E=sqrt(k*k+ms*ms)-m, ret;
  ret=k*k*fermi_function(E,nu,temper,flimit);
  return ret;
}

double sn_fermion::deg_density_T_fun(double k, const double &temper) {
  double E=sqrt(k*k+ms*ms), ret;
  if (method==direct) {
    E-=m;
    ret=k*k*(E-nu)/temper/temper*
      fermi_function(E,nu,temper,flimit)*
      (1.0-fermi_function(E,nu,temper,flimit));
  } else {
    ret=(2.0*k*k/temper+E*E/temper-E*(nu+m)/temper-k*k*(nu+m)/temper/E)*
      fermi_function(E-m,nu,temper,flimit);
  }
  return ret;
}

double sn_fermion::deg_density_mu_fun(double k, const double &temper) {
  double E=sqrt(k*k+ms*ms), ret;
  if (method==direct) {
    E-=m;
    ret=k*k/temper*fermi_function(E,nu,temper,flimit)*
      (1.0-fermi_function(E,nu,temper,flimit));
  } else {
    ret=(E*E+k*k)/E*fermi_function(E-m,nu,temper,flimit);
  }
  return ret;
}

double sn_fermion::deg_energy_fun(double k, const double &temper) {
  double E=sqrt(k*k+ms*ms)-m, ret;
  ret=k*k*E*fermi_function(E,nu,temper,flimit);
  return ret;
}

double sn_fermion::deg_entropy_fun(double k, const double &temper) {
  double E=sqrt(k*k+ms*ms)-m, nx, ret, nm1;
  
  // If the argument to the exponential is really small, then the
  // value of the integrand is just zero
  if (((E-nu)/(temper))<-200.0) {
    ret=0.0;
    // Otherwise, if the argument to the exponential is still small,
    // then addition of 1 makes us lose precision, so we use an
    // alternative:
  } else if (((E-nu)/temper)<-flimit) {
    // Should this really be E/T-nu/T or (E-nu)/T ?
    nm1=-exp(E/temper-nu/temper);
    ret=k*k*nm1*log(-nm1);
  } else {
    nx=fermi_function(E,nu,temper,flimit);
    if (nx==0.0) ret=0.0;
    else ret=-k*k*(nx*log(nx)+(1.0-nx)*log(1.0-nx));
  }
  if (!finite(ret)) {
    ret=0.0;
    O2SCL_ERR("Entropy not finite in sn_fermion::deg_entropy_fun().",
	      gsl_efailed);
  }

  return ret;
}
  
double sn_fermion::deg_entropy_T_fun(double k, const double &temper) {
  double E=sqrt(k*k+ms*ms), ret;
  if (method==direct) {
    E-=m;
    ret=k*k*fermi_function(E,nu,temper,flimit)*
      (1.0-fermi_function(E,nu,temper,flimit))*pow(E-nu,2.0)/pow(temper,3.0);
  } else {
    ret=(E-m-nu)/E/temper/temper*
      (pow(E,3.0)+3.0*E*k*k-(E*E+k*k)*(nu+m))*
      fermi_function(E-m,nu,temper,flimit);
  }
  return ret;
}

double sn_fermion::deg_density_ms_fun(double k, const double &temper) {
  double E=sqrt(k*k+ms*ms), ret;
  if (method==direct) {
    E-=m;
    ret=-k*k*ms/(E+m)/temper*fermi_function(E,nu,temper,flimit)*
      (1.0-fermi_function(E,nu,temper,flimit));
  } else {
    ret=-ms*fermi_function(E-m,nu,temper,flimit);
  }
  return ret;
}

double sn_fermion::density_fun(double u, const double &temper) {
  double k=u*(temper);
  double E=sqrt(k*k+ms*ms)-m, ret;
  ret=(temper)*k*k*fermi_function(E,nu,temper,flimit);
  return ret;
}

double sn_fermion::density_T_fun(double u, const double &temper) {
  double k=u*(temper);
  double E=sqrt(k*k+ms*ms), ret;
  if (method==direct) {
    E-=m;
    ret=k*k*(E-nu)/temper*
      fermi_function(E,nu,temper,flimit)*
      (1.0-fermi_function(E,nu,temper,flimit));
  } else {
    ret=(2.0*k*k/temper+E*E/temper-E*(nu+m)/temper-k*k*(nu+m)/temper/E)*
      temper*fermi_function(E-m,nu,temper,flimit);
  }
  return ret;
}

double sn_fermion::density_mu_fun(double u, const double &temper) {
  double k=u*(temper);
  double E=sqrt(k*k+ms*ms), ret;
  if (method==direct) {
    E-=m;
    ret=k*k*fermi_function(E,nu,temper,flimit)*
      (1.0-fermi_function(E,nu,temper,flimit));
  } else {
    ret=temper*(E*E+k*k)/E*fermi_function(E-m,nu,temper,flimit);
  }
  return ret;
}

double sn_fermion::energy_fun(double u, const double &temper) {
  double k=u*(temper);
  double E=sqrt(k*k+ms*ms)-m, ret;
  ret=k*k*E*fermi_function(E,nu,temper,flimit)*(temper);
  return ret;
}

double sn_fermion::entropy_fun(double u, const double &temper) {
  double k=u*(temper);
  double E=sqrt(k*k+ms*ms)-m, nx, ret, nm1;

  // If the argument to the exponential is really small, then the
  // value of the integrand is just zero
  if (((E-nu)/(temper))<-200.0) {
    ret=0.0;
    // Otherwise, if the argument to the exponential is still small,
    // then addition of 1 makes us lose precision, so we use an
    // alternative:
  } else if (((E-nu)/temper)<-30.0) {
    nm1=-exp(E/(temper)-nu/(temper));
    ret=k*k*nm1*log(-nm1)*(temper);
  } else {
    nx=fermi_function(E,nu,temper,flimit);
    if (nx==0.0) ret=0.0;
    else ret=-k*k*(nx*log(nx)+(1.0-nx)*log(1.0-nx))*temper;
  }

  return ret;
}

double sn_fermion::entropy_T_fun(double u, const double &temper) {
  double k=u*temper;
  double E=sqrt(k*k+ms*ms), ret;
  if (method==direct) {
    E-=m;
    ret=temper*k*k*fermi_function(E,nu,temper,flimit)*
      (1.0-fermi_function(E,nu,temper,flimit))*pow(E-nu,2.0)/pow(temper,3.0);
  } else {
    ret=(E-m-nu)/E/temper*
      (pow(E,3.0)+3.0*E*k*k-(E*E+k*k)*(nu+m))*
      fermi_function(E-m,nu,temper,flimit);
  }
  return ret;
}

double sn_fermion::density_ms_fun(double u, const double &temper) {
  double k=u*temper;
  double E=sqrt(k*k+ms*ms), ret;
  if (method==direct) {
    E-=m;
    ret=-k*k*ms/(E+m)/temper*fermi_function(E,nu,temper,flimit)*
      (1.0-fermi_function(E,nu,temper,flimit));
  } else {
    ret=-ms*fermi_function(E-m,nu,temper,flimit);
  }
  ret*=temper;
  return ret;
}

int sn_fermion::solve_fun(double x, double &yy, const double &temper) {
  double nden;
  
  nu=temper*x;
  
  // 6/6/03 - Should this be included? I think yes!
  if (non_interacting) mu=nu;
  
  if (nu/temper < deg_limit) {
    
    nden=nit->integ(*density_fun_f,0.0,0.0,temper);
    nden*=g/2.0/pi2;
    
    yy=(n-nden)/n;

  } else {
    
    double ulimit=sqrt(pow(flimit*temper+nu+m,2.0)-ms*ms);
    nden=dit->integ(*deg_density_fun_f,0.0,ulimit,temper);
    nden*=g/2.0/pi2;
    
    yy=(n-nden)/n;
  }

  return gsl_success;
}

int sn_fermion::pair_mu(const double temper) {
  int ret;

  if (non_interacting) { nu=mu; ms=m; }

  sn_fermion antip(ms,g);
  anti(antip);
  
  antip.massless_root=massless_root;
  antip.nit=nit;
  antip.dit=dit;
  antip.density_root=density_root;

  ret=calc_mu(temper);
  if (ret!=0) {
    O2SCL_ERR("Function calc_mu() failed in sn_fermion::pair_mu().",ret);
  }

  ret=antip.calc_mu(temper);
  if (ret!=0) {
    O2SCL_ERR2("calc_mu() failed for antiparticles in ",
		   "sn_fermion::pair_mu().",ret);
  }

  n-=antip.n;
  pr+=antip.pr;
  ed+=antip.ed;
  en+=antip.en;
  dsdT+=antip.dsdT;
  dndT+=antip.dndT;
  dndmu+=antip.dndmu;

  return ret;
}

int sn_fermion::pair_density(const double temper) {
  double nex;
  int ret;
  
  if (temper<=0.0) return fermion::calc_density_zerot();

  if (non_interacting==true) { nu=mu; ms=m; }
  
  double T=temper;

  nex=nu/temper;
  funct_mfptr<sn_fermion,const double> mf(this,&sn_fermion::pair_fun);
  ret=density_root->solve(nex,temper,mf);
  nu=nex*temper;

  if (non_interacting==true) { mu=nu; }
  
  ret=pair_mu(temper);

  return ret;

}

int sn_fermion::pair_fun(double x, double &yy, const double &temper) {
  double nden;
  
  nu=temper*x;
  
  if (non_interacting) mu=nu;
  
  if (nu/temper < deg_limit) {
    
    nden=nit->integ(*density_fun_f,0.0,0.0,temper);
    nden*=g/2.0/pi2;
    yy=nden;

  } else {
    
    double ulimit=sqrt(pow(flimit*temper+nu+m,2.0)-ms*ms);
    nden=dit->integ(*deg_density_fun_f,0.0,ulimit,temper);
    nden*=g/2.0/pi2;
    yy=nden;
  }
  
  if (inc_rest_mass) nu=-temper*x;
  else nu=-temper*x-2.0*m;
  
  if (nu/temper < deg_limit) {
    
    nden=nit->integ(*density_fun_f,0.0,0.0,temper);
    nden*=g/2.0/pi2;
    yy-=nden;
    
  } else {
    
    double ulimit=sqrt(pow(flimit*temper+nu+m,2.0)-ms*ms);
    nden=dit->integ(*deg_density_fun_f,0.0,ulimit,temper);
    nden*=g/2.0/pi2;
    yy-=nden;
  }
  
  yy=yy/n-1.0;

  return gsl_success;
}

template<> int io_tlate<sn_fermion>::input(cinput *co, in_file_format *ins, 
					   sn_fermion *f) {
  /*
  deriv_fermion *g=(deriv_fermion *)f;
  deriv_fermion *h=(deriv_fermion *)(&f->unc);
  std::string t;
  co->object_in("deriv_fermion",ins,g,t);
  co->object_in("deriv_fermion",ins,h,t);
  ins->int_in(f->method);
  ins->double_in(f->deg_limit);
  ins->double_in(f->flimit);
  */
  return 0;
}

template<> int io_tlate<sn_fermion>::output(coutput *co, 
					    out_file_format *outs, 
					    sn_fermion *f) {
  /*
  deriv_fermion *g=(deriv_fermion *)f;
  deriv_fermion *h=(deriv_fermion *)(&f->unc);
  std::string t;
  co->object_out("deriv_fermion",outs,g,0,0,t);
  co->object_out("deriv_fermion",outs,h,0,0,t);
  outs->int_out(f->method);
  outs->double_out(f->deg_limit);
  outs->double_out(f->flimit);
  */
  return 0;
}

template<> const char *io_tlate<sn_fermion>::type() {
  return "sn_fermion";
}

